/**
 * UNICOS
 * Copyright (C) CERN 2014 All rights reserved
 */
/**@file

@name LIBRARY: unRecipeFunctions.ctl

@author: Ivan Prieto Barreiro (EN-ICE-SIC)

Creation Date: April 2011 

version 1.4.0

Modification History:
  01/11/2011: Ivan
    - v1.1.0: The device type names have been modified to use the "CPC_" prefix.
	
  01/08/2012: Ivan
	- v1.2.0 : Adding new functionalities to the component:
		- Dynamic creation of recipe classes.
		- Recipe activation in multiple PLCs.

  09/12/2013: Ivan
	- v1.3.4 : Bug fixing and improvements in existing functionalities.
    
  09/09/2014: Ivan
    - v1.4.0 : Bug fixing and improvements in existing functionalities.
		
External Functions: 
  .  
  
Internal Functions:
  .
*/

#uses "fwAccessControl/fwAccessControl.ctc"
#uses "fwDevice/fwDevice.ctl"
#uses "fwGeneral/fwException.ctl"
#uses "fwGeneral/fwGeneral.ctl"
#uses "fwGeneral/fwExceptionHandling.ctl"
#uses "fwConfigurationDB/fwConfigurationDB.ctl"
#uses "fwConfigurationDB/fwConfigurationDB_Recipes.ctl"
#uses "fwConfigurationDB/fwConfigurationDB_Setup.ctl"
#uses "unGenericFunctions_core.ctl"
#uses "unGenericObject.ctl"
#uses "unGenericDpFunctions.ctl"
#uses "unExportTree.ctl"
#uses "unExportDevice.ctl"
#uses "unGraphicalFrame.ctl"
#uses "unDistributedControl.ctl"
#uses "unImportDevice.ctl"
#uses "unSelectDeselectHMI.ctl"
#uses "unSendMessage.ctl"
#uses "unProgressBar.ctl"
#uses "unSystemIntegrity_unicosPLC.ctl"
#uses "unSystemIntegrity.ctl"
#uses "unicos_declarations.ctl"
#uses "unicos_declarations_core.ctl"
#uses "unConfigGenericFunctions.ctl"
#uses "unConfigSOFT_FE.ctl"
#uses "libunCore/unApplicationFilter.ctl"
#uses "cpcObjects/cpcControllerConfig.ctl"
#uses "cpcObjects/cpcProcessControlObjectConfig.ctl"
#uses "cpcGenericObject.ctl"
#uses "unRcpConstant_declarations.ctl"
#uses "unRcpFunctions_activation.ctl"
#uses "unRcpDbFunctions.ctl"

//@{
const int unRecipeFunctions_ERROR_NoRecipeClass=-1;
const int unRecipeFunctions_ERROR_InvalidElements=-2;
const int unRecipeFunctions_ERROR_NoRecipeInstance=-3;

/********************************************************** 
 *  Indexes for the elements in the recipe instance data  *
 **********************************************************/
const int RCP_INSTANCE_ID            = 1;
const int RCP_INSTANCE_NAME          = 2;
const int RCP_INSTANCE_CLASS         = 3;
const int RCP_INSTANCE_AUTHOR        = 4;
const int RCP_INSTANCE_VERSION       = 5;
const int RCP_INSTANCE_CREATED       = 6;
const int RCP_INSTANCE_LAST_USAGE    = 7;
const int RCP_INSTANCE_LAST_MODIFIED = 8;
const int RCP_INSTANCE_LAST_ACT_USER = 9;
const int RCP_INSTANCE_LAST_MOD_USER = 10;
const int RCP_INSTANCE_VALUES        = 11;
const int RCP_INSTANCE_PREDEFINED    = 12;


/******************************************************* 
 *  Indexes for the elements in the recipe class data  *
 *******************************************************/
const int RCP_CLASS_NAME                 =  1;
const int RCP_CLASS_TYPE                 =  2;
const int RCP_CLASS_DESC                 =  3;
const int RCP_CLASS_ELEMENTS             =  4;
const int RCP_CLASS_ELEMENTS_DESC        =  5;
const int RCP_CLASS_ELEMENTS_RANGE       =  6;
const int RCP_CLASS_ELEMENTS_UNIT        =  7;
const int RCP_CLASS_INSTANCES            =  8;
const int RCP_CLASS_EDITABLE             =  9;
const int RCP_CLASS_LAST_MODIFIED        = 10;
const int RCP_CLASS_LAST_ACTIVATED       = 11;
const int RCP_CLASS_LAST_ACTIVATION_TIME = 12;


/**************************************************************************
 *      Constants used as return value while getting the initial recipes  *
 **************************************************************************/
const int RCP_INITIAL_OK		 =  0;	// The initial recipe of the class has been found
const int RCP_INITIAL_NO_DEFINED = -1;	// There is no initial recipe for the recipe class
const int RCP_INITIAL_TOO_MANY	 = -2;	// There are several initial recipes for the class
const int RCP_INITIAL_UNDEFINED	 = -3;	// Other error while getting the initial recipe			


/************************************
 *      Constants to define colors  *
 ************************************/
const string CELL_EDIT_COLOR 			   = "{204,255,153}";
const string CELL_VALUE_OUT_OF_RANGE_COLOR = "{255,64,64}";
const string CELL_ERROR_COLOR              = "{255,0,0}";
const string CELL_BG_COLOR                 = "{255,255,255}";
const string INACTIVE_RCP_COLOR            = "{251,255,72}";
const string ACTIVE_RCP_COLOR              = "{0,245,0}";

/****************************** 
 *      Global   Variables    *
 *****************************/
global mapping unRecipe_rangeElements;  // Map where key=device type, value=dpe containing the range definition for the device type
global int g_editingRowNumber = -1;
global string g_editingColName = "";
global bool g_bCreatingRecipeInstance = false;
// Map containing the selected recipe in each opened faceplace
// Key: $sDpName of the faceplate
// Value: Dp name of the selected recipe in the faceplate
global mapping unRecipeFaceplate_selectedRecipe;
//--------------------------------------------------------------------------------------------------------------------------
/** 
 * Gets the formatted values of a set of DPEs.
 * @param dpNames         - [IN] List of dp names in the UNICOS format.
 * @param values          - [IN] List of values for the specified dpNames.
 * @param formattedValues - [OUT] Formatted values.
 * @param exceptionInfo   - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_formatValues(dyn_string dpNames, dyn_string values, 
                                     dyn_anytype &formattedValues, dyn_string &exceptionInfo)
{
  string rangeElement, deviceType, rangeMin, rangeMax, pvssDpName;
  dyn_string dsOnlineDpes;
  float fFormattedValue, fValue;
  bool bOk;
  
  if (dynlen(dpNames)!=dynlen(values)) {
    fwException_raise(exceptionInfo, "ERROR", "The number of DPEs is different than the number of values!", "");
    return;
  }  
  
  for (int i=1; i<=dynlen(dpNames); i++) {  
    dynClear(exceptionInfo);
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(dpNames[i], pvssDpName);  
    // Get the DPE name (e.g. .ProcessOutput.MPosR")
    string dp=unGenericDpFunctions_getDpName(pvssDpName);
    int pos = strpos(pvssDpName, dp);
    string element = substr(pvssDpName, pos+strlen(dp), strlen(pvssDpName)-strlen(dp));  

    if (!dpExists(pvssDpName)) {
      formattedValues[i] = values[i];
      continue;
    }
    
    deviceType = dpTypeName(pvssDpName);

	// Removes the white spaces to format properly the exponential values (1.5 e1 => 1.5e1)
	strreplace(values[i], " ", "");
	if (deviceType=="CPC_DigitalParameter") {	
		if (values[i]==0) {
			formattedValues[i] = "FALSE";
			continue;
		}
		else if (values[i]==1) {
			formattedValues[i] = "TRUE";
			continue;
		}
			
		string value = strtoupper(values[i]);
		if (value=="TRUE" || value=="FALSE") {
			formattedValues[i] = value;
		} 
		continue;
	}
		
    // Get the dpe where the range is defined for the current recipe element
    if (!mappingHasKey(unRecipe_rangeElements, deviceType+":"+element)){
      _unRecipeFunctions_getRangeElementFromPVSSDPE(pvssDpName, rangeElement, exceptionInfo);  
      unRecipe_rangeElements[deviceType+":"+element] = rangeElement;
    } else {
      rangeElement = unRecipe_rangeElements[deviceType+":"+element];    
    }   
    
    string format="";
    _unRecipeFunctions_getInstanceRange(pvssDpName, rangeElement, rangeMin, rangeMax, exceptionInfo);
    if (rangeMin!="" && rangeMax!=""){
      format = dpGetFormat(unGenericDpFunctions_getDpName(pvssDpName)+rangeElement);
    }
    
    if (format!="") {
      formattedValues[i]=unGenericObject_FormatValue(format, values[i]);
	  continue;
    } 
    
	// if the value is a float, format the value using the online DPE format
	if (dpElementType(dpNames[i]) == DPEL_FLOAT) {
        string sDpName = unGenericDpFunctions_getDpName(pvssDpName)+rangeElement;
        _unRecipeFunctions_getOnlineValueDpes(makeDynString(unGenericDpFunctions_getDpName(pvssDpName)+rangeElement), dsOnlineDpes, exceptionInfo);
        if (!dynlen(exceptionInfo)) {
		  format = dpGetFormat(dsOnlineDpes[1]);
		  unGenericObject_StringToFloat(values[i], fValue, bOk);
		  if (!bOk) {
			fwException_raise(exceptionInfo, "ERROR", "Float conversion error :" + values[i], "");
		  }
		  formattedValues[i] = unGenericObject_FormatValue(format, fValue);
        } else {
          formattedValues[i] = values[i];
        }
	} else {
		// Applies the default PVSS format
		string s = values[i];  
		formattedValues[i] = s;
	}
  }  
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Show the online values in the recipe instance panel for the specified recipe instance.
 * @param sRecipeDpName - [IN] String containing the recipe instance DP name.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_displayOnlineValues(string sRecipeDpName, dyn_string &exceptionInfo)
{
  dyn_string dsDpNames;
  dyn_anytype daOnlineValues;
  
  if (!shapeExists("RecipeElementsCopy")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    DebugTN(exceptionInfo);
    return;	
  }
  
  _unRecipeFunctions_normalizeDp(sRecipeDpName);
  _unRecipeFunctions_writeInRecipeLog("Loading recipe online values.", true);
  dpSet(sRecipeDpName + ".ProcessInput.State", UN_RCP_INSTANCE_STATE_LOADING_VALUES);
  getValue("RecipeElementsCopy", "getColumnN", 1, dsDpNames);
  
  unRecipeFunctions_getOnlineValues(dsDpNames, daOnlineValues, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    DebugTN(exceptionInfo); 
    return;
  }
  
  _unRecipeFunctions_editRecipeValuesAndShowDifferences(daOnlineValues, exceptionInfo);
  _unRecipeFunctions_checkRecipeValues(dsDpNames, daOnlineValues, TRUE, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2], true, true );
  }
  //(dyn_string dsDps, dyn_anytype values, bool bShowErrorLines, dyn_string &exceptionInfo) 
  dpSet(sRecipeDpName + ".ProcessInput.State", UN_RCP_INSTANCE_STATE_EDIT);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Edit the values of the recipe displayed in the table and highlight the values that are different.
 * @param daValues      - [IN]  Recipe values.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_editRecipeValuesAndShowDifferences(dyn_anytype daValues, dyn_string &exceptionInfo) {
  int iIndexCol, iDpCol, iAliasCol, iDescriptionCol, iValueCol, iUnitCol, iRangeCol;
  dyn_string dsDpNames, dsAlias, dsDescription, dsUnit, dsRange;
  dyn_anytype daOnlineValues, daRecipeValues;
  dyn_int diIndex, diRowStyles, diRowStyleIndex, diDiffRowIndex;
  
  if (!shapeExists("RecipeElements") || !shapeExists("RecipeElementsCopy")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    DebugTN(exceptionInfo);
    return;	
  }

  getMultiValue("RecipeElementsCopy", "nameToColumn", "Index", iIndexCol,
                "RecipeElementsCopy", "nameToColumn", "DeviceDp", iDpCol,
                "RecipeElementsCopy", "nameToColumn", "Alias", iAliasCol,
                "RecipeElementsCopy", "nameToColumn", "Description", iDescriptionCol,
                "RecipeElementsCopy", "nameToColumn", "Value", iValueCol,
                "RecipeElementsCopy", "nameToColumn", "Unit", iUnitCol,
                "RecipeElementsCopy", "nameToColumn", "Range", iRangeCol,
                "RecipeElementsCopy", "getColumnN", iIndexCol, diIndex,
                "RecipeElementsCopy", "getColumnN", iDpCol, dsDpNames,
                "RecipeElementsCopy", "getColumnN", iAliasCol, dsAlias,
                "RecipeElementsCopy", "getColumnN", iDescriptionCol, dsDescription,
                "RecipeElementsCopy", "getColumnN", iValueCol, daRecipeValues,
                "RecipeElementsCopy", "getColumnN", iUnitCol, dsUnit,
                "RecipeElementsCopy", "getColumnN", iRangeCol, dsRange);

  _unRecipeFunctions_getRowStylesForValueComparison(daRecipeValues, daValues, diRowStyleIndex, diRowStyles, diDiffRowIndex);
  setMultiValue("RecipeElements", "deleteAllLines",
     "RecipeElements", "appendLines", dynlen(daRecipeValues),
     "Index", diIndex,
     "DeviceDp", dsDpNames,
     "Alias", dsAlias,
     "Description", dsDescription,
     "Value", daValues,
     "Unit", dsUnit,
     "Range", dsRange);
  
  _unRecipeFunctions_setTableColumnBGColor("RecipeElements", "Value", CELL_EDIT_COLOR);
  setMultiValue("RecipeElements", "rowFontType", diRowStyleIndex, diRowStyles);
  unRecipeFunctions_writeDifferentOnlineValuesReport(diDiffRowIndex, diIndex, dsDpNames, dsAlias, daRecipeValues, daValues);
  _unRecipeFunctions_writeInRecipeLog("Recipe values loaded.", true);
  

}
//------------------------------------------------------------------------------------------------------------------------
void unRecipeFunctions_writeDifferentOnlineValuesReport(dyn_int diDiffRowIndex, dyn_int diIndex, dyn_string dsDpNames, 
            dyn_string dsAlias, dyn_anytype daRcpValues, dyn_anytype daOnlineValues) {

  int iLen = dynlen(diDiffRowIndex);
  if (iLen>0) {
    string sLogMessages = "Recipe values and loaded values are different:";
    for (int i=1; i<=iLen; i++) {
      string sDpe = dsDpNames[(diDiffRowIndex[i])];
      dyn_string sDpeSplit = strsplit(sDpe, ".");
      string sAlias = dsAlias[(diDiffRowIndex[i])] + "." + sDpeSplit[(dynlen(sDpeSplit))];
      sLogMessages += "\nDifferent values on row index " + diDiffRowIndex[i] /*diIndex[(diDiffRowIndex[i])]  */
                   + ". Device: " + sAlias + " => Recipe Value: " + daRcpValues[(diDiffRowIndex[i])] 
                   + ", Loaded Value: " + daOnlineValues[(diDiffRowIndex[i])];
    }
    _unRecipeFunctions_writeInRecipeLog(sLogMessages, false, true);
  } else {
    _unRecipeFunctions_writeInRecipeLog("Recipe values and online values are equal.", false, true);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Compares the value lists daValues1 and daValues2 and set:
 *  - Row indexes (diRowStyleIndex).
 *  - Row styles (diRowStyles): normal (0) if the values are the same, bold (1) if the values are different.
 *  - Index of rows with different values (diDiffRowIndex).
 * @param daValues1       - [IN] First list of values to compare.
 * @param daValues2       - [IN] Second list of values to compare.
 * @param diRowStyleIndex - [OUT] Row index for the styles.
 * @param diRowStyles     - [OUT] Styles for the table rows.
 * @param diDiffRowIndex  - [OUT] Index of the rows with different values.
 * @return FALSE if the lenght of the lists is different, otherwise TRUE.
 */
bool _unRecipeFunctions_getRowStylesForValueComparison(dyn_anytype daValues1, dyn_anytype daValues2, dyn_int &diRowStyleIndex, dyn_int &diRowStyles, dyn_int &diDiffRowIndex) 
{
  if (dynlen(daValues1) != dynlen(daValues2) ) {
    return false;
  }
  
  int iLen = dynlen(daValues1);
  for (int i=1; i<=iLen; i++) {
    dynAppend(diRowStyleIndex, (i-1));
    if (daValues1[i] != daValues2[i]) {
      dynAppend(diRowStyles, 1);
      dynAppend(diDiffRowIndex, i);
    } else {
      dynAppend(diRowStyles, 0);
    }
  }

  return true;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the online values of a recipe instance.
 * @param dsDpNames 	- [IN] DP Names of the elements which online values are required.
 * @param values 		- [OUT] Online formatted values corresponding to the recipe class DPEs.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */ 
void unRecipeFunctions_getOnlineValues(dyn_string dsDpNames, dyn_anytype &values, dyn_string &exceptionInfo)
{
  dyn_string outputDpNames;
  dyn_anytype formattedValues;
  dyn_dyn_mixed recipeObject;

  dynClear(values);
  
   // Get recipe dpes
  _unRecipeFunctions_getOnlineValueDpes (dsDpNames, outputDpNames, exceptionInfo);
  dpGet(outputDpNames, values);
  _unRecipeFunctions_formatValues(dsDpNames, values, formattedValues, exceptionInfo);  
  
  values = formattedValues;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the range of the specified DP.
 * @param sDpName - [IN] The DP which range is required.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return String containing the valid range for the specified DP (format: [min, max]).
 */
string unRecipeFunctions_getDeviceRange(string sDpName, dyn_string &exceptionInfo)
{
  string range, deviceType, dp, element, rangeMin, rangeMax, rangeElement;
  
  if (!dpExists(sDpName)) {
    return "";
  }
  
  deviceType = dpTypeName(sDpName);   
  dp=unGenericDpFunctions_getDpName(sDpName);
  int pos = strpos(sDpName, dp);
  element = substr(sDpName, pos+strlen(dp), strlen(sDpName)-strlen(dp));
  dyn_string splitElement = strsplit(element, ".");

  // If the device type is a DigitalParameter the range is [FALSE, TRUE]
  if (deviceType=="CPC_DigitalParameter") {
    range = "[FALSE, TRUE]";
    return range;
  }
       
  if (!mappingHasKey(unRecipe_rangeElements, deviceType+":"+element)){
    _unRecipeFunctions_getRangeElementFromPVSSDPE(sDpName, rangeElement, exceptionInfo);
    if (dynlen(exceptionInfo)) {
        return "";
    }
    unRecipe_rangeElements[deviceType+":"+element] = rangeElement;
  } else {
    rangeElement = unRecipe_rangeElements[deviceType+":"+element];
  }

  // Get the range values
  _unRecipeFunctions_getInstanceRange(sDpName, rangeElement, rangeMin, rangeMax, exceptionInfo);       
  if (rangeMin!="" && rangeMax!="") {
    range = "[" + rangeMin +", " + rangeMax + "]";  
  }
	
  return range;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the measurement unit of a DP.
 * @param sDpName - [IN] The datapoint name.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return String containing the measurement unit of the specified DP.
 */
string unRecipeFunctions_getDpeUnit(string sDpName, dyn_string &exceptionInfo) {
  string deviceType, dp, element, rangeElement;
  
  if (!dpExists(sDpName)) {
    return "";
  }
  
  deviceType = dpTypeName(sDpName);   
  dp=unGenericDpFunctions_getDpName(sDpName);
  int pos = strpos(sDpName, dp);
  element = substr(sDpName, pos+strlen(dp), strlen(sDpName)-strlen(dp));
  dyn_string splitElement = strsplit(element, ".");
  
  // Hardcoding values for Controller: MTi, MTd, MTds (IS-1456)
  if (deviceType == UN_CONFIG_CPC_CONTROLLER_DPT_NAME 
        && (element==".ProcessOutput.MTi" || element==".ProcessOutput.MTd" || element==".ProcessOutput.MTds") ) 
  {
    return "s.";
  }      
  
  if (!mappingHasKey(unRecipe_rangeElements, deviceType+":"+element)){
    _unRecipeFunctions_getRangeElementFromPVSSDPE(sDpName, rangeElement, exceptionInfo);
    if (dynlen(exceptionInfo)) {
      return "";
    }
    unRecipe_rangeElements[deviceType+":"+element] = rangeElement;
  } else {
    rangeElement = unRecipe_rangeElements[deviceType+":"+element];
  }
  
  return dpGetUnit(dp + rangeElement);
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the DPEs associated to each device type in a recipe type.
 * @param sRecipeType - [IN] Recipe type name.
 * @param map - [OUT] Mapping variable where Key=Device type name, Value=List containing the DPEs 
 *              included in the recipe type for the specified key (device type).
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getTypeMapping(string sRecipeType, mapping &map, dyn_string &exceptionInfo)
{
  dyn_dyn_string rtData;
  _fwConfigurationDB_getRecipeTypeData(sRecipeType,rtData,exceptionInfo);
    
  if (dynlen(exceptionInfo)) {
    return;
  }
   
  int len = dynlen(rtData);
  for (int i=1; i<=len; i++){
    if (strtoupper(rtData[i][5])=="TRUE") {
      string deviceType = rtData[i][2];
      dyn_string items;
      if (mappingHasKey(map,deviceType)){
        items = map[deviceType];
      }
      dyn_string splitElement = strsplit(rtData[i][3], ".");
      dynAppend(items, splitElement[dynlen(splitElement)]);
      map[deviceType] = items;
    }
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Save the recipe values.
 * @param recipeObject - [IN] Recipe instance data.
 * @param values - [IN/OUT] Values for the recipe elements.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */ 
void _unRecipeFunctions_saveRecipeValues(dyn_dyn_mixed recipeObject, dyn_anytype &values, dyn_string &exceptionInfo)
{
  dynClear(exceptionInfo);
  
  dyn_anytype formattedValues;
  string instanceName = recipeObject[fwConfigurationDB_RO_META_ORIGNAME];
  dyn_string elements = recipeObject[fwConfigurationDB_RO_DPE_NAME];
  
  // The number of elements in the recipe must be the same as the number of values
  if (dynlen(elements)!=dynlen(values)){
    fwException_raise(exceptionInfo,"ERROR","The number of elements in the recipe " + 
                      "is different than the number of values.","");
    return;      
  }  
  
  _unRecipeFunctions_formatValues(elements, values, formattedValues, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    return;
  }

  // Save the recipe values in the JCOP recipe
  recipeObject[fwConfigurationDB_RO_VALUE] = formattedValues;
  recipeObject[fwConfigurationDB_RO_META_LASTMODIFICATIONTIME] = getCurrentTime();
  recipeObject[fwConfigurationDB_RO_META_LASTMODIFICATIONUSER] = getUserName();
  fwConfigurationDB_saveRecipeToCache(recipeObject, "", instanceName, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    return;
  }
  
  values = formattedValues;
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the element where the range is defined for a PVSS DPE
 * @param sDpName - DPE name in the PVSS format (e.g. dist_1:gcs-CFP_STP_CVSF18-SF18-CPC_AnalogParameter-00001.ProcessOutput.ValRq)
 * @param sElement - Variable to store the element name that contains the range (e.g. .ProcessInput.CurValSt)
 * @param exceptionInfo - Standard exception handling variable.
 */
void _unRecipeFunctions_getRangeElementFromPVSSDPE(string sDpName, string &sElement, dyn_string &exceptionInfo)
{
    string deviceType;
    dyn_string dpeInfo;
    
    sElement = "";
    dynClear(exceptionInfo);
    
    if (!dpExists(sDpName)) {
      fwException_raise(exceptionInfo,"ERROR","The data point element: "+sDpName+
                        " doesn't exist.",unRecipeFunctions_ERROR_InvalidElements);
      return;
    }
    
    deviceType = dpTypeName(sDpName);
    int pos = strpos(sDpName, ".");
    if (pos<0) {
      return;
    }

    if (!dpExists(deviceType+"_RcpInfo.DpeInfo")) {
      // If the information about what is the DPE containing the range definition doesn't exist,
      // returns the DPE name 
      string dp=unGenericDpFunctions_getDpName(sDpName);
      int pos = strpos(sDpName, dp);
      sElement = substr(sDpName, strlen(dp), strlen(sDpName)-strlen(dp));
      return;
    }
        
    sElement = substr(sDpName, pos, strlen(sDpName)-pos);
    dpGet(deviceType+"_RcpInfo.DpeInfo", dpeInfo);

    for (int i=1; i<=dynlen(dpeInfo); i++){
      string currDpeInfo = dpeInfo[i];
      dyn_string splitInfo = strsplit(currDpeInfo, ":");
      if (dynlen(splitInfo)!=2) {
        fwException_raise(exceptionInfo,"ERROR","The recipe info for the device type: "+deviceType+
                          " is wrong.",unRecipeFunctions_ERROR_InvalidElements);
        return;        
      }
      
      splitInfo[1] = strltrim(splitInfo[1]);
      splitInfo[1] = strrtrim(splitInfo[1]);
      if (splitInfo[1] == sElement) {
        sElement = strltrim(splitInfo[2]);
        sElement = strrtrim(sElement);
        return;
      }
    }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the min. and max. values defined in a range for a DPE.
 * @param sDpe - [IN] Data point element for whom the range values are required.
 * @param sRangeElement - [IN] DPE where the range is defined for the 'sDpe' parameter (e.g.: ".ProcessInput.CurValSt")
 * @param sRangeMin - [OUT] Minimum value of the range definition.
 * @param sRangeMax - [OUT] Maximum value of the range definition.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getInstanceRange(string sDpe, string sRangeElement, string &sRangeMin, 
                                         string &sRangeMax, dyn_string &exceptionInfo)
{
  string dpName = unGenericDpFunctions_getDpName(sDpe)+sRangeElement;
  if (!dpExists(dpName)){
    fwException_raise(exceptionInfo,"ERROR","The specified data point doesn't exist: "+dpName,"");
    return;        
  }
  
  string min="",max="";
  dpGet(dpName+":_pv_range.._min", min,
        dpName+":_pv_range.._max", max);
    
  if (min=="" && max=="") {
    sRangeMin="";
    sRangeMax="";
  } else {
    string format = dpGetFormat(dpName);
    sRangeMin = unGenericObject_FormatValue(format, min);
    sRangeMax = unGenericObject_FormatValue(format, max);     
  }
}  

//--------------------------------------------------------------------------------------------------------------------------
/** 
 * Try to lock the devices specified.
 * @param dsDpNames           - [IN] DP names of the devices to lock.
 * @param dsLockedDeviceDp    - [OUT] DP names of the devices that are already locked.
 * @param dsLockedDeviceAlias - [OUT] Alias of the devices that are already locked.
 * @param exceptionInfo       - [OUT] Standard exception handling variable.
 * @return TRUE if the devices were locked, otherwise FALSE.
 */
synchronized bool _unRecipeFunctions_lockDevices(dyn_string dsDpNames, dyn_string &dsLockedDeviceDp, 
    dyn_string &dsLockedDeviceAlias, dyn_string &exceptionInfo) 
{
  int iLen, iErr;
  dyn_string dsIsLockedDp, dsIsLockedValue, dsLockedByDp, dsLockRequestValue;
  dyn_int diLockedByValue;
  
  iLen = dynlen(dsDpNames);
  if (iLen == 0) {
    fwException_raise(exceptionInfo, "ERROR", "The list of devices to lock is empty.", "");
    return false;
  }
  
  for (int i=1; i<=iLen; i++) {
    dynAppend(dsIsLockedDp, dsDpNames[i] + ":_lock._original._locked");
    dynAppend(dsLockedByDp, dsDpNames[i] + ":_lock._original._man_nr");
    dynAppend(dsLockRequestValue, "true");
  }
  
  // Get the 'locked' value of the devices
  iErr = dpGet(dsIsLockedDp, dsIsLockedValue);
  if (_unRecipeFunctions_handleDpGetSetError(iErr, exceptionInfo)) {
	return false;
  }
  
  // Check if all the devices are unlocked
  for (int i=1; i<=iLen; i++) {
	if (dsIsLockedValue[i] == true) {
      dynAppend(dsLockedDeviceDp, dsDpNames[i]);
	  dynAppend(dsLockedDeviceAlias, dpGetAlias(dsDpNames[i]));
	}
  }
  
  // If some devices are locked return
  if (dynlen(dsLockedDeviceDp)) {
	return false;
  }
  
  // Try to lock all the required devices
  iErr = dpSetWait(dsIsLockedDp, dsLockRequestValue);
  if (_unRecipeFunctions_handleDpGetSetError(iErr, exceptionInfo)) {
	return false;
  }
  
  // Get the 'locked' value of the devices
  iErr = dpGet(dsIsLockedDp, dsIsLockedValue,
			   dsLockedByDp, diLockedByValue);
  if (_unRecipeFunctions_handleDpGetSetError(iErr, exceptionInfo)) {
	return false;
  }
  
  // Check if all the devices are locked by the current manager
  for (int i=1; i<=iLen; i++) {
	if (dsIsLockedValue[i] == false || diLockedByValue[i] != myManNum()) {
	  dynAppend(dsLockedDeviceDp, dsDpNames[i]);
      dynAppend(dsLockedDeviceAlias, dpGetAlias(dsDpNames[i]));
	} else {
      _unRecipeFunctions_writeInRecipeLog("Device locked: " + dpGetAlias(dsDpNames[i]), true);
    }
  }
  
  // If the lock failed in any device, unlock all the devices
  if (dynlen(dsLockedDeviceDp)) {
    string sAliasLockError;
    fwGeneral_dynStringToString(dsLockedDeviceAlias, sAliasLockError, ", ");
    _unRecipeFunctions_writeInRecipeLog("The following devices could not be locked: " + sAliasLockError, true);
    _unRecipeFunctions_writeInRecipeLog("Unlocking devices ..", true);
	_unRecipeFunctions_unlockDevices(dsDpNames);
	return false;
  }
    
  return true;
}
//--------------------------------------------------------------------------------------------------------------------------
/** 
 * Try to unlock the devices specified.
 * @param dsDpNames           - [IN] DP names of the devices to unlock.
 * @param exceptionInfo       - [OUT] Standard exception handling variable.
 * @return TRUE if the devices were unlocked, otherwise FALSE.
 */
synchronized bool _unRecipeFunctions_unlockDevices(dyn_string dsDpNames, dyn_string &exceptionInfo) {
  int iErr;
  int iManNum = myManNum();
  bool bSelect = false;
  dyn_string dsLockedDpName, dsLockedByDpName, dsLockedValue, dsAlias;
  dyn_int diLockedByValue;

  dynClear(exceptionInfo);
  
  int iLen = dynlen(dsDpNames);
  if (iLen == 0) {
    fwException_raise(exceptionInfo, "ERROR", "The list of devices to lock is empty.", "");
    return false;
  }
  
  for (int i=1; i<=iLen; i++) {
    dynAppend(dsAlias, dpGetAlias(dsDpNames[i]));
	dynAppend(dsLockedDpName,   dsDpNames[i] + ":_lock._original._locked");
	dynAppend(dsLockedByDpName, dsDpNames[i] + ":_lock._original._man_nr");
  }
  
  // Get the 'locked' state of the devices and the manager number that locked them
  iErr = dpGet(dsLockedDpName, dsLockedValue,
			   dsLockedByDpName, diLockedByValue);			   
  if (_unRecipeFunctions_handleDpGetSetError(iErr, exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Failure while getting the lock state of the devices: '" + dsAlias + "'", true);
    _unRecipeFunctions_writeInRecipeLog("The devices cannot be unlocked", true);    
    return false;
  }
  
  for (int i=1; i<=iLen; i++) {
    // Check if the device is locked by the current manager
    if (dsLockedValue[i] == true) {
      if (diLockedByValue[i] == iManNum) {
        // Release the device
	    iErr = dpSetWait(dsLockedDpName[i], 0);
        if (_unRecipeFunctions_handleDpGetSetError(iErr, exceptionInfo)) {
          _unRecipeFunctions_writeInRecipeLog("Failure while unlocking the device: " + dsAlias[i], true);
        } else {
          _unRecipeFunctions_writeInRecipeLog("Device unlocked: " + dsAlias[i] , true);
        }
      } else {
        _unRecipeFunctions_writeInRecipeLog("The device '" + dsAlias[i]
          + "' is locked by a different manager (" + diLockedValue[i] + ") and it cannot be unlocked", true);
      }        
    }
  }
  
  return (dynlen(exceptionInfo) == 0);
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Enable/disable some graphical components of the recipe instance panel (required for the recipe activation).
 * @param value - [IN] The value for the enabled property of the graphical components.
 */
void _unRecipeFunctions_setRecipeInstancePanelComponentsEnabled(bool value) {
  if (shapeExists("RecipeClasses")) {
	setValue("RecipeClasses", "enabled", value);
  }
	
  if (shapeExists("RecipeInstances")) {
	  setValue("RecipeInstances", "enabled", value);
  }
	
  if (shapeExists("RecipeElements")) {
	  setValue("RecipeElements", "enabled", value);
  }
	
  if (shapeExists("PcoList")) {
	  setValue("PcoList", "enabled", value);
  }
  
  if (shapeExists("ApplicationList")) {
    setValue("ApplicationList", "enabled", value);
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Handle dpGet/dpSet errors.
 * @param err           - [IN] Value returned by dpGet/dpSet
 * @param exceptionInfo - [OUT] Standard exception handling variable.  
 */
bool _unRecipeFunctions_handleDpGetSetError (int err, dyn_string &exceptionInfo) {
  if((err != 0) || (dynlen(getLastError()) > 0)) {
    fwException_raise(exceptionInfo, "ERROR",
                      "Failed to get/set value from/to dpe (dpGet/dpSetWait).",
                      getLastError());
    return true;
  } 

  return false;
}
//------------------------------------------------------------------------------------------------------------------------

/**
 * Get the PLC address of a DPE.
 * @param dpe 			- [IN] The Datapoint element which address will be calculated.
 * @param deviceType 	- [IN] The device type name.
 * @param address 		- [OUT] The PLC address.
 * @param exceptionInfo - [OUT] tandard exception handling variable.
 */
void _unRecipeFunctions_getDpeAddress(string dpe, string deviceType, 
                                     string &address, dyn_string &exceptionInfo)
{
  string addressType, temp;
  dpGet(dpe + ":_address.._drv_ident", addressType);
  string sFunction = "DRV_" + addressType + "_convertToUnicosAddress";
  if(!isFunctionDefined(sFunction)) {
    fwException_raise(exceptionInfo, "ERROR", "The function " +sFunction+ " isn't defined.", "");  
	  DebugTN(exceptionInfo);
    return;
  }	

  evalScript(address,	"string main(string sDeviceDpeName, string sDeviceType)"+
  										"{	string temp;"+
  										"	temp = "+sFunction+"(sDeviceDpeName, sDeviceType);"+
  										"	return temp;"+
  										"}", makeDynString(), dpe, deviceType);     
  
  // If it's an S7 address, gets the offset
  if (addressType=="S7"){
    string fullAddress = address;
    string offset;
    
    int pos=strpos(fullAddress, ".");
    if(pos<0) {
      fwException_raise(exceptionInfo, "ERROR", "The S7 address format is invalid: " + fullAddress, "");
      return;
    }
  
    string temp = substr(fullAddress, pos+1, strlen(fullAddress)-pos);
    int len = strlen(temp);
    int j=0;
    for (int i=0; i<=len; i++){
      if ((temp[i]>='0' && temp[i]<='9') || temp[i]=='\0'){
        offset = offset + temp[i];
        j++;
      }
    }
    address = offset;  
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Set the bit value in a register variable.
 * @param manRegVal - [IN/OUT] Register where the bit value will be modified.
 * @param position  - [IN] Bit position in the register variable.
 * @param value		- [IN] New value for the bit in the register.
 */
void _unRecipeFunctions_setBit(anytype &manRegVal, int position, bool value) {
  bit64 bitVariable;
  bitVariable = manRegVal;
  setBit(bitVariable, position, value);
  manRegVal = bitVariable;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the bit value in a register variable.
 * @param manRegVal - [IN] Register where the bit value will read.
 * @param position  - [IN] Bit position in the register variable.
 * @param value		- [IN] Value of the bit in the register.
 */
void _unRecipeFunctions_getBit(anytype manRegVal, int position, int &value) {
  bit64 bitVariable;
  bitVariable = manRegVal;
  value = getBit(bitVariable, position);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Gets the DB number of a S7 address for a DPE.
 * @param sDpe - [IN] DPE which DB number is required.
 * @param sDeviceType - [IN] Device type name of the dpe.
 * @param sDb - [OUT] DB number of the DPE S7 address.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getS7AddressDBNumber(string sDpe, string sDeviceType, string &sDb, dyn_string &exceptionInfo)
{
  string fullAddress = DRV_S7_convertToUnicosAddress(sDpe, sDeviceType);

  int pos=strpos(fullAddress, ".");
  if(pos<0) {
    fwException_raise(exceptionInfo, "ERROR", "The S7 address format is invalid: " + fullAddress, "");
    return;
  }
  
  sDb = substr(fullAddress, 0, pos);
  strreplace(sDb, "DB", "");
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Gets the offset of an S7 address where a DPE is located.
 * @param sDpe - [IN] DPE which offset is required.
 * @param sDeviceType - [IN] Device type name of the dpe.
 * @param sOffset - [OUT] Offset of the DPE S7 address.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getS7AddressOffset(string sDpe, string sDeviceType, string &sOffset, dyn_string &exceptionInfo)
{
  string fullAddress = DRV_S7_convertToUnicosAddress(sDpe, sDeviceType);

  int pos=strpos(fullAddress, ".");
  if(pos<0) {
    fwException_raise(exceptionInfo, "ERROR", "The S7 address format is invalid: " + fullAddress, "");
    return;
  }
  
  string temp = substr(fullAddress, pos+1, strlen(fullAddress)-pos);
  int len = strlen(temp);
  int j=0;
  for (int i=0; i<=len; i++){
    if ((temp[i]>='0' && temp[i]<='9') || temp[i]=='\0'){
      sOffset = sOffset + temp[i];
      j++;
    }
  }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Gets the recipe elements according to the distributed system.
 * @param elements - [IN] Recipe data point elements.
 * @param values - [IN] Recipe values.
 * @param systemDpes - [OUT] Map containing the recipe elements and values for each distributed system.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */  
void _unRecipeFunctions_getElementsBySystem(dyn_string elements, dyn_anytype values, 
                                            mapping &systemDpes, dyn_string &exceptionInfo)
{
  string pvssDpe;
  dyn_dyn_string data;
  mappingClear(systemDpes);
  
  int len = dynlen(elements);
  if (len!=dynlen(values)) {
    fwException_raise(exceptionInfo, "ERROR", "The number of DPEs and values of the recipe must be equal.", "");
    return;
  }  
  
  // Inserts the DPEs in the mapping according to the system name
  for (int i=1; i<=len; i++) {
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(elements[i], pvssDpe);
    string systemName = unGenericDpFunctions_getSystemName(pvssDpe);
    
    dynClear(data);
    if (mappingHasKey(systemDpes, systemName)) {
      data = systemDpes[systemName];
    }

    dynAppend(data[1], pvssDpe);
    dynAppend(data[2], values[i]);
    systemDpes[systemName] = data;
  }  
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the DPEs that contains the online values (ProcessInput) for the specified DPEs (ProcessOutput)
 * @param dpNames       - [IN] List of DPEs which online value DPEs are required.
 * @param outputDpes    - [OUT] List of online value DPEs.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getOnlineValueDpes (dyn_string dpNames, dyn_string &outputDpes, dyn_string &exceptionInfo) {
	int len = dynlen(dpNames);
	for (int i=1; i<=len; i++) {    
        if (!dpExists(dpNames[i])) {
          fwException_raise(exceptionInfo, "ERROR", "The datapoint " + dpNames[i] + " doesn't exist.", "");
          continue;
        }
		string dpType = dpTypeName(dpNames[i]);
		
		if (dpType=="CPC_DigitalParameter") {
			strreplace(dpNames[i], "ProcessOutput.ManReg01", "ProcessInput.PosSt");
		} else if (dpType=="CPC_AnalogAlarm") {
			strreplace(dpNames[i], "ProcessOutput", "ProcessInput");
			dpNames[i] += "St";
		} else if (dpType=="CPC_Controller") {
			if (strpos(dpNames[i], "ProcessOutput.MSPH") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MSPH", "ProcessInput.ActSPH");
			} else if (strpos(dpNames[i], "ProcessOutput.MSPL") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MSPL", "ProcessInput.ActSPL");
			} else if (strpos(dpNames[i], "ProcessOutput.MSP") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MSP", "ProcessInput.ActSP");
			} else if (strpos(dpNames[i], "ProcessOutput.MOutH") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MOutH", "ProcessInput.ActOutH");
			} else if (strpos(dpNames[i], "ProcessOutput.MOutL") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MOutL", "ProcessInput.ActOutL");
			} else if (strpos(dpNames[i], "ProcessOutput.MKc") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MKc", "ProcessInput.ActKc");
			} else if (strpos(dpNames[i], "ProcessOutput.MTi") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MTi", "ProcessInput.ActTi");
			} else if (strpos(dpNames[i], "ProcessOutput.MTd") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MTd", "ProcessInput.ActTd");
			} else if (strpos(dpNames[i], "ProcessOutput.MTds") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MTds", "ProcessInput.ActTds");
			} 
		} else if (dpType=="CPC_WordParameter" || dpType=="CPC_AnalogParameter") {
			strreplace(dpNames[i], "ProcessOutput.MPosR", "ProcessInput.PosSt");
		} 
	}
	
	outputDpes = dpNames;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the DPEs that contains the default values for the specified DPEs 
 * @param dpNames       - [IN] List of DPEs which online value DPEs are required.
 * @param outputDpes    - [OUT] List of online value DPEs.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getDefaultValueDpes (dyn_string dpNames, dyn_string &outputDpes, dyn_string &exceptionInfo) {
	int len = dynlen(dpNames);
	for (int i=1; i<=len; i++) {    
        if (!dpExists(dpNames[i])) {
          fwException_raise(exceptionInfo, "ERROR", "The datapoint " + dpNames[i] + " doesn't exist.", "");
          continue;
        }
		string dpType = dpTypeName(dpNames[i]);
		
		if (dpType=="CPC_DigitalParameter") {
			strreplace(dpNames[i], "ProcessOutput.ManReg01", "ProcessOutput.DfltVal");
		} else if (dpType=="CPC_AnalogAlarm") {
      // Nothing to do
		} else if (dpType=="CPC_Controller") {
			if (strpos(dpNames[i], "ProcessOutput.MSPH") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MSPH", "ProcessOutput.DefSPH");
			} else if (strpos(dpNames[i], "ProcessOutput.MSPL") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MSPL", "ProcessOutput.DefSPL");
			} else if (strpos(dpNames[i], "ProcessOutput.MSP") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MSP", "ProcessOutput.DefSP");
			} else if (strpos(dpNames[i], "ProcessOutput.MOutH") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MOutH", "ProcessOutput.DefOutH");
			} else if (strpos(dpNames[i], "ProcessOutput.MOutL") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MOutL", "ProcessOutput.DefOutL");
			} else if (strpos(dpNames[i], "ProcessOutput.MKc") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MKc", "ProcessOutput.DefKc");
			} else if (strpos(dpNames[i], "ProcessOutput.MTi") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MTi", "ProcessOutput.DefTi");
			} else if (strpos(dpNames[i], "ProcessOutput.MTd") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MTd", "ProcessOutput.DefTd");
			} else if (strpos(dpNames[i], "ProcessOutput.MTds") >= 0) {
				strreplace(dpNames[i], "ProcessOutput.MTds", "ProcessOutput.DefTds");
			} 
		} else if (dpType=="CPC_WordParameter" || dpType=="CPC_AnalogParameter") {
			strreplace(dpNames[i], "ProcessOutput.MPosR", "ProcessOutput.DfltVal");
		} 
	}
	
	outputDpes = dpNames;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if the recipe values are the same than the online values.
 * @param dpes          - [IN] Recipe elements data points.
 * @param values        - [IN] Recipe values.
 * @param errDpe        - [OUT] If the recipe values are different than the online values, this variable
 *      will contain the first dpe where the values are different.
 * @param exceptionInfo - [OUT] Standard error handling variable.
 * @return TRUE if the recipe values are the same than the online values, otherwise FALSE.
 */
 bool _unRecipeFunctions_checkRecipeOnlineValues(dyn_string dpes, dyn_string values, string &errDpe, dyn_string &exceptionInfo)
{ 
  bool bOk;
  dyn_string outputDpes;
  dyn_anytype onlineValues;
  string sFormat, sOnlineValue;
  float fOnlineValue, fRcpValue;;
  
  if (dynlen(dpes)==0) {
	return true;
  }
  
  // Get the DPEs for the online values
  _unRecipeFunctions_getOnlineValueDpes (dpes, outputDpes, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    return FALSE;
  }
  dpGet(outputDpes, onlineValues);

  int len = dynlen(onlineValues);
  for (int i=1; i<=len; i++) {	  
	  float fval, frcpval, fDiff;
      int ival, ircpval;
      bool bval, brcpval;        
	  
      if (!dpExists(outputDpes[i])) {
        fwException_raise(exceptionInfo, "ERROR", "The datapoint " + outputDpes[i] + " doesn't exist.", "");
        continue;
      }
      
	    string dpType = dpTypeName(outputDpes[i]);
	    if (dpType=="CPC_DigitalParameter") {
		    bval = onlineValues[i];
		    brcpval = values[i];
		    if (bval!=brcpval) {
			    errDpe = dpes[i];
			    return FALSE;
		    }
		    continue;	
	    }
	
      int elementType = dpElementType(outputDpes[i]); 
      switch(elementType){
        case DPEL_FLOAT:
          sFormat = dpGetFormat(outputDpes[i]);
          // Format the online value (e.g. values like '10.579999923706' will be converted to '10.580')
          sOnlineValue = unGenericObject_FormatValue(sFormat, onlineValues[i]);			
          // Get the online value as float
          unGenericObject_StringToFloat(sOnlineValue, fOnlineValue, bOk);
          // Get the recipe value as float
          unGenericObject_StringToFloat(values[i], fRcpValue, bOk);
          // Compare the float values			
          if (fRcpValue != fOnlineValue) {
            errDpe = dpes[i];
            return FALSE;
          }
          break;
        case DPEL_INT:
          ircpval = (int)values[i];          
          if (onlineValues[i]!=ircpval) {
            errDpe = dpes[i];
            return FALSE;
          }             
          break;        
        case DPEL_BOOL:
          brcpval = (bool)values[i];
          if (onlineValues[i]!=brcpval) {
            errDpe = dpes[i];
            return FALSE;
          }  
          break;        
        default:
          value = (string)values[j];           
      }
  }
  
  return TRUE;
} 

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the status of the recipe instance (Activated, inactive, no plc connection,...).
 * @param recipeObject  - [IN] Recipe instance data from fwConfigDB.
 * @param lastActivated - [IN] String containing the name of the last activated recipe of the class.
 * @param status 	    - [OUT] Recipe status.
 * @param errMsg        - [OUT] Error message for the user.
 * @param exceptionInfo - [OUT] Standard error handling variable.
 */ 
void unRecipeFunctions_getRecipeStatus(dyn_dyn_mixed recipeObject, string lastActivated, int &status, 
                                       dyn_string &errMsg, dyn_string &exceptionInfo)
{
  bool active;
  dyn_string statusMessage;
  string errDpe;
  string rcpName 		 = recipeObject[fwConfigurationDB_RO_META_ORIGNAME];
  dyn_string rcpElements = recipeObject[fwConfigurationDB_RO_DPE_NAME];
  dyn_string rcpValues   = recipeObject[fwConfigurationDB_RO_VALUE];
  dyn_string rcpTypes	 = recipeObject[fwConfigurationDB_RO_ELEMENT_TYPE];
  
  // Check the PLCs status
  unRecipeFunctions_getPlcStatus(rcpElements, active, errMsg, exceptionInfo);
  if(active==false){
    status = RCP_NO_PLC_CONNECTION;
    return;
  }
  
  // Check if all the DPEs used in the recipe exist
  status = unRecipeFunctions_checkRecipeStructure(rcpName, rcpElements, rcpValues, rcpTypes, exceptionInfo);
  if (dynlen(exceptionInfo)){
    return;
  }  
  
  // Check if the last activated recipe of this type is the current one
  lastActivated = unExportTree_removeSystemName(lastActivated);
  rcpName       = unExportTree_removeSystemName(rcpName);
  if (lastActivated != rcpName){
    status = RCP_INACTIVE;
    return;
  }      
  
  //Debug("unRecipeFunctions_getRecipeStatus - Getting online values \n");
  bool bOnlineValuesOk = _unRecipeFunctions_checkRecipeOnlineValues(rcpElements, rcpValues, errDpe, exceptionInfo);
  if(!bOnlineValuesOk || dynlen(exceptionInfo)){
    status = RCP_INACTIVE;  
    return;  
  }  

  status = RCP_ACTIVATED;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if the recipe structure is coherent:
 *   - The number of DPEs and the number of values is the same
 *   - The values match the expected data type.
 * @param sRcpName      - [IN] Recipe instance name.
 * @param dsRcpElements - [IN] DPEs of the recipe.
 * @param dsRcpValues   - [IN] Recipe values.
 * @param dsRcpTypes    - [IN] Recipe value types.
 * @param exceptionInfo - [OUT] Standard error handling variable.
 * @return The result of the recipe structure check (bad types, bad structure ...)
 */
int unRecipeFunctions_checkRecipeStructure(string sRcpName, dyn_string dsRcpElements, dyn_string dsRcpValues, dyn_string dsRcpTypes,
                                            dyn_string &exceptionInfo)
{
  float fval;
  int ival, elementType;
  bool bval;
  dyn_string exception;
  string deviceType = "", lowerValue;
  
  dynClear(exceptionInfo);
  
  // Check if the number of elements is equal to the number of values
  if (dynlen(dsRcpElements)!=dynlen(dsRcpValues)){
    fwException_raise(exceptionInfo, "ERROR", "unRecipeFunctions_checkRecipeStructure: " +
                      "The number of elements is different than the number of values: " + sRcpName, "");
    return RCP_BAD_STRUCTURE;
  }
  
  // Check if the recipe value types match the data point element types
  int numElements = dynlen(dsRcpElements);
  for (int i=1; i<=numElements; i++){    
    if (!dpExists(dsRcpElements[i])) {
        fwException_raise(exceptionInfo, "ERROR", "unRecipeFunctions_checkRecipeStructure: " +
			          "The recipe DPE does not exist: " + dsRcpElements[i], "");
        return RCP_BAD_DPES;
    }
	deviceType = dpTypeName(dsRcpElements[i]);

	if (deviceType=="CPC_DigitalParameter") {
		elementType = DPEL_BOOL;
	} else {
		elementType = dsRcpTypes[i];
    }
    dynClear(exception);
	
    switch(elementType){
      case DPEL_FLOAT:       
        unConfigGenericFunctions_checkFloat(dsRcpValues[i], fval, exception); 
        if(dynlen(exception)){ 
          dynAppend(exceptionInfo, exception);  
          return RCP_BAD_TYPES;
        }
        break;
      case DPEL_INT:
        unConfigGenericFunctions_checkInt(dsRcpValues[i], ival, exception); 
        if(dynlen(exception)) {
          dynAppend(exceptionInfo, exception);   
          return RCP_BAD_TYPES;
        }
        break;
      case DPEL_BOOL:
		lowerValue = strtolower(dsRcpValues[i]);
		if (lowerValue!="true" && lowerValue!="false") {
			fwException_raise(exceptionInfo, "ERROR", "unRecipeFunctions_checkRecipeStructure: " +
			          "Wrong boolean value: " + dsRcpValues[i], "");
            return RCP_BAD_TYPES;
        }
        break;
      default:
        fwException_raise(exceptionInfo, "ERROR", "unRecipeFunctions_checkRecipeStructure: " +
                          "Unknown DPType: " + dpTypeName(dsRcpElements[i]), "");
        return RCP_BAD_TYPES;
    }
  }
  
  return -1;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if the recipe values are valid according to the defined ranges.
 * @param recipeObject    - [IN]  Recipe object.
 * @param values          - [IN]  Recipe values.
 * @param bShowErrorLines - [IN]  TRUE if the values out of range must be highlighted in the table.
 * @param exceptionInfo   - [OUT] Standard exception handling variable.
 * @return TRUE if the recipe values are Ok according to the ranges, otherwise FALSE.
 */
bool unRecipeFunctions_checkRecipeValues (dyn_dyn_mixed recipeObject, dyn_anytype &values, 
                      bool bShowErrorLines, dyn_string &exceptionInfo)
{
  return _unRecipeFunctions_checkRecipeValues(recipeObject[fwConfigurationDB_RO_DPE_NAME], values, bShowErrorLines, exceptionInfo);
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Compares the recipe values and online values of a list of recipes.
 * @param dsRcpDpList [IN] - List of recipe instance datapoints.
 */
void unRecipeFunctions_compareOnlineValues(dyn_string dsRcpDpList) {
  if (!shapeExists("RcpTable")) {
    _unRecipeFunctions_writeInRecipeLog("Error: The shape 'RcpTable' is missing.", TRUE, TRUE);
    return;
  }  
  
  
  int iDpCol;
  dyn_string dsAllRcpDps;
  
  getMultiValue("RcpTable", "nameToColumn", "RcpInstanceDp", iDpCol,
                "RcpTable", "getColumnN", iDpCol, dsAllRcpDps); 
  
  int i, iLen;
  iLen = dynlen(dsRcpDpList);
  for (i=1; i<=iLen; i++) {
    string sRcpDp, sRcpName;
    dyn_dyn_mixed recipeObject;
    dyn_string exceptionInfo, dsDpNames, dsAlias;  
    dyn_anytype daOnlineValues, daRecipeValues;
    dyn_int diRowStyleIndex, diRowStyles, diDiffRowIndex;
    
    int iPos = dynContains(dsAllRcpDps, dsRcpDpList[i]);
    sRcpDp = dsRcpDpList[i];
    _unRecipeFunctions_normalizeDp(sRcpDp);
    sRcpName  = unGenericDpFunctions_getAlias(sRcpDp + ".");
    _unRecipeFunctions_writeInRecipeLog("Checking online values of recipe " + sRcpName, TRUE, TRUE);
    _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "yellow");
    _unRecipeFunctions_getRecipeObjectFromDp(sRcpDp, recipeObject, exceptionInfo);
    if (dynlen(exceptionInfo)) { 
      DebugTN(exceptionInfo); 
      continue; 
    }
    daRecipeValues = recipeObject[fwConfigurationDB_RO_VALUE];
    dsDpNames = recipeObject[fwConfigurationDB_RO_DPE_NAME];
    dsAlias = recipeObject[fwConfigurationDB_RO_DP_NAME];
    unRecipeFunctions_getOnlineValues(dsDpNames, daOnlineValues, exceptionInfo);
    if (dynlen(exceptionInfo)) { 
      DebugTN(exceptionInfo); 
      continue; 
    }    
  
    _unRecipeFunctions_getRowStylesForValueComparison(daRecipeValues, daOnlineValues, diRowStyleIndex, diRowStyles, diDiffRowIndex);
    unRecipeFunctions_writeDifferentOnlineValuesReport(diDiffRowIndex, diDiffRowIndex, dsDpNames, dsAlias, daRecipeValues, daOnlineValues);
    
    if (dynlen(diDiffRowIndex)) {
      _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "red");
    } else {
      _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "green");
    }
  }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Saves the online values of the recipe elements into the recipe.
 * @param dsRcpDps [IN] - Datapoint names of the recipe instances.
 */
public void unRecipeFunctions_adoptOnlineValues(dyn_string dsRcpDps) {
  int i, iLen, iDpCol;
  dyn_string dsAllRcpDps;
  
  getMultiValue("RcpTable", "nameToColumn", "RcpInstanceDp", iDpCol,
                "RcpTable", "getColumnN", iDpCol, dsAllRcpDps); 
 
  iLen = dynlen(dsRcpDps);
  for (i=1; i<=iLen; i++) {
    dyn_dyn_mixed recipeObject;
    dyn_string dsDpNames, dsDpNames, exceptionInfo;
    dyn_anytype daOnlineValues;
    string sAlias = unGenericDpFunctions_getAlias(dsRcpDps[i]);

    int iPos = dynContains(dsAllRcpDps, dsRcpDps[i]);
    
    _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "yellow");
    if (!unRecipeFunctions_lockRecipe(dsRcpDps[i], sAlias)) {
      _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "red");
      continue;
    }
    
    _unRecipeFunctions_getRecipeObjectFromDp(dsRcpDps[i], recipeObject, exceptionInfo);
    if (dynlen(exceptionInfo)) {
      _unRecipeFunctions_writeInRecipeLog("Exception loading the recipe '" + sAlias + "' : " + exceptionInfo[2], TRUE, TRUE);
      _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "red");
      continue;
    }
        
    dsDpNames = recipeObject[fwConfigurationDB_RO_DPE_NAME];
    unRecipeFunctions_getOnlineValues(dsDpNames, daOnlineValues, exceptionInfo);
    if (dynlen(exceptionInfo)){
      _unRecipeFunctions_writeInRecipeLog("Exception loading the recipe '" + sAlias + "' " + exceptionInfo[2], TRUE, TRUE);
      _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "red");
      continue;
    }
    _unRecipeFunctions_writeInRecipeLog("Saving online values in recipe: " + sAlias, TRUE, TRUE);
    _unRecipeFunctions_saveRecipeValues(recipeObject, daOnlineValues, exceptionInfo);
    if (dynlen(exceptionInfo)){
      _unRecipeFunctions_writeInRecipeLog("Exception saving the values of recipe '" + sAlias + "' : " + exceptionInfo[2], TRUE, TRUE);
      _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "red");
      continue;
    }
    _unRecipeFunctions_setTableCellBGColor("RcpTable", (iPos-1), "RcpActivationStatus", "green");
    dpSet(dsRcpDps[i] + ".ProcessInput.State", "");
  }
}
    
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Locks the recipe instance specified.
 * @param sRcpDp [IN] - Datapoint name of the recipe instace.
 * @param sAlias [IN] - Recipe alias.
 * @return TRUE if and only if the recipe instance was locked.
 */  
private bool unRecipeFunctions_lockRecipe(string sRcpDp, string sAlias) {
  bool bLocked, bSelected, bSystemOk;
  string sSelectedManager, sSystemName, sRecipeState;
  dyn_string exceptionInfo;
  
  unSelectDeselectHMI_isLocked(sRcpDp, bLocked, sSelectedManager);
  string localManager = unSelectDeselectHMI_getSelectedState(bLocked, sSelectedManager);
  bSelected = (localManager == "S");
  sSystemName = unGenericDpFunctions_getSystemName(sRcpDp);
  bSystemOk = (sSystemName == getSystemName());
  dpGet(sRcpDp + ".ProcessInput.State", sRecipeState);
  
  if (bLocked && !bSelected) {
    _unRecipeFunctions_writeInRecipeLog("Error: The recipe '" + sAlias + "' is selected by a different manager (" + sSelectedManager + ").", TRUE, TRUE);
    return false;
  }
  
  if (sRecipeState != "") {
    _unRecipeFunctions_writeInRecipeLog("Error: The recipe '" + sAlias + "' is in state (" + sRecipeState + ") and cannot be locked.", TRUE, TRUE);
    return false;
  }

  dpSet(sRcpDp + ".ProcessInput.State", UN_RCP_INSTANCE_STATE_LOCKED);

  return true;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Internal function to check if the recipe values are valid according to the defined ranges.
 * @param dsDps           - [IN]  Datapoint elements of the recipe values.
 * @param values          - [IN]  Recipe values.
 * @param bShowErrorLines - [IN]  TRUE if the values out of range must be highlighted in the table.
 * @param exceptionInfo   - [OUT] Standard exception handling variable.
 * @return TRUE if the recipe values are Ok according to the ranges, otherwise FALSE.
 */
bool _unRecipeFunctions_checkRecipeValues (dyn_string dsDps, dyn_anytype values, bool bShowErrorLines, dyn_string &exceptionInfo) {  
  int iRetVal;
  dyn_int diErrorLines;

  dynClear(exceptionInfo);  
  int iLen = dynlen(dsDps);
  
  // Check if the number DPs matches the number of values.
  if (iLen != dynlen(values)) {
    fwException_raise(exceptionInfo, "ERROR", "The number of elements in the recipe " + 
                      "is different than the number of values.", "");
    return FALSE;
  }
  
  for (int i=1; i<=iLen; i++){
    dyn_string exception;
	// Check if the recipe value is valid
    iRetVal = unRecipeFunctions_checkRecipeValue(dsDps[i], values[i], exception);

	if (dynlen(exception)) {
      dynAppend(exceptionInfo, exception);
    }	  
	  
	if (iRetVal == -1) {
      dynAppend(diErrorLines, i);
    } 
  }
  
  if (dynlen(diErrorLines)) {
    fwException_raise(exceptionInfo, "ERROR", "The value of some recipe elements is out of range.", "");
  }
  
  if (bShowErrorLines) {
    _unRecipeFunctions_showValuesOutOfRange(diErrorLines);
  }
  
  return (dynlen(diErrorLines) == 0);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if the recipe value is valid according to the defined range.
 * @param sDpName - [IN] Data point name of the recipe element.
 * @param value   - [IN] Value of the recipe element.
 * @param exceptionInfo - [OUT] Standard exception handling routine.
 * @return 0 if the value is in the defined range or if no range is defined, otherwise -1.
 */
int unRecipeFunctions_checkRecipeValue (string sDpName, anytype value, dyn_string &exceptionInfo)
{
	string sVal, range, min, max;
	
	range = unRecipeFunctions_getDeviceRange(sDpName, exceptionInfo);	
	if (dynlen(exceptionInfo)) {
		return -1;
	}
	
	// No range defined => Value is valid
	if (range=="") {
		return 0;	
    }
		
	// Check if the range is FALSE, TRUE
	if (range=="[FALSE, TRUE]") {
		if (value=="0" || value=="1" || strtoupper(value)=="TRUE" || strtoupper(value)=="FALSE") {
			return 0;
        }
		return -1;
    }	
		
    _unRecipeFunctions_getRangeValues(range, min, max, exceptionInfo);
	
    if (dynlen(exceptionInfo)) {
      return -1;
    } 
	
    float fmin, fmax, fval;
    fmin = (float)min;
    fmax = (float)max;
	
	// Removes the white space in the exponential value (1.5 e1 => 1.5e1) before converting to float
	sVal = value;
	strreplace(sVal, " ", "");	
    fval = (float)sVal;

    if (fmin>fval || fval>fmax){
      return -1;
    } 
	
	return 0;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the min. and max. values defined in a range string
 * @param range - [IN] String containing a range definition (i.e.: [0.0, 100.0])
 * @param min - [OUT] Minimum value defined in the range.
 * @param max - [OUT] Maximum value defined in the range.
 * @param exceptionInfo - [OUT] Standard exception handling routine.
 */
void _unRecipeFunctions_getRangeValues(string range, string &min, string &max, dyn_string &exceptionInfo)
{  
  string rangeElement, deviceType, rangeMin, rangeMax;

  min = "";
  max = "";
  
  string tmpRange = range;
  strreplace(tmpRange, "[", "");
  strreplace(tmpRange, "]", "");
  strreplace(tmpRange, " ", "");
  

  dyn_string rangeSplit = strsplit(tmpRange, ",");
  if (dynlen(rangeSplit)!=2)  {
    fwException_raise(exceptionInfo, "ERROR", "The specified range format is incorrect: " + range, "");
    return;
  }

  min = rangeSplit[1];
  max = rangeSplit[2];
  
  strreplace(min, " ", "");
  strreplace(max, " ", "");
  
  min = strrtrim(min);
  min = strltrim(min);
  max = strrtrim(max);
  max = strltrim(max);
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Gets the PLC status of a recipe.
 * @param recipeElements - [IN] List of DPEs.
 * @param bActive        - [OUT] True if the PLCs are active, otherwise false.
 * @param statusMessage  - [OUT] PLC status message.
 * @param exceptionInfo  - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_getPlcStatus(dyn_string recipeElements, bool &bActive, 
                                    dyn_string &statusMessage, dyn_string &exceptionInfo)
{
  int iRes;
  bActive = true;
	
  // Get the list of PLCs used in the recipe
  dyn_string plcList;
  for (int i=1; i<=dynlen(recipeElements); i++){
    if (!dpExists(recipeElements[i])) {
        fwException_raise(exceptionInfo, "ERROR", "The recipe element DPE does not exist: " + recipeElements[i], "");
        return;
    }
    
    string sSystemName = unGenericDpFunctions_getSystemName(recipeElements[i]);
    string sPlc = sSystemName + unGenericObject_GetPLCNameFromDpName(recipeElements[i]);
    if (!dynContains(plcList, sPlc)){
        dynAppend(plcList, sPlc);
    }
  }
	
  //check PLC status
  for (int i=1; i<=dynlen(plcList); i++){
    //sPlcName - corresponding dp name in "_Mod_Plc" internal DPT
    int iSystemIntegrityAlarmValue = 0;
    bool bSystemIntegrityAlarmEnabled = false;
    string sSystemName = unGenericDpFunctions_getSystemName(plcList[i]);
    string sPlcName = substr(plcList[i], strlen(sSystemName));	
    iRes = dpGet(sSystemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + sPlcName + ".alarm", iSystemIntegrityAlarmValue);
    if (iRes < 0)	{
      dynAppend(exceptionInfo, "unRecipeFunctions_getPlcStatus: failed to get current state of plc - " + sPlcName +"\n");
      return;
    }
	
    iRes = dpGet(sSystemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + sPlcName + ".enabled", bSystemIntegrityAlarmEnabled);
    if (iRes < 0)	{
     	dynAppend(exceptionInfo, "unRecipeFunctions_getPlcStatus: failed to get current state of plc - " + sPlcName+"\n");
        return;
    }	
    
    if(!bSystemIntegrityAlarmEnabled || (iSystemIntegrityAlarmValue > c_unSystemIntegrity_no_alarm_value)) {
        bActive = false;
        dynAppend(statusMessage, sPlcName);
    }
  }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Callback function executed when the PLC status has changed.
 * The properties table of the recipe instance will be reloaded.
 * @param sAlarmDpe - [IN] DP Name of the PLC alarm state.
 * @param iSystemIntegrityAlarmValue - [IN] Value of the PLC alarm state.
 * @param sEnabledDpe - [IN] DP Name of the system integrity alarm enabled.
 * @param bSystemIntegrityAlarmEnabled - [IN] Value of the system integrity alarm enabled.
 */
void _unRecipeFunctions_plcStatusCB (string sAlarmDpe, int iSystemIntegrityAlarmValue,
									 string sEnabledDpe, bool bSystemIntegrityAlarmEnabled)
{	
	if (!shapeExists("LoadedRecipeDp")) {
		return;
	}

	_unRecipeFunctions_loadPropertiesTableInfo();
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Disconnect from the PLC status of a recipe instance.
 * @param sRcpInstanceDp - [IN] DP Name of the recipe instance.
 * @param exceptionInfo  - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_disconnectPlcStatus(string sRcpInstanceDp, dyn_string &exceptionInfo)
{
  dyn_string dsRecipeElements, dsPlcList;
  dyn_dyn_mixed recipeObject;
  string sPlcName, sSystemName;
  int rc;

  _unRecipeFunctions_getRecipeObjectFromDp(sRcpInstanceDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    DebugTN(exceptionInfo); 
    return; 
  }  
  dsRecipeElements = recipeObject[fwConfigurationDB_RO_DPE_NAME];
  
  // Get the list of PLCs used in the recipe
  for (int i=1; i<=dynlen(dsRecipeElements); i++){
     sSystemName = unGenericDpFunctions_getSystemName(dsRecipeElements[i]);
     sPlcName    = sSystemName + unGenericObject_GetPLCNameFromDpName(dsRecipeElements[i]);
     if (!dynContains(dsPlcList, sPlcName)){
       dynAppend(dsPlcList, sPlcName);
     }
  }
  
  // connect to PLC status
  for (int i=1; i<=dynlen(dsPlcList); i++){
    sSystemName = unGenericDpFunctions_getSystemName(dsPlcList[i]);
    sPlcName    = substr(dsPlcList[i], strlen(sSystemName));	
    rc = dpDisconnect("_unRecipeFunctions_plcStatusCB", 
			  sSystemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + sPlcName + ".alarm", 
			  sSystemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + sPlcName + ".enabled" );
  }	
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Establish a DP connection to the PLC status of all the PLCs containing devices of the recipe instance.
 * @param sRcpInstanceDp - [IN] DP Name of the recipe instance.
 * @param exceptionInfo  - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_connectPlcStatus(string sRcpInstanceDp, dyn_string &exceptionInfo)
{
  dyn_string dsRecipeElements, dsPlcList;
  dyn_dyn_mixed recipeObject;
  string sPlcName, sSystemName;

  _unRecipeFunctions_getRecipeObjectFromDp(sRcpInstanceDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    DebugTN(exceptionInfo); 
    return; 
  }
  
  dsRecipeElements = recipeObject[fwConfigurationDB_RO_DPE_NAME];
  
  // Get the list of PLCs used in the recipe
  for (int i=1; i<=dynlen(dsRecipeElements); i++){
    if (!dpExists(dsRecipeElements[i])) {
        fwException_raise(exceptionInfo, "ERROR", "The recipe element DPE does not exist: " + dsRecipeElements[i], "");
        return;
    }
    sSystemName = unGenericDpFunctions_getSystemName(dsRecipeElements[i]);
    sPlcName = sSystemName + unGenericObject_GetPLCNameFromDpName(dsRecipeElements[i]);
    
    if (!dynContains(dsPlcList, sPlcName)){
       dynAppend(dsPlcList, sPlcName);
    }
  }
  
  // connect to PLC status
  for (int i=1; i<=dynlen(dsPlcList); i++){
    sPlcName = dsPlcList[i];	
    sSystemName = unGenericDpFunctions_getSystemName(dsRecipeElements[i]);
    sPlcName = substr(sPlcName, strlen(sSystemName));
	dpConnect("_unRecipeFunctions_plcStatusCB", 
			  sSystemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + sPlcName + ".alarm", 
			  sSystemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + sPlcName + ".enabled");
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the recipe classes linked to the PCO sPcoAlias.
 * @param sPcoAlias - [IN] The PCO alias which recipe classes muat be loaded.
 * @param exceptionInfo - [OUT] Standard exception handling routine.
 * @param bLoadFirstClass - [IN] TRUE if the data of the first recipe class must be loaded in the panel. 
 * @param sApplication - [IN] Application name for the recipe classes to load.
 * @param bOnlyConnectedSystems - [IN]  TRUE if only the devices of the connected systems must be obtained.
 * @return The number of recipe classes added.
 */
int unRecipeFunctions_loadRecipeClasses(string sPcoAlias, dyn_string &exceptionInfo, bool bLoadFirstClass=TRUE, string sApplication="*", 
         bool bOnlyConnectedSystems=FALSE)
{
  int iLines = 0;
  string sType, sDeleteName, sAppendName;
  dyn_string recipeClassNames, recipeClassDps;
  
  if (!shapeExists("RecipeClasses")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    return 0;
  }

  bool bLocalOnly = FALSE;
  bool bAddSystemName = FALSE;
  unRecipeFunctions_getRecipeClasses(recipeClassNames, recipeClassDps, sPcoAlias, bLocalOnly, bAddSystemName, sApplication, bOnlyConnectedSystems);  
  
  // Get the type of the RecipeClasses element
  getValue("RecipeClasses", "type", sType);
  if (sType=="Table") {
    iLines = _unRecipeFunctions_loadRecipeClassesInTable(recipeClassNames, recipeClassDps, bLoadFirstClass);
  } else if (sType=="ComboBox") {
    iLines = _unRecipeFunctions_loadRecipeClassesInComboBox(recipeClassNames);
  } else {
    _unRecipeFunctions_writeInRecipeLog("Error: The type of the 'RecipeClasses' shape is unknown.");
  }
  
  return iLines;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the recipe class list in a table.
 * @param rcpClassNames   - [IN] List containing the recipe class names.
 * @param rcpClassDps     - [IN] List containing the recipe class datapoints.
 * @param bLoadFirstClass - [IN] TRUE if the data of the first recipe class must be loaded in the panel.
 * @return The number of recipe classes added.
 */
int _unRecipeFunctions_loadRecipeClassesInTable(dyn_string rcpClassNames, dyn_string rcpClassDps, bool bLoadFirstClass=TRUE) {
  int i, iLines;
  dyn_int lines;
  string sRcpClass, sRcpDp, sPvssDpe;
  dyn_string exceptionInfo;

  iLines = dynlen(rcpClassNames);
  setValue("RecipeClasses", "deleteAllLines");
  if (iLines > 0) {
    setValue("RecipeClasses", "appendLines", iLines, 
			 "RecipeClass", rcpClassNames,
			 "RecipeClassDp", rcpClassDps);
	// Select the first recipe class in the table
	if (bLoadFirstClass==true) {
		unRecipeFunctions_selectRecipeClass(rcpClassDps[1], exceptionInfo, TRUE);
	}
  } else {
	setValue("RecipeClasses", "appendLine", "RecipeClass", "-None-");
	unRecipeFunctions_clearRecipeClassData();
  }
  
  return iLines;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the recipe class list in a combo box.
 * @param rcpClassNames - [IN] List containing the recipe class names.
 * @return The number of recipe classes added.
 */
int _unRecipeFunctions_loadRecipeClassesInComboBox(dyn_string rcpClassNames) {
  int i, iLines;
  
  iLines = dynlen(rcpClassNames);
  setValue("RecipeClasses", "deleteAllItems");
  if (iLines > 0) {
    for (int i=1; i<=iLines; i++) {
	  setValue("RecipeClasses", "appendItem", rcpClassNames[i]);
	}
  } else {
	setValue("RecipeClasses", "appendItem", "-None-");
  }
  
  return iLines;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Clear all the data from the recipe class panel.
 */
void unRecipeFunctions_clearRecipeClassData() {
  if (shapeExists("RecipeClass")) {
    RecipeClass.text = "";
  }
	
  if (shapeExists("RecipeType")) {
    RecipeType.text = "";
  }
	
  if (shapeExists("RecipeClassComment")) { 
    RecipeClassComment.text = "";
  }
	
  if (shapeExists("LastModifiedTime")) {
    LastModifiedTime.text = "";
  }
	
  if (shapeExists("LastModifiedUser")) {
    LastModifiedUser.text = "";
  }
	
  if (shapeExists("LastActivatedRecipe")) {
    LastActivatedRecipe.text = "";
  }
  
  if (shapeExists("LastActivationTime")) {
    LastActivationTime.text = "";
  }
  
  if (shapeExists("LastActivationUser")) {
    LastActivationUser.text = "";
  }
  
  if (shapeExists("LoadedRcpDp")) {
    LoadedRcpDp.text = "";
  }
	
  if (shapeExists("RCCElements.RecipeElementsTable")) {
    setValue("RCCElements.RecipeElementsTable", "deleteAllLines");
  }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Create a predefined recipe instance to store the default privileges.
 * @param sClassName - [IN] Recipe class name.
 * @param sAppName   - [IN] Application name.
 * @param exceptionInfo - [OUT] Standar exception handling variable.
 */
void _unRecipeFunctions_createPredefinedRecipeInstance(string sClassName, string sAppName, dyn_string &exceptionInfo) {
  bool bEvent;
  int iDeviceNumber;
  string sDpName, sOperatorPrivileges, sExpertPrivileges, sAdminPrivileges, sPvssDpe, sPrefix, sFrontEndName;
  dyn_string dsOperatorPrivileges, dsExpertPrivileges, dsAdminPrivileges;
  
  // Check if the predefined recipe instance already exists
  sDpName = _unGenericDpFunctions_dpAliasToName(sClassName + "/" + UN_RCP_PREDEFINED_NAME);
  if (sDpName != "") {
    // The predefined recipe of the class already exists
	return;
  }
  
  // Get the recipe front end name
  unImportDevice_initialize();
  unImportDevice_getApplicationCharacteristics(bEvent, sPrefix);
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(UN_RECIPE_FE_ALIAS, sPvssDpe);
  if (sPvssDpe==UN_RECIPE_FE_ALIAS) {
	// The RCP_FE doesn't exist, create it
	bool bOk = unRecipeFunctions_createRecipeFrontEnd(sAppName, sPrefix, bEvent, exceptionInfo);
	if (bOk == false) {
		fwException_raise(exceptionInfo, "ERROR", "The " + UN_RECIPE_FE_ALIAS + " SOFT_FE couldn't be created.", "");
		return;
	}
	unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(UN_RECIPE_FE_ALIAS, sPvssDpe);
  }

  _unRecipeFunctions_createNewRecipeDp(UN_CONFIG_INTERNAL_UNRCPINSTANCE_DPT_NAME, UN_RECIPE_FE_ALIAS, sAppName, sDpName, iDeviceNumber); 
  dpCreate(sDpName, UN_CONFIG_INTERNAL_UNRCPINSTANCE_DPT_NAME);
  dpSetAlias(sDpName + ".", sClassName + "/" + UN_RCP_PREDEFINED_NAME);
  
  // Set the default privileges for the default recipe instance
  _unRecipeFunctions_getDeviceDefaultPrivileges(UN_CONFIG_UNRCPINSTANCE_DPT_NAME, sOperatorPrivileges, sExpertPrivileges, 
		sAdminPrivileges, exceptionInfo);
		
  dpSet(sDpName + ".accessControl.operator", sOperatorPrivileges,
	    sDpName + ".accessControl.expert", sExpertPrivileges,
		sDpName + ".accessControl.admin", sAdminPrivileges);
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Delete the predefined recipe instance of a recipe class.
 * @param sClassName - [IN] Recipe class name.
 * @return TRUE if the predefined recipe instance was deleted correctly, otherwise FALSE.
 */
private bool _unRecipeFunctions_deletePredefinedRecipeInstance(string sClassName) {
  int iRetVal;
  string sDpName, sPvssDpe;
  
  // Check if the predefined recipe instance already exists
  sDpName = _unGenericDpFunctions_dpAliasToName(sClassName + "/" + UN_RCP_PREDEFINED_NAME);
  if (sDpName == "") {
    // The predefined recipe of the class doesn't exist
    return false;
  }
  
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sDpName, sPvssDpe);
  iRetVal = dpDelete(sPvssDpe);
  
  return (iRetVal==0);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the default privileges of a recipe device (recipe class or recipe instance).
 * @param sDpType 			  - [IN] Datapoint type which default privileges are required.
 * @param sOperatorPrivileges - [OUT] Default privileges for the operator.
 * @param sExpertPrivileges   - [OUT] Default privileges for the expert.
 * @param sAdminPrivileges    - [OUT] Default privileges for the administrator.
 * @param exceptionInfo       - [OUT] Standard exception handling variable.
 */
_unRecipeFunctions_getDeviceDefaultPrivileges(string sDpType, string & sOperatorPrivileges, string & sExpertPrivileges, 
		string & sAdminPrivileges, dyn_string exceptionInfo) 
{
  string sFunction;
  dyn_string dsOperatorPrivileges, dsExpertPrivileges, dsAdminPrivileges;
  
  if (sDpType == UN_CONFIG_UNRCPCLASS_DPT_NAME) {
	sFunction = "_unRecipeFunctions_getDefaultClassPrivileges";
  } else if (sDpType == UN_CONFIG_UNRCPINSTANCE_DPT_NAME) {
    sFunction = "_unRecipeFunctions_getDefaultInstancePrivileges";
  } else {
    fwException_raise(exceptionInfo, "ERROR", "_unRecipeFunctions_getDeviceDefaultPrivileges - The DPType " +sDpType+ " is unknown." ,"");
    DebugTN(exceptionInfo);
    return;
  }
  
  string sFunctionDefinition = "dyn_string main(string sRoleType)"+
  										"{	dyn_string temp;"+
  										"	"+sFunction+"(sRoleType, temp);"+
  										"	return temp;"+
  										"}";
										
  evalScript(dsOperatorPrivileges, sFunctionDefinition, makeDynString(), "Operator");     
  evalScript(dsExpertPrivileges, sFunctionDefinition, makeDynString(), "Expert");  
  evalScript(dsAdminPrivileges, sFunctionDefinition, makeDynString(), "Admin");  
  fwGeneral_dynStringToString(dsOperatorPrivileges, sOperatorPrivileges, ",");
  fwGeneral_dynStringToString(dsExpertPrivileges, sExpertPrivileges, ",");
  fwGeneral_dynStringToString(dsAdminPrivileges, sAdminPrivileges, ",");
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Update the recipe class privileges.
 * @param sDpName - [IN] Recipe class DP name.
 */
void unRecipeFunctions_updateRecipeClassPrivileges(string sDpName) {
  int iPos;
  string sDeviceOperatorPrivileges, sDeviceExpertPrivileges, sDeviceAdminPrivileges;
  string sDefaultOperatorPrivileges, sDefaultExpertPrivileges, sDefaultAdminPrivileges;
  dyn_string dsDeviceOperatorPrivileges, dsDeviceExpertPrivileges, dsDeviceAdminPrivileges;
  dyn_string dsDefaultOperatorPrivileges, dsDefaultExpertPrivileges, dsDefaultAdminPrivileges;
  
  // Get the default privileges for the recipe class
  _unRecipeFunctions_getDefaultClassPrivileges("operator", dsDefaultOperatorPrivileges);
  _unRecipeFunctions_getDefaultClassPrivileges("expert", dsDefaultExpertPrivileges);
  _unRecipeFunctions_getDefaultClassPrivileges("admin", dsDefaultAdminPrivileges);
  fwGeneral_dynStringToString(dsDefaultOperatorPrivileges, sDefaultOperatorPrivileges, ",");
  fwGeneral_dynStringToString(dsDefaultExpertPrivileges, sDefaultExpertPrivileges, ",");
  fwGeneral_dynStringToString(dsDefaultAdminPrivileges, sDefaultAdminPrivileges, ",");
  
  // Get the device privileges
  dpGet(sDpName + ".statusInformation.accessControl.operator", sDeviceOperatorPrivileges,
	    sDpName + ".statusInformation.accessControl.expert", sDeviceExpertPrivileges,
		sDpName + ".statusInformation.accessControl.admin", sDeviceAdminPrivileges);
		  		  
  if (sDeviceOperatorPrivileges=="" && sDeviceExpertPrivileges=="" && sDeviceAdminPrivileges=="") {
	// Apply the default privileges
	dpSet(sDpName + ".statusInformation.accessControl.operator", sDefaultOperatorPrivileges,
		  sDpName + ".statusInformation.accessControl.expert", sDefaultExpertPrivileges,
		  sDpName + ".statusInformation.accessControl.admin", sDefaultAdminPrivileges);
  } else {
	// There are some privileges for the device, add the new privileges
	dsDeviceOperatorPrivileges = strsplit(sDeviceOperatorPrivileges, ",");
	dsDeviceExpertPrivileges   = strsplit(sDeviceExpertPrivileges, ",");
	dsDeviceAdminPrivileges    = strsplit(sDeviceAdminPrivileges, ",");
		
	// Replace the 'Apply' privilege by 'ApplyChanges' (from version 1.1.x to 1.2.0)
	if ( (iPos = dynContains(dsDeviceOperatorPrivileges, "Apply")) > 0)
		dsDeviceOperatorPrivileges[iPos] = "ApplyChanges";
	else if ( (iPos = dynContains(dsDeviceExpertPrivileges, "Apply")) > 0)
		dsDeviceExpertPrivileges[iPos] = "ApplyChanges";
	else if ( (iPos = dynContains(dsDeviceAdminPrivileges, "Apply")) > 0)
		dsDeviceAdminPrivileges[iPos] = "ApplyChanges";
			
	_unRecipeFunctions_addPrivileges(dsDefaultOperatorPrivileges, dsDeviceOperatorPrivileges, dsDeviceExpertPrivileges, dsDeviceAdminPrivileges);
	_unRecipeFunctions_addPrivileges(dsDefaultExpertPrivileges, dsDeviceExpertPrivileges, dsDeviceOperatorPrivileges, dsDeviceAdminPrivileges);
	_unRecipeFunctions_addPrivileges(dsDefaultAdminPrivileges, dsDeviceAdminPrivileges, dsDeviceOperatorPrivileges, dsDeviceExpertPrivileges);
		
	// Remove the operator obsolete privileges 
	_unRecipeFunctions_removeObsoletePrivileges(dsDeviceOperatorPrivileges, dsDefaultOperatorPrivileges,
			dsDefaultExpertPrivileges, dsDefaultAdminPrivileges);

	// Remove the expert obsolete privileges 
	_unRecipeFunctions_removeObsoletePrivileges(dsDeviceExpertPrivileges, dsDefaultOperatorPrivileges,
			dsDefaultExpertPrivileges, dsDefaultAdminPrivileges);
				
	// Remove the admin obsolete privileges 
	_unRecipeFunctions_removeObsoletePrivileges(dsDeviceAdminPrivileges, dsDefaultOperatorPrivileges, 
				dsDefaultExpertPrivileges, dsDefaultAdminPrivileges);	
				
	fwGeneral_dynStringToString(dsDeviceOperatorPrivileges, sDeviceOperatorPrivileges, ",");
	fwGeneral_dynStringToString(dsDeviceExpertPrivileges, sDeviceExpertPrivileges, ",");
	fwGeneral_dynStringToString(dsDeviceAdminPrivileges, sDeviceAdminPrivileges, ",");
		
	// Apply the device privileges
	dpSet(sDpName + ".statusInformation.accessControl.operator", sDeviceOperatorPrivileges,
		  sDpName + ".statusInformation.accessControl.expert", sDeviceExpertPrivileges,
		  sDpName + ".statusInformation.accessControl.admin", sDeviceAdminPrivileges);
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Update the recipe instances privileges.
 * @param sDpName - [IN] DP name of the recipe instance.
 * @param sClassName - [IN] Recipe class name.
 */
void unRecipeFunctions_updateRecipeInstancePrivileges(string sDpName, string sClassName) {
  string sPredefinedRcp;
  string sDeviceAcDomain,sDeviceOperatorPrivileges, sDeviceExpertPrivileges, sDeviceAdminPrivileges;
  string sDefaultAcDomain, sDefaultOperatorPrivileges, sDefaultExpertPrivileges, sDefaultAdminPrivileges;
  dyn_string dsDefaultOperatorPrivileges, dsDefaultExpertPrivileges, dsDefaultAdminPrivileges;
  dyn_string dsDeviceOperatorPrivileges, dsDeviceExpertPrivileges, dsDeviceAdminPrivileges;

  // Get the default privileges for the recipe instance from the predefined recipe instance
  unRecipeFunctions_getRecipeInstancePredefinedPrivileges(sClassName, sPredefinedRcp, sDefaultAcDomain, sDefaultOperatorPrivileges, sDefaultExpertPrivileges, sDefaultAdminPrivileges);
  _unRecipeFunctions_normalizeDp(sDpName);
  
  // Get the device privileges
  dpGet(sDpName + ".statusInformation.accessControl.accessControlDomain", sDeviceAcDomain,
        sDpName + ".statusInformation.accessControl.operator", sDeviceOperatorPrivileges,
        sDpName + ".statusInformation.accessControl.expert", sDeviceExpertPrivileges,
        sDpName + ".statusInformation.accessControl.admin", sDeviceAdminPrivileges);
		  		  
  // If the AcDomain is empty on the device, it applies the default AcDomain 
  if (sDeviceAcDomain == "" && sDefaultAcDomain!="") {
      dpSet(sDpName + ".statusInformation.accessControl.accessControlDomain", sDefaultAcDomain);
  }  
  
  if (sDeviceOperatorPrivileges=="" && sDeviceExpertPrivileges=="" && sDeviceAdminPrivileges=="") {
    // Apply the default privileges
    int iRes = dpSet(
      sDpName + ".statusInformation.accessControl.operator", sDefaultOperatorPrivileges,
      sDpName + ".statusInformation.accessControl.expert", sDefaultExpertPrivileges,
      sDpName + ".statusInformation.accessControl.admin", sDefaultAdminPrivileges);

    if (iRes<0) {
      dyn_errClass err = getLastError();
      DebugTN(err);
    }
	} else {
	  // There are some privileges for the device, add the new privileges
	  dsDeviceOperatorPrivileges = strsplit(sDeviceOperatorPrivileges, ",");
	  dsDeviceExpertPrivileges   = strsplit(sDeviceExpertPrivileges, ",");
	  dsDeviceAdminPrivileges    = strsplit(sDeviceAdminPrivileges, ",");
    dsDefaultOperatorPrivileges= strsplit(sDefaultOperatorPrivileges, ",");
    dsDefaultExpertPrivileges  = strsplit(sDefaultExpertPrivileges, ",");
    dsDefaultAdminPrivileges    = strsplit(sDefaultAdminPrivileges, ",");
			
	  // Add the operator default privileges
	  _unRecipeFunctions_addPrivileges(dsDefaultOperatorPrivileges, dsDeviceOperatorPrivileges,
    dsDeviceExpertPrivileges, dsDeviceAdminPrivileges);
				
	  // Add the expert default privileges
	  _unRecipeFunctions_addPrivileges(dsDefaultExpertPrivileges, dsDeviceExpertPrivileges,
    dsDeviceOperatorPrivileges, dsDeviceAdminPrivileges);
				
	  // Add the admin default privileges
	  _unRecipeFunctions_addPrivileges(dsDefaultAdminPrivileges, dsDeviceAdminPrivileges,
    dsDeviceOperatorPrivileges, dsDeviceExpertPrivileges);
		
	  // Remove the operator obsolete privileges 
	  _unRecipeFunctions_removeObsoletePrivileges(dsDeviceOperatorPrivileges, dsDefaultOperatorPrivileges,
    dsDefaultExpertPrivileges, dsDefaultAdminPrivileges);

	  // Remove the expert obsolete privileges 
	  _unRecipeFunctions_removeObsoletePrivileges(dsDeviceExpertPrivileges, dsDefaultOperatorPrivileges,
    dsDefaultExpertPrivileges, dsDefaultAdminPrivileges);
				
	  // Remove the admin obsolete privileges 
	  _unRecipeFunctions_removeObsoletePrivileges(dsDeviceAdminPrivileges, dsDefaultOperatorPrivileges, 
    dsDefaultExpertPrivileges, dsDefaultAdminPrivileges);				
		
	  fwGeneral_dynStringToString(dsDeviceOperatorPrivileges, sDeviceOperatorPrivileges, ",");
	  fwGeneral_dynStringToString(dsDeviceExpertPrivileges, sDeviceExpertPrivileges, ",");
	  fwGeneral_dynStringToString(dsDeviceAdminPrivileges, sDeviceAdminPrivileges, ",");
		
	  // Apply the device privileges
	  int iRes = dpSet(
		  sDpName + ".statusInformation.accessControl.operator", sDeviceOperatorPrivileges,
		  sDpName + ".statusInformation.accessControl.expert", sDeviceExpertPrivileges,
		  sDpName + ".statusInformation.accessControl.admin", sDeviceAdminPrivileges);
	}
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Auxiliary function to add the default privileges to the privileges list.
 * If none of the dyn_strings (dsDevicePrivileges1, dsDevicePrivileges2, dsDevicePrivileges3) contains the privileges
 * included in dsDefaultPrivileges, the privilege will be added to dsDevicePrivileges1. Otherwise (the privilege is included
 * in any dyn_string) no actions will be done.
 * @param dsDefaultPrivileges - [IN] The list of privileges to add.
 * @param dsDevicePrivileges1 - [OUT] The list where add the privileges if they don't exist in any dyn_string.
 * @param dsDevicePrivileges2 - [IN] Device privileges list.
 * @param dsDevicePrivileges3 - [IN] Device privileges list.
 */
void _unRecipeFunctions_addPrivileges(dyn_string dsDefaultPrivileges, dyn_string &dsDevicePrivileges1, 
					dyn_string dsDevicePrivileges2, dyn_string dsDevicePrivileges3)
{
  int i, iLen;
  string sPrivilege;
  
  iLen = dynlen(dsDefaultPrivileges);
  for (i=1; i<=iLen; i++) {
    sPrivilege = dsDefaultPrivileges[i];
	
	if (dynContains(dsDevicePrivileges1, sPrivilege) || 
        dynContains(dsDevicePrivileges2, sPrivilege) ||
        dynContains(dsDevicePrivileges3, sPrivilege) ) {
            continue;
    }
		
	// The current privilege is not assigned to any role,
	// add it to the default role.
	dynAppend(dsDevicePrivileges1, sPrivilege);
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Auxiliary function to remove the obsolete privileges from the privileges list.
 * @param dsDevicePrivileges - [OUT] List where the obsolete parameters will be removed.
 * @param dsDefaultPrivileges1 - [IN] Default privileges list 1 (e.g. operator).
 * @param dsDefaultPrivileges2 - [IN] Default privileges list 2 (e.g. expert). 
 * @param dsDefaultPrivileges3 - [IN] Default privileges list 3 (e.g. admin).
 */
void _unRecipeFunctions_removeObsoletePrivileges(dyn_string & dsDevicePrivileges, dyn_string dsDefaultPrivileges1, 
				dyn_string dsDefaultPrivileges2, dyn_string dsDefaultPrivileges3) 
{
	int i, iLen;
	
	iLen = dynlen(dsDevicePrivileges);
	for (i=1; i<=iLen; i++) {
	  string sPrivilege = dsDevicePrivileges[i];
	  
	  if (dynContains(dsDefaultPrivileges1, sPrivilege) ||
          dynContains(dsDefaultPrivileges2, sPrivilege) || 
          dynContains(dsDefaultPrivileges3, sPrivilege)) 
      {
		continue;
      }

	  // Remove the obsolete privilege
	  dynRemove(dsDevicePrivileges, i);
	  iLen--;
	  i--;
	}
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the device mapping from a recipe class.
 * @param rcpClassObject - [IN] The recipe class object from the JCOP recipe class definition.
 * @param rcpDeviceDpes  - [OUT] Mapping where: Key=device name, Value=list of device DPEs included in the recipe.
 */
void _unRecipeFunctions_getRecipeClassDeviceMapping(dyn_mixed rcpClassObject, mapping & rcpDeviceDpes) {
  int i, iLen;
  dyn_string rcpElements;
  
  // For each device in the recipe, get all the DPEs included in the recipe and
  // store them in the mapping rcpDeviceDpes as a string (comma separated)
  rcpElements = rcpClassObject[fwConfigurationDB_RCL_ELEMENTS];
  iLen = dynlen(rcpElements);
  for (i=1; i<=iLen; i++) {
	string sRcpElement = rcpElements[i];
    string sAlias = substr(sRcpElement, 0, strpos(sRcpElement, "."));
	dyn_string rcpElementSplit = strsplit(sRcpElement, ".");
	string rcpAliasDpes = "";
	if (mappingHasKey(rcpDeviceDpes, sAlias)) {
	  rcpAliasDpes = rcpDeviceDpes[sAlias];
	}
	
	if (strlen(rcpAliasDpes)) {
	  rcpAliasDpes += ", ";
    }
	rcpAliasDpes += rcpElementSplit[dynlen(rcpElementSplit)];
	rcpDeviceDpes[sAlias] = rcpAliasDpes;
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the recipe class data in the recipe class panel.
 * @param sRcpClassDp - [IN] Datapoint name of the recipe class which data will be loaded in the panel.
 */
void unRecipeFunctions_loadRecipeClassData(string sRcpClassDp) {
  string sSystemName = "", sClassName;
  dyn_mixed recipeClassInfo;
  dyn_string dsDeviceType, dsAlias, dsElements, dsErrorRows, exceptionInfo;
  mapping rcpDeviceDpes;
  
  _unRecipeFunctions_normalizeDp(sRcpClassDp);
  sRcpClassDp += ".";
  
  sSystemName = unGenericDpFunctions_getSystemName(sRcpClassDp);
  dpGet(sRcpClassDp + "ProcessInput.ClassName", sClassName);
  unRecipeFunctions_getRecipeClassObject(sRcpClassDp, recipeClassInfo, exceptionInfo);
  
  RecipeClass.text         = recipeClassInfo[fwConfigurationDB_RCL_CLASSNAME];
  RecipeType.text          = recipeClassInfo[fwConfigurationDB_RCL_RECIPETYPE];
  RecipeClassComment.text  = recipeClassInfo[fwConfigurationDB_RCL_DESCRIPTION];
  LastModifiedTime.text    = recipeClassInfo[fwConfigurationDB_RCL_LAST_MODIFIED_TIME];
  LastModifiedUser.text    = recipeClassInfo[fwConfigurationDB_RCL_LAST_MODIFIED_USER];
  LastActivatedRecipe.text = recipeClassInfo[fwConfigurationDB_RCL_LAST_ACTIVATED_RECIPE];
  LastActivationTime.text  = recipeClassInfo[fwConfigurationDB_RCL_LAST_ACTIVATED_TIME];
  LastActivationUser.text  = recipeClassInfo[fwConfigurationDB_RCL_LAST_ACTIVATED_USER];
  
  // For each device in the recipe, get all the DPEs included in the recipe and
  // store them in the mapping rcpDeviceDpes as a string (comma separated)
  _unRecipeFunctions_getRecipeClassDeviceMapping(recipeClassInfo, rcpDeviceDpes);
  
  // Add the recipe elements to the table
  setValue("RCCElements.RecipeElementsTable", "deleteAllLines");
  
  dyn_string devices = _unRecipeFunctions_getRcpClassDeviceOrder(sRcpClassDp, recipeClassInfo[fwConfigurationDB_RCL_DEVICES], exceptionInfo);
  if (dynlen(devices) > 0) {    
    if (dynlen(exceptionInfo)) {
      DebugTN("Warning: The list of devices can not be ordered: " + exceptionInfo[2]);
      dynClear(exceptionInfo);
    }
      
    setValue("RCCElements.RecipeElementsTable", "enabled", FALSE);
    dyn_string dsDeviceType, dsAlias, dsElements;
    for (int i=1; i<=dynlen(devices); i++) {
      string deviceType, pvssDpe, sAlias;
      unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sSystemName + devices[i], pvssDpe);
      if (dpExists(pvssDpe)) {
        deviceType = dpTypeName(pvssDpe);
      } else {
        dynAppend(dsErrorRows, i);
      }
      sAlias = devices[i];
      dynAppend(dsDeviceType, deviceType);
      dynAppend(dsAlias, sAlias);
      dynAppend(dsElements, rcpDeviceDpes[sAlias]);
    }   
    
    setMultiValue("RCCElements.RecipeElementsTable", "appendLines", dynlen(dsDeviceType),
               "DeviceType", dsDeviceType, 
               "DeviceName", dsAlias, 
               "DeviceElements", dsElements);
      
    if (dynlen(dsErrorRows)) {
      // Set the recipe instance in INVALID state
      dpSet(sRcpClassDp+"ProcessInput.State", UN_RCP_CLASS_STATE_INVALID);
      _unRecipeFunctions_setInvalidRows(dsErrorRows, "RCCElements.RecipeElementsTable");
      _unRecipeFunctions_writeInRecipeLog("Error: Some elements of the recipe class '" + sClassName + "' do not exist.");
      _unRecipeFunctions_writeInRecipeLog("Please remove the missing recipe elements.");
    }
  }

  setMultiValue(
    "RCCElements.RecipeElementsTable", "deleteSelection",
    "RCCElements.RecipeElementsTable", "enabled", TRUE);	
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if a recipe class definition is valid.
 * @param sClassDp      - [IN]  Datapoint name of the recipe class.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
bool _unRecipeFunctions_isRecipeClassDefinitionValid(string sClassDp, dyn_string &exceptionInfo) {
  string /*sClassName, */sSystemName, sPvssDpe;
  dyn_mixed recipeClassInfo;
  
  _unRecipeFunctions_normalizeDp(sClassDp);
  sSystemName = unGenericDpFunctions_getSystemName(sClassDp);
  unRecipeFunctions_getRecipeClassObject(sClassDp, recipeClassInfo, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    return FALSE;
  }
  
  dyn_string dsDevices = recipeClassInfo[fwConfigurationDB_RCL_DEVICES];
  int iLen = dynlen(dsDevices);
  for (int i=1; i<=iLen; i++) {
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sSystemName + dsDevices[i], sPvssDpe);
    if (!dpExists(sPvssDpe)) {
      fwException_raise(exceptionInfo, "ERROR", "The device " + dsDevices[i] + " does not exist." ,"");
      return FALSE;
    }
  }
  
  return TRUE;
}
//------------------------------------------------------------------------------------------------------------------------
dyn_string _unRecipeFunctions_getRcpClassDeviceOrder(string sRcpClassDp, dyn_string dsDevices, dyn_string &exceptionInfo) {
  int iLen;
  dyn_string dsOrderedDevices;
  
  dynUnique(dsDevices);
  dpGet(sRcpClassDp + "ProcessInput.DeviceOrder", dsOrderedDevices);
  if (dynlen(dsOrderedDevices)==0) {
    return dsDevices;
  }
  
  // Check if the lists don't contain the same elements
  if (!_unRecipeFunctions_areListsEqual(dsDevices, dsOrderedDevices)) {
    fwException_raise(exceptionInfo, "ERROR", "The device lists don't have the same devices.","");
    return dsDevices;
  }
  
  dpGet(sRcpClassDp + "ProcessInput.DeviceOrder", dsOrderedDevices);
  return dsOrderedDevices;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Select the specified recipe class in the recipe class panel.
 * @param sRcpClassDp   - [IN] Datapoint name of the recipe class.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @param bForceReload  - [IN] Boolean value to force the reload of the recipe class data.
 */ 
void unRecipeFunctions_selectRecipeClass(string sRcpClassDp, dyn_string &exceptionInfo, bool bForceReload=FALSE)
{
  int iPos, iRes, iRangeMin, iRangeMax;  
  string sClassName, sButtonPanel, sLoadedRcpDp/*, sSystemName*/;
  dyn_string classNames;
  
  // If the sRcpClassDp ends with ".", remove the dot from the DP
  _unRecipeFunctions_normalizeDp(sRcpClassDp);
 
  //sSystemName = unGenericDpFunctions_getSystemName(sRcpClassDp);
  sClassName  = unGenericDpFunctions_getAlias(sRcpClassDp + ".");
  getMultiValue("RecipeClasses", "getColumnN", 0, classNames,
				"LoadedRcpDp", "text", sLoadedRcpDp);
  
  iPos = dynContains(classNames, sClassName);
  if (iPos<=0) {
	return;
  }

  // Check if the recipe class is already loaded or the reload has been forced
  if (sRcpClassDp != sLoadedRcpDp || bForceReload)
  {	
    // Select the recipe class in the table (if it's not already selected)
    dyn_int diSelLines;    
    getValue("RecipeClasses", "getSelectedLines", diSelLines);

    if (!dynContains(diSelLines, iPos-1)) {
        RecipeClasses.selectLineN(iPos-1);
    }
	
	if (shapeExists("RCCElements")) {
	  removeSymbol(myModuleName(), myPanelName(), "RCCElements");
	}

    addSymbol(myModuleName(), myPanelName(), "vision/unRecipe/UnRcpClass/unUnRcpClass_ClassElements.pnl",
				"RCCElements", makeDynString("$sDpName:"+sRcpClassDp), 190, 220, 0, 1, 1);
	
	unRecipeFunctions_loadRecipeClassData(sRcpClassDp);
	setValue("LoadedRcpDp", "text", sRcpClassDp);		
  }
  
  getValue("RecipeClasses", "lineRangeVisible", iRangeMin, iRangeMax);
  if (iRangeMin > (iPos-1) || (iPos-1) > iRangeMax) {
    setValue("RecipeClasses", "lineVisible", (iPos-1));
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the recipe class data from a recipe instance datapoint.
 * @param sRcpInstanceDp - [IN] Datapoint name of the recipe instance.
 * @param sRcpClassName  - [OUT] Name of the recipe class related to the recipe instance.
 * @param sRcpClassDp    - [OUT] Datapoint name of the recipe class related to the recipe instance.
 * @param sPcoLink       - [OUT] Device link of the recipe class.
 */
void _unRecipeFunctions_getRecipeClassDataFromInstance(string sRcpInstanceDp, string & sRcpClassName, string & sRcpClassDp,
		string & sPcoLink) 
{
  string sSystemName;
  
  _unRecipeFunctions_normalizeDp(sRcpInstanceDp);
  sSystemName = unGenericDpFunctions_getSystemName(sRcpInstanceDp);
  
  // Get the recipe class name of the recipe instance
  if (!dpExists(sRcpInstanceDp)) {
    return;
  }
  
  dpGet(sRcpInstanceDp + ".ProcessInput.ClassName", sRcpClassName);
  if (sRcpClassName == "") {
    // If the recipe is in a remote system and this is disconnected, the recipe class name will be empty
    _unRecipeFunctions_writeInRecipeLog("Error: Could not get data associated to the recipe class");
    return;
  }
  
  // Get the recipe class datapoint name and the recipe class pco link 
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sSystemName + sRcpClassName, sRcpClassDp);
  _unRecipeFunctions_normalizeDp(sRcpClassDp);
  if (!dpExists(sRcpClassDp)) {
    return;
  }
  
  dpGet(sRcpClassDp + ".statusInformation.deviceLink", sPcoLink);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Selects the specified recipe instance in the recipe instance panel.
 * @param sRcpInstanceDp - [IN] Datapoint of the recipe instance which will be selected.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
bool unRecipeFunctions_selectRecipeInstance(string sRcpInstanceDp, dyn_string &exceptionInfo)
{
  int iPos, iSelected, pcoPos;
  dyn_int selectedLines;
  bool bInitial, bExists;
  string sFaceplatePanel, sButtonPanel, sClassName, sInstanceName, sLoadedRcpDp, sPreviousRcpDp, sState;
  string sRcpClassDp, sPcoLink, sSelectedPco, sJcopRcpClassDp;
  dyn_string rcpClassesLoaded, rcpInstancesDp, pcoItems;  
    
  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeClasses") || !shapeExists("RecipeInstances") || 
      !shapeExists("RecipeClassLabel") || !shapeExists("RecipeInstanceLabel")) {
  	fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	  return false;
  }

  // Check if the DP name exists
  if (!dpExists(sRcpInstanceDp)) {
    fwException_raise(exceptionInfo, "ERROR", "The DP name " + sRcpInstanceDp + " doesn't exist.","");
    return false;
  }
  
  // Get the recipe class data from the recipe instance datapoint
  _unRecipeFunctions_getRecipeClassDataFromInstance(sRcpInstanceDp, sClassName, sRcpClassDp, sPcoLink);
  
  // Get the recipe instance name 
  dpGet(sRcpInstanceDp+".ProcessInput.InstanceName", sInstanceName);
  // Selects the PCO link in the combo box (if necessary)
  getValue("PcoList", "selectedText", sSelectedPco);
  if (sPcoLink != sSelectedPco) {  
    string sPcoDp, sApplication;
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sPcoLink, sPcoDp);
    sApplication = unGenericDpFunctions_getApplication(sRcpInstanceDp);
	  unRecipeFunctions_selectPcoRecipes(sPcoDp, exceptionInfo, false, sApplication);
  } 
  
  // Get the recipe classes loaded in the combo box
  getValue("RecipeClasses", "items", rcpClassesLoaded);
  iPos = dynContains(rcpClassesLoaded, sClassName);  
  if (iPos<=0) {
    // The recipe class of the recipe instance doesn't exist!
    iPos = dynContains(rcpClassesLoaded, "-None-");
    if (iPos<=0) {
      setValue("RecipeClasses", "appendItem", "-None-");
      setValue("RecipeClasses", "selectedPos", dynlen(rcpClassesLoaded)+1);
    } else {
      setValue("RecipeClasses", "selectedPos", iPos);
    }
    return false;
  } 
  
  getMultiValue("RecipeClasses", "items", rcpClassesLoaded,
			    "RecipeClasses", "selectedPos", iSelected);
  iPos = dynContains(rcpClassesLoaded, sClassName);
  
  // Selects the recipe class in the combo box
  if (iPos!=iSelected) {   
    setValue("RecipeClasses", "selectedPos", iPos);
    _unRecipeFunctions_loadRecipeInstances(sClassName);
  } 
    
  // Selects the recipe instance in the table
  getMultiValue("RecipeInstances", "getSelectedLines", selectedLines,
				"RecipeInstances", "getColumnN", 1, rcpInstancesDp,
				"LoadedRecipeDp", "text", sLoadedRcpDp);
  iPos = dynContains(rcpInstancesDp, sRcpInstanceDp);
  if (iPos<=0) { 
    // The required recipe doesn't exist, an error occurred in the dynContains method
    // or there aren't selected lines in the table
    return false;
  } 
  
  if (dynlen(selectedLines)<=0 || (iPos-1)!=selectedLines[1] || sLoadedRcpDp=="" || sLoadedRcpDp!=sRcpInstanceDp) {
	  setValue("RecipeInstances", "selectLineN", (iPos-1));
   
	  // add  device action
	  if (shapeExists("buttonActivate")) {
      removeSymbol(myModuleName(), myPanelName(), "RCPButton");
      removeSymbol(myModuleName(), myPanelName(), "RCPElementsTable");
	  }
			
	  // get the button panel of the device from UNICOS configuration
	  unGenericDpFunctions_getContextPanel(sRcpInstanceDp, sButtonPanel, exceptionInfo);
	  if(dynlen(exceptionInfo)) {
      DebugTN(exceptionInfo); 
      return false;
    }

	  addSymbol(myModuleName(), myPanelName(), sButtonPanel, "RCPButton", makeDynString("$sDpName:"+sRcpInstanceDp), 40, 620, 0, 1, 1);
	  addSymbol(myModuleName(), myPanelName(), "vision/unRecipe/UnRcpInstance/unUnRcpInstance_RecipeElements.pnl", 
				"RCPElementsTable", makeDynString(), 190, 193, 0, 1, 1);

	  unRecipeFunctions_loadRecipeInstanceData(); 
	  setMultiValue("RecipeClassLabel", "text", sClassName,
					"RecipeInstanceLabel", "text", sInstanceName,
					"LoadedRecipeDp", "text", sRcpInstanceDp);	
  } 
	  
  // Forces the reload of the recipe state to animate the Select button
  dpGet(sRcpInstanceDp+".ProcessInput.State", sState);
  dpSet(sRcpInstanceDp+".ProcessInput.State", sState);

  // Animate the Selection Progress Bar
  startThread("unRecipeFunctions_animateSelectionProgressBar", sRcpInstanceDp);
  
  //DebugTN("END = unRecipeFunctions_selectRecipeInstance " + dpGetAlias(sRcpInstanceDp + ".") + "\n");
  
  return true;
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Callback function executed when the recipe class elements have been modified.
 * The function will enable a warning message to indicate that the recipe must be reloaded.
 */
void _unRecipeFunctions_dpConnectRcpClassElementsCB(string sRcpClassElementsDp, dyn_string dsValues) {
  _unRecipeFunctions_enableRecipeInstanceWarning(true, UN_RCP_WARNING_MESSAGE_CLASS_MODIFIED);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Function to show/hide the recipe instance warning message.
 * @param bEnable  - [IN] Flag to specify the visibility of the warning message.
 * @param sMessage - [IN] Message to display in the WarningText label.
 */
void _unRecipeFunctions_enableRecipeInstanceWarning(bool bEnable, string sMessage="") {
  if (shapeExists("WarningIcon")) {
	setValue("WarningIcon", "visible", bEnable);
  }
	
  if (shapeExists("WarningText")) {
    setValue("WarningText", "text", sMessage);
	setValue("WarningText", "visible", bEnable);	
  }
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Callback function executed when the recipe values have been modified.
 * The function will enable a warning message to indicate that the recipe must be reloaded.
 */
void _unRecipeFunctions_dpConnectRcpValuesCB(string sRcpValuesDp, dyn_string dsValues) {
  bool bVisible;
  string sRecipeName, sRecipeDp;
  dyn_string dsLoadedValues, dsRcpDpSplit, exceptionInfo;
  dyn_dyn_mixed recipeObject;
  
  getValue("WarningText", "visible", bVisible);
  if (bVisible) {
    return;
  }
  
  int iCol = RecipeElements.nameToColumn("Value");
  getValue("RecipeElements", "getColumnN", iCol, dsLoadedValues);
  
  int iLen = dynlen(dsLoadedValues);
  if (iLen != dynlen(dsValues)) {
    _unRecipeFunctions_enableRecipeInstanceWarning(true, UN_RCP_WARNING_MESSAGE_VALUES_MODIFIED);
    return;
  }
  
  // Compares the recipe object values with the values loaded in the recipe table
  dsRcpDpSplit = strsplit(sRcpValuesDp, ".");  
  dpGet(dsRcpDpSplit[1] + ".RecipeName", sRecipeName);
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sRecipeName, sRecipeDp);
  _unRecipeFunctions_getRecipeObjectFromDp (sRecipeDp, recipeObject, exceptionInfo); 
  if (dynlen(exceptionInfo)) {
    return;
  }

  for (int i=1; i<=iLen; i++) {
    if (dsLoadedValues[i] != recipeObject[fwConfigurationDB_RO_VALUE][i]) {
      _unRecipeFunctions_enableRecipeInstanceWarning(true, UN_RCP_WARNING_MESSAGE_VALUES_MODIFIED);
      return;
    }
  }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Gets a mapping where the key is the application name and the value is a dyn_string containing the PCOs in that application.
 * @param bLocalOnly - [IN] TRUE if and only if the local applications must be included, otherwise FALSE.
 * @param bOnlyPcosWithRecipes - [IN] TRUE if and only if the PCOs with recipes must be included, otherwise FALSE.
 */
// TODO: Rename function to end with suffix 'Mapping'
public mapping unRecipeFunctions_getApplicationPcos(bool bLocalOnly=TRUE, bool bOnlyPcosWithRecipes=FALSE) {
  int i, iLen;
  mapping mApplicationPcos;
  dyn_string dsSelectedDeviceTypes, dsRetDpNames, dsRetAppl, dsPcoNames, dsPcoNamesWithRecipes, dsPcoDpsWithRecipes;
  
  dsSelectedDeviceTypes = makeDynString(UN_CONFIG_CPC_PROCESSCONTROLOBJECT_DPT_NAME);
  _unRecipeFunctions_getFilteredDevices(dsSelectedDeviceTypes, dsRetDpNames, dsPcoNames, dsRetAppl, bLocalOnly);
  
  iLen = dynlen(dsPcoNames);
  if (iLen == 0) {
    return mApplicationPcos;    
  }
  
  if (bOnlyPcosWithRecipes) {
    unRecipeFunctions_getPcosWithRecipes(dsPcoNamesWithRecipes, dsPcoDpsWithRecipes, bLocalOnly, "*", dsPcoNames, dsRetDpNames);
  }
  
  for (i=1; i<=iLen; i++) {
     string key = dsRetAppl[i];
     string sPcoName = dsPcoNames[i];
     if (bOnlyPcosWithRecipes && !dynContains(dsPcoNamesWithRecipes, sPcoName)) {
       continue;
     }
    _unRecipeFunctions_mappingInsertValue(mApplicationPcos, key, sPcoName); 
  }
  
  return mApplicationPcos;
}
//------------------------------------------------------------------------------------------------------------------------
public dyn_string unRecipeFunctions_getApplicationPcoList(mapping mApplicationPcos, string sApplication) {
  dyn_string result;
  
  dynAppend(result, "*");
  if (sApplication =="*") {
    int i, iLen;
    dyn_string keys = mappingKeys(mApplicationPcos);
    iLen = dynlen(keys);
    for (i=1; i<=iLen; i++) {
      string key = keys[i];
      dyn_string value = mApplicationPcos[key];
      dynAppend(result, value);
    }
  } else if (mappingHasKey(mApplicationPcos, sApplication)) {
    dyn_string dsApplicationPcos = mApplicationPcos[sApplication];
    dynAppend(result, dsApplicationPcos);    
  }
  
  return result;
} 
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Function to select the PCO recipes in the recipe instance panel.
 * @param sPcoDp - [IN] DP Name of the PCO.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @param bSelectFirstInstance - [IN] Flag to specify if the first recipe must be selected (TRUE by default).
 * @param sApplication - [IN] Application name where the PCOs must be loaded.
 */
void unRecipeFunctions_selectPcoRecipes(string sPcoDp, dyn_string &exceptionInfo, bool bSelectFirstInstance=true, string sApplication="*") {
  int iNumRecipes;
  bool bLocalOnly = FALSE;
  string sPcoAlias, sRcpClassName, sRcpInstanceDp, sApplicationName;
  dyn_string dsPcoList, dsRcpClasses, dsPcoDpNames, dsRetAppl;
  
  if (!shapeExists("PcoList") || !shapeExists("RecipeInstances") || !shapeExists("PcoLabel")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    return;
  }
    
  _unRecipeFunctions_normalizeDp(sPcoDp);
  sPcoDp = sPcoDp + ".";	
	
  sPcoAlias = unGenericDpFunctions_getAlias(sPcoDp);
  PcoLabel.text = sPcoAlias;

  // Get the PCOs loaded in the combo box and select the sPcoAlias if it exists
  _unRecipeFunctions_loadPcosWithRecipes(sApplication); 
  if (!unRecipeFunctions_selectItemInComboBox("PcoList", sPcoAlias)) {
    return;
  }
  
  // Load the recipe classes for the selected PCO
  unRecipeFunctions_loadRecipeClasses(sPcoAlias, exceptionInfo);
  
  // Load the recipe instances of the selected recipe class
  getValue("RecipeClasses", "items", dsRcpClasses);  
  sRcpClassName = dsRcpClasses[1];
  if (sRcpClassName == "-None-") {
    setValue("RecipeInstances", "deleteAllLines");
    return; 
  } 
  
  _unRecipeFunctions_loadRecipeInstances(sRcpClassName);

  // Select the first recipe instance (if it exists)
  getValue("RecipeInstances", "lineCount", iNumRecipes);
  if (bSelectFirstInstance && iNumRecipes>0) {
    getValue("RecipeInstances", "cellValueRC", 0, "RecipeInstancesDP", sRcpInstanceDp);
    unRecipeFunctions_selectRecipeInstance(sRcpInstanceDp, exceptionInfo);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Selects an item in a combo box.
 * @param sShape - [IN] Combo box name.
 * @param sSelect - [IN] Item to select.
 * @param indexOtherwise - [IN] Index to select in case the selected item does not exist.
 */
public bool unRecipeFunctions_selectItemInComboBox(string sShape, string sSelect, int indexOtherwise=1) {  
  dyn_string dsItems;
  
  if (!shapeExists(sShape)) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    return false;
  }    
  
  getValue(sShape, "items", dsItems);
  int iPos = dynContains(dsItems, sSelect);
  if (iPos>0) {
    setValue(sShape, "selectedPos", iPos);
  } else if (dynlen(dsItems)>1 && indexOtherwise>0) {
    setValue(sShape, "selectedPos", indexOtherwise);
  } else {
    return false;
  }
  
  return true;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Method used to animate the selection progress bar in the recipe instance panel.
 * @param sDpName - [IN] Data point name of the selected recipe instance.
 */
void unRecipeFunctions_animateSelectionProgressBar (string sDpName) 
{
  int iTimeOut;
  dyn_string exceptionInfo;
  string sRemainingTime, sSelectedManager, sSelectedRecipeDp;
  bool bSelect;

  // Get the selected recipe instance in the panel
  unRecipeFunctions_getSelectedRecipeDp(sSelectedRecipeDp, exceptionInfo); 
  if (sDpName != sSelectedRecipeDp) {
	unProgressBar_setPosition("SelectionProgressBar", 0);  
	return;
  }
	
  /** Selection Progress Bar animation **/
  if (shapeExists("SelectionProgressBar") && shapeExists("buttonSelect")) {
    delay (0, 50);
    dpGet(sDpName + ".statusInformation.selectedManager", sSelectedManager);
    bSelect = (sSelectedManager != "");
    if (bSelect) {
      dpGet(getSystemName() + "_unSelectDeselect.remainingTime", sRemainingTime,
		         getSystemName() + "_unSelectDeselect.timeOut", iTimeOut);      
      
      unProgressBar_setMax("SelectionProgressBar", iTimeOut);
      unProgressBar_setPosition("SelectionProgressBar", iTimeOut);

	  while (strsplit(sRemainingTime, ";")[3] != "0" && bSelect) {
		unRecipeFunctions_getSelectedRecipeDp(sSelectedRecipeDp, exceptionInfo);
		if (sDpName != sSelectedRecipeDp) {
			return;
		}
        dpGet(getSystemName() + "_unSelectDeselect.remainingTime", sRemainingTime);
        unProgressBar_setPosition("SelectionProgressBar", strsplit(sRemainingTime, ";")[3]);
		delay(5);
		// If the DP was removed, return
		if (!dpExists(sDpName + ".statusInformation.selectedManager"))
			return;
		dpGet(sDpName + ".statusInformation.selectedManager", sSelectedManager);
		bSelect = (sSelectedManager != "");
	  }
    }
    
	unRecipeFunctions_getSelectedRecipeDp(sSelectedRecipeDp, exceptionInfo);
	if (sDpName != sSelectedRecipeDp) {
		return;
	}
	unProgressBar_setPosition("SelectionProgressBar", 0);     
  } 
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the list of PCOs with recipes
 * @param dsPcoNamesWithRecipes - [OUT] Alias of the PCOs with recipes found.
 * @param dsPcoDpsWithRecipes   - [OUT] Dp names of the PCOs with recipes found.
 * @param bLocalOnly - [IN] TRUE if and only if the PCOs on the local application must be queried, otherwise FALSE.
 * @param sApplication - [IN] Applications where to look for the PCOs with recipes.
 * @param dsAllPcos - [IN] List of PCO aliases, if not provided it will search for the PCOs in the specified applications and systems.
 * @param dsAllPcoDps - [IN] List of PCO dp names, if not provided it will search for the PCO dpnames in the specified applications and systems.
 */
public void unRecipeFunctions_getPcosWithRecipes(dyn_string &dsPcoNamesWithRecipes, dyn_string &dsPcoDpsWithRecipes,
        bool bLocalOnly=FALSE, string sApplication ="*", dyn_string dsAllPcos=makeDynString(), dyn_string dsAllPcoDps=makeDynString()) {
  int i, iLen, j;  
  dyn_string dsRetAppl, dsRetPcoDpNames, dsRetPcoNames;
  dyn_dyn_anytype resultTable;
  
  if (dynlen(dsAllPcos) == 0 || dynlen(dsAllPcoDps) ==0) {
    _unRecipeFunctions_getFilteredDevices(makeDynString(UN_CONFIG_CPC_PROCESSCONTROLOBJECT_DPT_NAME), 
        dsAllPcoDps, dsAllPcos, dsRetAppl, bLocalOnly, makeDynString(sApplication));
  }
  
  // Get the list of PCOs with recipe classes   
  iLen = dynlen(dsAllPcoDps);
  dyn_string dsQueriedSystems;
  for (i = 1; i<=iLen; i++) {    
    string sSystemName = unGenericDpFunctions_getSystemName(dsAllPcoDps[i]); 
    if (dynContains(dsQueriedSystems, sSystemName)) {
      continue;
    }
    dynAppend(dsQueriedSystems, sSystemName);
    string sQuery = "SELECT '.statusInformation.deviceLink:_online.._value' FROM '*' REMOTE '" + sSystemName + "' WHERE _DPT=\"UnRcpClass\" "; 
    dpQuery(sQuery, resultTable);   	  
    
    int jLen = dynlen(resultTable);
    for (j = 2; j<=jLen; j++) {
      if (!dynContains(dsPcoNamesWithRecipes, resultTable[j][2])) {
        int iPos = dynContains(dsAllPcos, resultTable[j][2]);
        if (iPos > 0 ) {
          string sPcoDp = dsAllPcoDps[iPos];
          dynAppend(dsPcoNamesWithRecipes, resultTable[j][2]);
          dynAppend(dsPcoDpsWithRecipes, sPcoDp);      
        }
      }
    }
  } 
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the PCO List in the PcoList combo box.
 * @param sApplication - [IN] Application name where to load the PCOs from.
 */
public void _unRecipeFunctions_loadPcosWithRecipes(string sApplication) {
	dyn_string dsPcoNamesWithRecipes, dsPcoDpsWithRecipes, dsPcoDpNames, dsPcoList, dsRetAppl;
	
	// Check that all the necessary shapes exist
	if (!shapeExists("PcoList")) {
		fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
		return;
	}
                                        
  // Get the list of PCOs with recipe classes  
  bool bLocalOnly=FALSE;
  unRecipeFunctions_getPcosWithRecipes(dsPcoNamesWithRecipes, dsPcoDpsWithRecipes, bLocalOnly, sApplication);   
  	
	//PcoList.appendItem("*");
  dynInsertAt(dsPcoNamesWithRecipes, "*", 1);
  PcoList.deleteAllItems();
  PcoList.items = dsPcoNamesWithRecipes;
} 
 
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Method used to load the recipe class list associated to a PCO
 * @param sPcoLink Alias of the PCO which recipe classes will be loaded or "*" for all the available classes.
 */ 
public void _unRecipeFunctions_loadRecipeClassList(string sPcoLink="*", string sApplication="*")
{
	dyn_dyn_anytype resultTable;
	dyn_string recipeClassNames, recipeClassDps, recipeClassComments, exceptionInfo;
	dynClear(recipeClassNames);
	dynClear(recipeClassComments);

	// Check that all the necessary shapes exist
	if (!shapeExists("RecipeClasses")) {
		fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
		return;
	}	
	//unRecipeFunctions_getRecipeClasses(dyn_string &dsRecipeClassNames, dyn_string &dsRecipeClassDps, 
  //      string sPcoLink="*", bool bLocalOnly=FALSE, bool bAddSystemName=FALSE, string sApplication="*"
  bool bLocalOnly=FALSE;
  bool bAddSystemName=FALSE;
	unRecipeFunctions_getRecipeClasses(recipeClassNames, recipeClassDps, sPcoLink, bLocalOnly, bAddSystemName, sApplication);
	if (dynlen(recipeClassNames)==0) {
		setValue("RecipeClasses", "appendItem", "-None-");
	} else {
		for (int i=1; i<=dynlen(recipeClassNames); i++) {
			setValue("RecipeClasses", "appendItem", recipeClassNames[i]);
		}
	}
} 

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the list of PCO aliases.
 * @param dsPcoNames - [OUT] List of PCO aliases.
 * @param dsPcoDps   - [OUT] List of PCO Dps.
 * @param bLocalOnly - [IN]  TRUE if only the devices from the local system must be returned, otherwise FALSE.
 */
void unRecipeFunctions_getPcoList(dyn_string & dsPcoNames, dyn_string &dsPcoDps, bool bLocalOnly=FALSE) 
{
  dyn_string dsSelectedDeviceTypes, dsRetAppl;
  dynClear(dsPcoNames);
  dynClear(dsPcoDps);
  
  dsSelectedDeviceTypes = makeDynString(UN_CONFIG_CPC_PROCESSCONTROLOBJECT_DPT_NAME);
  _unRecipeFunctions_getFilteredDevices(dsSelectedDeviceTypes, dsPcoDps, dsPcoNames, dsRetAppl, bLocalOnly);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the list of devices filtering the applications and the device types specified.
 * @param dsSelectedDevType     - [IN]  List of selected device types.
 * @param dsRetDpName 	        - [OUT] List of DP Names found.
 * @param dsRetAlias            - [OUT] List of devices aliases found.
 * @param dsRetAppl             - [OUT] List of applications of the found devices.
 * @param bLocalOnly            - [IN]  TRUE if only the devices from the local system must be returned, otherwise FALSE.
 * @param dsSelectedApplication - [IN]  List of applications where to look for the devices.
 */
public void _unRecipeFunctions_getFilteredDevices(dyn_string dsSelectedDevType, dyn_string &dsRetDpName, dyn_string &dsRetAlias, 
            dyn_string &dsRetAppl, bool bLocalOnly=FALSE, dyn_string dsSelectedApplication=makeDynString("*")) 
{
  dyn_string dsSystemRetDpName, dsSystemRetAlias, dsSystemRetAppl, dsSystemNames;
  if (bLocalOnly) {
    dsSystemNames = makeDynString(getSystemName());
  } else {
    dsSystemNames = mappingKeys(g_m_ddsTreeDeviceOverviewDeviceList);
  }
  
  for (int i=1; i<=dynlen(dsSystemNames); i++) {
    _unRecipeFunctions_getFilteredDevicesBySystemName(dsSystemNames[i], dsSelectedDevType, dsSystemRetDpName, dsSystemRetAlias, dsSystemRetAppl, dsSelectedApplication);
    dynAppend(dsRetDpName, dsSystemRetDpName);
    dynAppend(dsRetAlias, dsSystemRetAlias);
    dynAppend(dsRetAppl, dsSystemRetAppl);
  }
}
//------------------------------------------------------------------------------------------------------------------------
public void _unRecipeFunctions_getFilteredDevicesBySystemName(string sSystemName, dyn_string dsSelectedDevType, dyn_string &dsRetDpName, 
         dyn_string &dsRetAlias, dyn_string &dsRetAppl, dyn_string dsSelectedApplication=makeDynString("*")) {
  bit32 b32SelectState;
  dyn_string dsSelectedSystem, dsSelectedFE, dsSelectedAppl, dsSelectedSubsystem1, dsSelectedSubsystem2, dsAliasSelected ;
  dyn_string dsExcludedAliasSelected, dsDpName, dsAlias, dsFE, dsFEDpName, dsFEDPType, dsAppl, dsDevType, dsSubsystem1, dsSubsystem2;
  dyn_string dsRetFE, dsRetFEDpName, dsRetFEDPType, dsRetDevType, dsRetSubsystem1, dsRetSubsystem2;
  dyn_dyn_string dsSystemData;
  
  // Set the bits of the SelectState variable
  setBit(b32SelectState, FILTER_APPLICATION, TRUE);
  setBit(b32SelectState, FILTER_DEVICETYPE, TRUE);

  dsSelectedSystem = makeDynString("*");
  dsSelectedFE = makeDynString("*");
  dsSelectedSubsystem1 = makeDynString("*");
  dsSelectedSubsystem2 = makeDynString("*");
  dsAliasSelected = makeDynString("*");
  dsExcludedAliasSelected = makeDynString();
  
  // Get the system data  
  dsSystemData = g_m_ddsTreeDeviceOverviewDeviceList[sSystemName];
  if (!dynlen(dsSystemData)) {
    return;
  }
  dsDpName = dsSystemData[TREE_DEVICE_DPNAME];
  dsAlias = dsSystemData[TREE_DEVICE_ALIAS];
  dsFE = dsSystemData[TREE_DEVICE_FE];
  dsFEDpName = dsSystemData[TREE_DEVICE_FEDPNAME];
  dsFEDPType = dsSystemData[TREE_DEVICE_FEDPTYPE];
  dsDevType = dsSystemData[TREE_DEVICE_TYPE];
  dsSubsystem1 = dsSystemData[TREE_DEVICE_SUBSYSTEM1];
  dsSubsystem2 = dsSystemData[TREE_DEVICE_SUBSYSTEM2];
  dsAppl = dsSystemData[TREE_DEVICE_FEAPPLICATION];
  
  // Get the visible applications
  _unRecipeFunctions_getVisibleApplicationsList(dsSelectedAppl);
  
  CtrlTreeDeviceOverview_getSelectedDeviceList(
    b32SelectState, sSystemName, dsSelectedSystem, dsSelectedFE, dsSelectedAppl,
	dsSelectedDevType, dsSelectedSubsystem1, dsSelectedSubsystem2, dsAliasSelected, dsExcludedAliasSelected,
	dsDpName, dsAlias, dsFE, dsFEDpName, dsFEDPType, dsAppl, dsDevType, dsSubsystem1, dsSubsystem2,
	dsRetDpName, dsRetAlias, dsRetFE, dsRetFEDpName, dsRetFEDPType, dsRetAppl, dsRetDevType, dsRetSubsystem1, dsRetSubsystem2);
	  
  // Remove the DPNames and Alias that don't exist (the list may not be synchronized)
  int iLen = dynlen(dsRetDpName);
  for (int i=1; i<=iLen; i++) {
	  if (!dpExists(dsRetDpName[i]) || (!dynContains(dsSelectedApplication, dsRetAppl[i]) && !dynContains(dsSelectedApplication, "*"))) {
	    dynRemove(dsRetDpName, i);
	    dynRemove(dsRetAlias, i);
      dynRemove(dsRetAppl, i);
	    iLen--;
	    i--;
	  }
  }  
}
//------------------------------------------------------------------------------------------------------------------------
/**	 
 * Get the list of recipe class names and recipe class DPs linked to a PCO.
 * @param dsRecipeClassNames    - [OUT] List of recipe class names.
 * @param dsRecipeClassDps      - [OUT] List of recipe class DPs.
 * @param sPcoLink		          - [IN]  PCO alias which recipe classes are requested.
 * @param bLocalOnly            - [IN]  TRUE if only the recipe classes of the local system must be loaded, otherwise FALSE.
 * @param bAddSystemName        - [IN]  TRUE if the system name must be added to the recipe class name
 * @param bOnlyConnectedSystems - [IN]  TRUE if only the devices of the connected systems must be obtained.
 * @param sApplication          - [IN]  Name of the application for which the recipe classes are required.
 */
void unRecipeFunctions_getRecipeClasses(dyn_string &dsRecipeClassNames, dyn_string &dsRecipeClassDps, 
        string sPcoLink="*", bool bLocalOnly=FALSE, bool bAddSystemName=FALSE, string sApplication="*", bool bOnlyConnectedSystems=FALSE)
{
	int iLen;
	string pvssDpe, sRcpClassPcoLink, sSystemName;
	dyn_string dsSelectedDevType, dsAppl, exceptionInfo;
	dyn_dyn_anytype resultTable;
	
	dynClear(dsRecipeClassNames);
	dynClear(dsRecipeClassDps);
	
	// Get the list of recipe classes
	dsSelectedDevType = makeDynString(UN_CONFIG_UNRCPCLASS_DPT_NAME);	
	_unRecipeFunctions_getFilteredDevices(dsSelectedDevType, dsRecipeClassDps, dsRecipeClassNames, dsAppl, bLocalOnly);
				
	if (sPcoLink!="*" || sApplication!="*") {
	  // Removes all the recipe classes which are not linked with the selected PCO or belong to the specified application
	  iLen = dynlen(dsRecipeClassDps);
	  for (int i=1; i<=dynlen(dsRecipeClassDps); i++) {
	    if (!dpExists(dsRecipeClassDps[i] + ".statusInformation.deviceLink:_online.._value")) {
		    continue;
      }
	    dpGet(dsRecipeClassDps[i] + ".statusInformation.deviceLink:_online.._value", sRcpClassPcoLink);
      string sRcpApplication = unGenericDpFunctions_getApplication(dsRecipeClassDps[i]);
      bool bRemoveElement = (sPcoLink!="*" && sRcpClassPcoLink!=sPcoLink) || (sApplication!="*" && sApplication!=sRcpApplication);
      if (bRemoveElement) {
        dynRemove(dsRecipeClassDps, i);
        dynRemove(dsRecipeClassNames, i);
        iLen--;
        i--;
      }
	  }
  }
	
  if (bOnlyConnectedSystems) {
    iLen = dynlen(dsRecipeClassDps);
    for (int i=1; i<=iLen; i++) {
      sSystemName = unGenericDpFunctions_getSystemName(dsRecipeClassDps[i]);
      if(!unRecipeFunctions_isRemoteSystemConnected(sSystemName, exceptionInfo)) {
        dynRemove(dsRecipeClassDps, i);
        dynRemove(dsRecipeClassNames, i);
        iLen--;
        i--;
      }
    }
  }
  
  // Add the system name to the recipe class names (if specified)
  if (bAddSystemName) {
    iLen = dynlen(dsRecipeClassDps);
    for (int i=1; i<=iLen; i++) {
      sSystemName = unGenericDpFunctions_getSystemName(dsRecipeClassDps[i]);
      dsRecipeClassNames[i] = sSystemName + dsRecipeClassNames[i];
    }
  }
    
	// Sort alphabetically the recipe class names and the recipe class DP Names in a consistent way
	_unRecipeFunctions_sortAscDynLists(dsRecipeClassNames, dsRecipeClassDps, exceptionInfo);
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the list of visible applications.
 * @param dsApplications - [OUT] List of visible applications.
 */
void _unRecipeFunctions_getVisibleApplicationsList(dyn_string &dsApplications) synchronized(g_unicosHMI_dsApplicationList) { 
  dynClear(dsApplications);
  
  // Check if the application filter is being used
  if (false == g_unicosHMI_bApplicationFilterInUse) {
	dynAppend(dsApplications, "*");
	return;
  }
  
  fwGeneral_stringToDynString(g_unicosHMI_dsApplicationList, dsApplications);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the recipe instances in the table (recipe instances panel).
 * @param sRecipeClass - [IN] Recipe class name which instances will be loaded.
 */
synchronized void _unRecipeFunctions_loadRecipeInstances(string sRecipeClass="")
{
  dyn_string dsRcpNames, dsRcpDps, exceptionInfo;
  dyn_dyn_anytype resultTable;
    	
  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeInstances") || !shapeExists("RecipeClasses")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    return;
  }  	
	
  if (sRecipeClass=="") {
    // Get the name of the recipe class selected in the combo box
    getValue("RecipeClasses", "text", sRecipeClass); 
    if (sRecipeClass=="") {
      return;
    }
  }
  
  setValue("RecipeInstances", "deleteAllLines");   
  _unRecipeFunctions_getRecipeInstancesOfClass(sRecipeClass, dsRcpNames, dsRcpDps); 
  _unRecipeFunctions_sortAscDynLists(dsRcpNames, dsRcpDps, exceptionInfo);
  setValue("RecipeInstances", "appendLines", dynlen(dsRcpNames),
           "RecipeInstances", dsRcpNames, 
           "RecipeInstancesDP", dsRcpDps);
} 
 
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get all the recipe instances of a class.
 * @param sClassName 		    - [IN] Recipe class name.
 * @param dsRcpInstanceNames    - [OUT] Recipe instance names.
 * @param dsRcpInstanceDps      - [OUT] Recipe instance datapoint elements.
 */
void _unRecipeFunctions_getRecipeInstancesOfClass(string sClassName, dyn_string & dsRcpInstanceNames, 
			dyn_string &dsRcpInstanceDps) 
{
  int iLen;
  string sInstClassName, sInstName;
  dyn_dyn_anytype resultTable;
  dyn_string dsSelectedDevType, dsAppl;
  
  dsSelectedDevType = makeDynString(UN_CONFIG_UNRCPINSTANCE_DPT_NAME);
  _unRecipeFunctions_getFilteredDevices(dsSelectedDevType, dsRcpInstanceDps, dsRcpInstanceNames, dsAppl);
  
  if (sClassName=="*") {
    return;
  }
	
  // Check if the recipe instances belong to the specified recipe class
  iLen = dynlen(dsRcpInstanceDps);
  for (int i=1; i<=iLen; i++) {
    dpGet(dsRcpInstanceDps[i] + ".ProcessInput.ClassName", sInstClassName,
		  dsRcpInstanceDps[i] + ".ProcessInput.InstanceName", sInstName);
	
	if (sInstClassName!=sClassName) {
	  dynRemove(dsRcpInstanceDps, i);
	  dynRemove(dsRcpInstanceNames, i);
	  iLen--;
	  i--;
	  continue;
	} 
	
	dsRcpInstanceNames[i] = sInstName;
  }
}
//------------------------------------------------------------------------------------------------------------------------ 
/** 
 * Auxiliary function to order alphabetically a list (list1) and reorder the second list (list2) in a consistent way.
 * @param list1 - [IN/OUT] List to be ordered alphabetically.
 * @param list2 - [IN/OUT] List to be ordered in the same way as the list1.
 * @param exceptionInfo - [OUT] Standard error handling variable.
 */
void _unRecipeFunctions_sortAscDynLists(dyn_string &list1, dyn_anytype &list2, dyn_string &exceptionInfo) {
  int iLen, iPos;
  dyn_string cpyList1;
  dyn_anytype cpyList2;
  
  if (dynlen(list1) != dynlen(list2)) {
    fwException_raise(exceptionInfo, "ERROR", "The number of elements in the list must be equal","");
	return;
  }  
  
  cpyList1 = list1;    
  dynSortAsc(cpyList1);
  
  iLen = dynlen(cpyList1);
  for (int i=1; i<=iLen; i++) {
    // Check the position of the curren element in the original list1
	iPos = dynContains(list1, cpyList1[i]);
	
	// Place the element of the second list in the proper order
	cpyList2[i] = list2[iPos];
  }
  
  list1 = cpyList1;
  list2 = cpyList2;
}
//------------------------------------------------------------------------------------------------------------------------ 
/**
 * Load the data of the recipe instance properties table.
 * @param sDpName - [IN] Datapoint element of the recipe instance.
 * @param recipeObject - [IN] Recipe instance data from fwConfigDB.
 */
synchronized void _unRecipeFunctions_loadPropertiesTableInfo(string sDpName="", dyn_dyn_mixed recipeObject=makeDynMixed()) {
  int iStatus, iRows;
  bool bActive;
  string sClassName, sSystemName, sClassDp, sCurrentStatus;
  dyn_int lines;
  dyn_string message, exceptionInfo;
  dyn_mixed recipeClassObject;

  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeProperties")) {
	  fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	  return;
  }     

  // Get the selected recipe DP name
  if (sDpName=="") {
	unRecipeFunctions_getSelectedRecipeDp(sDpName, exceptionInfo);
	if (dynlen(exceptionInfo) || sDpName=="") { 
        DebugTN(exceptionInfo); 
        return; 
    }
  }

  // Loads the recipe object data
  if (dynlen(recipeObject)==0) {
    _unRecipeFunctions_getRecipeObjectFromDp(sDpName, recipeObject, exceptionInfo); 
    if (dynlen(exceptionInfo)) { 
      DebugTN(exceptionInfo); 
      return; 
    }
  }

  sSystemName = unGenericDpFunctions_getSystemName(sDpName);
  sClassName  = recipeObject[fwConfigurationDB_RO_META_CLASSNAME];
  if (strpos(sClassName, sSystemName) != 0) {  
    sClassName  = sSystemName + recipeObject[fwConfigurationDB_RO_META_CLASSNAME];
  }
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sClassName, sClassDp);
  // Loads the recipe class info
  unRecipeFunctions_getRecipeClassObject(sClassDp, recipeClassObject, exceptionInfo);
  if (dynlen(exceptionInfo))  { 
    DebugTN(exceptionInfo); 
    return; 
  }
  
  // The current activation status will be set to avoid text blinking when this function is called from _unRecipeFunctions_plcStatusCB()
  getValue("RecipeProperties", "lineCount", iRows);
  if (iRows >= 2) {
    getValue("RecipeProperties", "cellValueRC", 2, "Value", sCurrentStatus);
  }
  
  // Fill the data in the recipe properties table.
  setValue("RecipeProperties", "deleteAllLines");
  dyn_string names = makeDynString("Class:", "Last Activated:", "Status:", "Class Desc.:", "Rcp. Desc.:");
  dyn_string values= makeDynString(recipeClassObject[fwConfigurationDB_RCL_CLASSNAME], 
                                   recipeClassObject[fwConfigurationDB_RCL_LAST_ACTIVATED_RECIPE], 
                                   sCurrentStatus, 
                                   recipeClassObject[fwConfigurationDB_RCL_DESCRIPTION], 
								   recipeObject[fwConfigurationDB_RO_META_COMMENT]);
  setValue("RecipeProperties", "appendLines", 5,
           "Name", names,
           "Value", values);
  
  // Get the recipe instance status
  unRecipeFunctions_getRecipeStatus(recipeObject, 
                                    recipeClassObject[fwConfigurationDB_RCL_LAST_ACTIVATED_RECIPE],
                                    iStatus, 
                                    message, 
                                    exceptionInfo);
				
  unRecipeFunctions_setRecipeStatusInfo(iStatus, message);
} 
 
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Set the recipe status in the RecipeProperties table.
 * @param status      - [IN] New status of the recipe instance.
 * @param message     - [IN] Contains a message to display when there is no PLC connection.
 */
synchronized void unRecipeFunctions_setRecipeStatusInfo(int status, dyn_string message=makeDynString())
{
  int len;
  string sCurrentStatus, sNewStatus, sNewColor;
  dyn_string exceptionInfo;
  
  // Check that all the necessary shapes exist
  if (!shapeExists("StatusColor") || !shapeExists("buttonActivate") || !shapeExists("RecipeProperties")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    return;
  }     

  StatusColor.visible(true);
  getValue("RecipeProperties", "cellValueRC", 2, "Value", sCurrentStatus);
  
  switch(status){
    case RCP_NO_PLC_CONNECTION:
      sNewStatus = "     [No connection to PLCs: "+message+"]";
      sNewColor  = CELL_ERROR_COLOR;
      break;
    case RCP_BAD_STRUCTURE:
      sNewStatus = "     [Wrong recipe structure]";
      sNewColor  = CELL_ERROR_COLOR;
      break;     
    case RCP_BAD_TYPES:
      sNewStatus = "     [Wrong value types]";
      sNewColor  = CELL_ERROR_COLOR;
      break;       
    case RCP_BAD_DPES:
      sNewStatus = "     [Wrong recipe elements]";
      sNewColor  = CELL_ERROR_COLOR;
      break;
    case RCP_INACTIVE:
      sNewStatus = "     [Inactive]";
      sNewColor  = INACTIVE_RCP_COLOR;
      break;
    case RCP_ACTIVATED:
      sNewStatus = "     [Active]";
      sNewColor  = ACTIVE_RCP_COLOR;
      break;
    case RCP_ACTIVATION_FAILED:
      sNewStatus = "     [Activation failed]";
      sNewColor  = CELL_ERROR_COLOR;
      break;	  
    case RCP_ACTIVATION_TRIGGERED:
      sNewStatus = "     [Activating]";
      sNewColor  = INACTIVE_RCP_COLOR;	
      break;
  }

  if (sNewStatus == sCurrentStatus) {
    return;
  }
    
  StatusColor.backCol(sNewColor);
  setValue("RecipeProperties", "cellValueRC", 2, "Value", sNewStatus);
}
 
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Load the recipe instance data in the panel.
 */
synchronized void unRecipeFunctions_loadRecipeInstanceData() {
  string sDpName, sClassName, sInstanceName;
  dyn_int diIndex;
  dyn_string exceptionInfo, dsDeviceDps, dsDeviceAlias, dsDescriptions, dsUnits, dsRanges, dsErrorIndex;  
  dyn_dyn_mixed recipeObject;

  // Get the selected recipe DP name
  unRecipeFunctions_getSelectedRecipeDp(sDpName, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error loading the recipe: " + exceptionInfo[2]);
    DebugTN(exceptionInfo); 
    return;
  } else if (sDpName=="") {
    _unRecipeFunctions_writeInRecipeLog("Error loading the recipe: The DP Name is empty");
    return;
  }
  
  _unRecipeFunctions_getRecipeObjectFromDp(sDpName, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    _unRecipeFunctions_writeInRecipeLog("Error loading the recipe: " +exceptionInfo[2]);
    DebugTN(exceptionInfo); 
    return; 
  }
	  
  // Load the RecipeProperties table
  _unRecipeFunctions_loadPropertiesTableInfo(sDpName, recipeObject);
  
  // Load the History table
  _unRecipeFunctions_loadHistoryTableInfo(sDpName, recipeObject);

  // Inserts the recipe elements in the table
  int len = dynlen(recipeObject[fwConfigurationDB_RO_DPE_NAME]);
  setMultiValue("RecipeElements", "enabled", FALSE,
                "RecipeElements", "deleteAllLines");
  
  for (int i=1; i<=len; i++) {  
    dyn_string tmp;
    string dpName, alias, deviceType, description, dpe, unit="", range="";    

    dpName = recipeObject[fwConfigurationDB_RO_DPE_NAME][i];
    alias = recipeObject[fwConfigurationDB_RO_DP_NAME][i];
    tmp = strsplit(dpName, ".");
    dpe = tmp[dynlen(tmp)];
    
    if (dpExists(dpName)) {
        description = unGenericDpFunctions_getDescription(unGenericDpFunctions_getDpName(dpName)) + " : " + dpe;
        range = unRecipeFunctions_getDeviceRange(recipeObject[fwConfigurationDB_RO_DPE_NAME][i], exceptionInfo);
        unit = unRecipeFunctions_getDpeUnit(dpName, exceptionInfo);
    } else {
        dynAppend(dsErrorIndex, i);
        description = " : " + dpe;
        range = "";
        unit  = "";
    }
    
    dynAppend(diIndex, i);
    dynAppend(dsDeviceDps, dpName);
    dynAppend(dsDeviceAlias, alias);
    dynAppend(dsDescriptions, description);
    dynAppend(dsUnits, unit);
    dynAppend(dsRanges, range);
  }
  
  dyn_anytype formattedValues;
  _unRecipeFunctions_formatValues(dsDeviceDps, recipeObject[fwConfigurationDB_RO_VALUE], 
                                  formattedValues, exceptionInfo);
  
  if (len>0) {
	  setValue("RecipeElements", "appendLines", len,
				 "Index", diIndex,
				 "DeviceDp", dsDeviceDps,
				 "Alias", dsDeviceAlias,
				 "Description", dsDescriptions,
				 "Value",  formattedValues,
				 "Unit", dsUnits,
				 "Range", dsRanges);
                 
      _unRecipeFunctions_copyRecipeInstanceToHiddenTable();
  }
  
  string sInitial = recipeObject[fwConfigurationDB_RO_META_PREDEFINED];
  setValue("RecipeInitialLabel", "text", sInitial);
  _unRecipeFunctions_resetRecipeElementsBgColor();
  
  // Check if the recipe has invalid elements
  if (dynlen(dsErrorIndex)) {
    // Set the recipe instance in INVALID state
    dpSet(sDpName+".ProcessInput.State", UN_RCP_INSTANCE_STATE_INVALID);
    dpGet(sDpName+".ProcessInput.ClassName", sClassName,
          sDpName+".ProcessInput.InstanceName", sInstanceName);
    _unRecipeFunctions_setInvalidRows(dsErrorIndex);
    _unRecipeFunctions_writeInRecipeLog("Error: Some elements of the recipe instance '" + sInstanceName + "' do not exist." );
    _unRecipeFunctions_writeInRecipeLog("Please edit the recipe class '" + sClassName + "' to remove the missing recipe elements.");
  }
  
  setMultiValue("RecipeElements", "columnHeader", 0, "Index",
				"RecipeElements", "enabled", TRUE,
				"RecipeElements", "adjustColumn", 2);
				
  // Check the range of the recipe values
  _unRecipeFunctions_checkRecipeValues (dsDeviceDps, recipeObject[fwConfigurationDB_RO_VALUE], TRUE, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2], true, true );
  }
  _unRecipeFunctions_enableRecipeInstanceWarning(false);
}

//------------------------------------------------------------------------------------------------------------------------ 
 /** 
  * Load the recipe instance data in the history table.
  * @param sDpName - [IN] Datapoint element of the recipe instance.
  * @param recipeObject - [IN] Contents of the recipe instance from fwConfigDB.
  */
synchronized void _unRecipeFunctions_loadHistoryTableInfo(string sDpName="", dyn_dyn_mixed recipeObject=makeDynMixed())
{
  string recipeClass, recipeInstance, sUser, sTime;
  time t, t0;
  dyn_int lines;
  dyn_string exceptionInfo;

  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeInstances") || !shapeExists("History") || !shapeExists("RecipeClasses")) {
	  fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	  return;
  }     

  // Get the selected recipe DP name
  if (sDpName=="") {
	unRecipeFunctions_getSelectedRecipeDp(sDpName, exceptionInfo);
	if (dynlen(exceptionInfo) || sDpName=="") { 
      DebugTN(exceptionInfo); 
      return; 
    }
  }
  
  // Loads the recipe object data
  if (dynlen(recipeObject)==0) {
	_unRecipeFunctions_getRecipeObjectFromDp(sDpName, recipeObject, exceptionInfo); 
	if (dynlen(exceptionInfo)) { 
      DebugTN(exceptionInfo); 
      return; 
    }
  }
  
  unRecipeFunctions_loadRecipeMetadataFromDb(sDpName, sUser, sTime);
  
  setValue("History", "deleteAllLines");
  dyn_string col1 = makeDynString("Creator:", "Last modifier:", "Last activator:", "Last saved in DB by:");
  dyn_string col2 = makeDynString(recipeObject[fwConfigurationDB_RO_META_AUTHOR], 
                                  recipeObject[fwConfigurationDB_RO_META_LASTMODIFICATIONUSER], 
                                  recipeObject[fwConfigurationDB_RO_META_LASTACTIVATIONUSER],
                                  sUser);
  dyn_string col3 = makeDynString("Creation time:", "Last modification time:", "Last activation time:", "Last saved in DB time:");
  dyn_string col4 = makeDynString(recipeObject[fwConfigurationDB_RO_META_CREATIONTIME], 
                                  recipeObject[fwConfigurationDB_RO_META_LASTMODIFICATIONTIME], 
                                  recipeObject[fwConfigurationDB_RO_META_LASTACTIVATIONTIME],
                                  sTime);  
  setValue("History", "appendLines", 4,
           "#1", col1,
           "#2", col2,
           "#3", col3,
           "#4", col4);
}  

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Remove the recipe instance data from the panel.
 */
synchronized void _unRecipeFunctions_clearInstanceTables() {
  setMultiValue("RecipeProperties", "deleteAllLines",
				"History", "deleteAllLines",
				"LoadedRecipeDp", "text", "",
				"StatusColor", "visible", false,
				"RecipeClassLabel", "text", "",
				"RecipeInstanceLabel", "text", "",
				"RecipeInitialLabel", "text", ""
				);
  
  if (shapeExists("RecipeElements")) {
	setValue("RecipeElements", "deleteAllLines");				
  }
}

//--------------------------------------------------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Methods to edit a recipe instance
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Put the recipe instance in Edit Mode so the recipe values can be modified.
 * @param sRecipeDp 	- [IN] DataPoint name of the recipe instance.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @param row			- [IN] Optional variable. It indicates the row number that is being edited.
 * @param column		- [IN] Optional variable. It indicates the name of the column that is being edited.
 */
void unRecipeFunctions_editRecipeInstance(string sRecipeDp, dyn_string &exceptionInfo, int row=-1, string column="") {  
  int iDpCol, iValueCol;
  string sSelectedDpName, sRecipeState;
  dyn_anytype daDeviceDps, daValues;
  
  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeElements") || !shapeExists("RecipeClasses") || !shapeExists("RecipeInstances")) {
	fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	return;
  }
  
  unRecipeFunctions_getSelectedRecipeDp(sSelectedDpName, exceptionInfo);
  if (sSelectedDpName!=sRecipeDp) {
	  // If the instance selection fails, returns
	  if (!unRecipeFunctions_selectRecipeInstance(sRecipeDp, exceptionInfo)) {
		  return;
    }
  }
	
  // Recipe instance in EDIT mode
  dpSet(sRecipeDp+".ProcessInput.State", UN_RCP_INSTANCE_STATE_EDIT);
	  
  // Modifies the color of the recipe values
  _unRecipeFunctions_setTableColumnBGColor("RecipeElements", "Value", CELL_EDIT_COLOR);
  setMultiValue("PcoList", "enabled", false,
        "ApplicationList", "enabled", false,
				"RecipeClasses", "enabled", false,
				"RecipeInstances", "enabled", false,
				"RecipeElements", "namedColumnEditable", "Value", true);
  
  // Check the range of the recipe values
  getMultiValue("RecipeElements", "nameToColumn", "DeviceDp", iDpCol,
                "RecipeElements", "nameToColumn", "Value", iValueCol,
                "RecipeElements", "getColumnN", iDpCol, daDeviceDps,
                "RecipeElements", "getColumnN", iValueCol, daValues);
           
  _unRecipeFunctions_checkRecipeValues(daDeviceDps, daValues, TRUE, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2], true, true );
  }
  
  g_editingRowNumber = row;
  g_editingColName = column;
  _unRecipeFunctions_copyRecipeInstanceToHiddenTable();
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Copy the RecipeElements table (Recipe instance) to a hidden table.
 */
void _unRecipeFunctions_copyRecipeInstanceToHiddenTable() {
  int i, iCols;
  dyn_int diIndex;
  dyn_dyn_string dsData;
  
  getValue("RecipeElements", "columnCount", iCols);
  
  for (i=0; i<iCols; i++) {
    if (i==0) {
      getValue("RecipeElements", "getColumnN", i, diIndex);
    } else {
      getValue("RecipeElements", "getColumnN", i, dsData[i]);
    }
  }
  
  setValue("RecipeElementsCopy", "deleteAllLines");
  setValue("RecipeElementsCopy", "appendLines", dynlen(dsData[1]), 
		   "Index", diIndex,
		   "DeviceDp", dsData[1],
		   "Alias", dsData[2],
		   "Description", dsData[3],
		   "Value", dsData[4],
		   "Unit", dsData[5],
		   "Range", dsData[6]);

  setValue("RecipeElementsCopy", "sort", true, "Index");  
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Save the values of the recipe instance.
 * @param sRecipeDp - [IN] The DP of the recipe instance.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @param bShowConfirmationMessage - [IN] Flag to specify if the confirmation message must be displayed (TRUE by default)
 * @return TRUE if the recipe values were saved correctly, otherwise FALSE.
 */ 
bool unRecipeFunctions_saveRecipeInstance(string sRecipeDp, dyn_string &exceptionInfo, bool bShowConfirmationMessage=TRUE)
{
  int colCount, iValColIndex = -1, iDpColIndex = -1, status;
  bool bAreEqual, bRangesOk;
  string colName;
  dyn_float df;
  dyn_string ds, message, dsDpNames, dsIndex;
  dyn_anytype values;
  dyn_dyn_mixed recipeObject;
  
  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeElements") || !shapeExists("RecipeElementsCopy") || !shapeExists("RecipeClasses") || !shapeExists("RecipeInstances")) {
    fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
    DebugTN(exceptionInfo);
    return FALSE;
  } 
  
  // Copy the table data to a hidden table (to order by index)
  _unRecipeFunctions_copyRecipeInstanceToHiddenTable();
  
  // Get the column index for the values
  getMultiValue("RecipeElementsCopy", "nameToColumn", "Value", iValColIndex,
				"RecipeElementsCopy", "nameToColumn", "DeviceDp", iDpColIndex);  
  if (iValColIndex<0) {
    fwException_raise(exceptionInfo, "ERROR", "The 'RecipeElements' table doesn't have a column named 'Value'.","");
    return FALSE;
  }
  getMultiValue("RecipeElementsCopy", "getColumnN", iValColIndex, values,
				"RecipeElementsCopy", "getColumnN", iDpColIndex, dsDpNames);
  
  _unRecipeFunctions_getRecipeObjectFromDp(sRecipeDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    DebugTN(exceptionInfo); 
    return FALSE;
  }
  
  // Check if the DPEs of the recipe are coherent (comparing the ones on the table and the ones saved)
  bAreEqual = _unRecipeFunctions_areListsEqual(recipeObject[fwConfigurationDB_RO_DPE_NAME], dsDpNames);
  if (!bAreEqual) {
    _unRecipeFunctions_writeInRecipeLog("Error: The DP names in the table don't match the recipe DP names. Please reload the recipe.");
    return FALSE;
  }
  
  // Check the recipe values
  bRangesOk = unRecipeFunctions_checkRecipeValues(recipeObject, values, TRUE, exceptionInfo);
  if (dynlen(exceptionInfo) || !bRangesOk) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
	return FALSE;
  } 

  // Asks for user confirmation
  if (bShowConfirmationMessage) {
	ChildPanelOnCentralReturn("fwGeneral/fwOkCancel.pnl",
	 						  "Save recipe values",
							  makeDynString("$sTitle:Save recipe values ",
											  "$text: Do you want to save the recipe values?"  ),
							  df,
							  ds);
	  
	  if (df[1]<=0) return FALSE;
  }
  
  _unRecipeFunctions_saveRecipeValues(recipeObject, values, exceptionInfo); 
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
    return FALSE;
  }
  recipeObject[fwConfigurationDB_RO_VALUE] = values;

  getMultiValue("RecipeElements", "getColumnN", 1, dsDpNames,
				"RecipeElements", "getColumnN", 0, dsIndex);
  // sets the formatted values in the table
  for (int i=1; i<=dynlen(dsIndex); i++) {
    int index = dsIndex[i];
	setValue("RecipeElements", "cellValueRC", (i-1), "Value", values[index]);
  }
  _unRecipeFunctions_resetRecipeElementsBgColor();
  _unRecipeFunctions_loadHistoryTableInfo(sRecipeDp, recipeObject);
  _unRecipeFunctions_loadPropertiesTableInfo(sRecipeDp, recipeObject);
  
  setMultiValue("PcoList", "enabled", true,
        "ApplicationList", "enabled", true,
				"RecipeClasses", "enabled", true,
				"RecipeInstances", "enabled", true,
				"RecipeElements", "namedColumnEditable", "Value", false);
  
  g_editingRowNumber = -1;
  g_editingColName = "";
  _unRecipeFunctions_writeInRecipeLog("The recipe " + recipeObject[fwConfigurationDB_RO_META_ORIGNAME] + " has been saved locally.", TRUE);
  dpSet(sRecipeDp+".ProcessInput.State", "");
  _unRecipeFunctions_enableRecipeInstanceWarning(false);
  
  return TRUE;
}
//------------------------------------------------------------------------------------------------------------------------
bool _unRecipeFunctions_showValuesOutOfRange(dyn_int diErrorLines) {
  // check if the range of any value is wrong
  if (dynlen(diErrorLines)) {
    int iLen=dynlen(diErrorLines);
	
	dyn_string dsElementsIndex;
	getValue("RecipeElements", "getColumnN", 0, dsElementsIndex);
    for (int i=1; i<=iLen; i++) {
      int index = (int)diErrorLines[i];
	  int iPos = dynContains(dsElementsIndex, index);
	  if (iPos>0) {
		setValue("RecipeElements", "cellBackColRC", (iPos-1), "Value", CELL_VALUE_OUT_OF_RANGE_COLOR);
	  }
    }
    return TRUE;
  }
  
  return FALSE;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Save the recipe values as default values in the devices.
 * @param sRecipeDp - [IN] The DP of the recipe instance.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */ 
void unRecipeFunctions_saveAsDeviceDefaultValues(string sRecipeDp, dyn_string &exceptionInfo) {
  dyn_string dsDefaultValueDpes;
  dyn_dyn_mixed recipeObject;  

  // Get the recipe class of the selected recipe instance
  _unRecipeFunctions_getRecipeObjectFromDp(sRecipeDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    DebugTN(exceptionInfo); 
    return;
  }

  _unRecipeFunctions_getDefaultValueDpes(recipeObject[fwConfigurationDB_RO_DPE_NAME], dsDefaultValueDpes, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    DebugTN(exceptionInfo); 
    return;
  }
  
  if(dpSet(dsDefaultValueDpes, recipeObject[fwConfigurationDB_RO_VALUE]) < 0) {
    fwException_raise(exceptionInfo,"ERROR", "Error saving the default values of the recipe devices.","");
    DebugTN(exceptionInfo);
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Save the values of the selected recipe in the initial recipe of the class.
 * @param sRecipeDp - [IN] The DP of the recipe instance which values must be save in the initial recipe.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */ 
void unRecipeFunctions_saveAsInitialRecipe(string sRecipeDp, dyn_string &exceptionInfo)
{
  int iCol;
  bool bRangesOk;
  string sRcpClassName, sRcpClassDp, sRcpInstanceName, sInitialRcpInstanceDp, sSystemName;
  dyn_dyn_mixed recipeObject, initialRecipeObject;
  dyn_string recipeValues;
  dyn_int errorLines;

  // Get the recipe class of the selected recipe instance
  _unRecipeFunctions_getRecipeObjectFromDp(sRecipeDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    DebugTN(exceptionInfo); 
    return;
  }
  
  sSystemName   = unGenericDpFunctions_getSystemName(sRecipeDp);
  sRcpClassName = recipeObject[fwConfigurationDB_RO_META_CLASSNAME];
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sSystemName + sRcpClassName, sRcpClassDp);
  
  // Get the initial recipe of the recipe class
  _unRecipeFunctions_getInitialRecipeOfClass(sRcpClassDp, sInitialRcpInstanceDp, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
    DebugTN(exceptionInfo);
    return;
  }
  
  // Get the values of the selected recipe
  _unRecipeFunctions_copyRecipeInstanceToHiddenTable();
  getMultiValue("RecipeElementsCopy", "nameToColumn", "Value", iCol,
    "RecipeElementsCopy", "getColumnN", iCol, recipeValues);
  
  // Get the initial recipe object
  _unRecipeFunctions_getRecipeObjectFromDp(sInitialRcpInstanceDp, initialRecipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error getting the data of the recipe instance with DP: " + sRecipeDp);
    DebugTN(exceptionInfo); 
    return;
  }
  
  // Check if the recipe values are in the valid range
  bRangesOk = unRecipeFunctions_checkRecipeValues (initialRecipeObject, recipeValues, FALSE, exceptionInfo);
  if (dynlen(exceptionInfo) || !bRangesOk) {
    _unRecipeFunctions_writeInRecipeLog("Error: The recipe values can't be saved. Some recipe elements are out of range.");
    return;
  } 
  
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
    return;
  }
  
  // Save the new values of the initial recipe
  dpGet(sInitialRcpInstanceDp + ".ProcessInput.InstanceName", sRcpInstanceName);
  initialRecipeObject[fwConfigurationDB_RO_VALUE] = recipeValues;  
  
  fwConfigurationDB_saveRecipeToCache(initialRecipeObject, "", sSystemName + sRcpClassName + "/" + sRcpInstanceName, exceptionInfo);
  
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
    return;
  }
  
  _unRecipeFunctions_writeInRecipeLog("The values of the recipe " + recipeObject[fwConfigurationDB_RO_META_ORIGNAME] + 
				" have been saved in the initial recipe: " + sRcpClassName + "/" + sRcpInstanceName);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the DP name of the initial recipe of the class 'sClassName'.
 * @param sClassDp      - [IN]  Datapoint name of the recipe class which initial recipe is required.
 * @param sInitialRcpDp - [OUT] DP Name of the initial recipe of the specified class.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return 0 if the initial recipe was found, otherwise returns a negative value
 */
int _unRecipeFunctions_getInitialRecipeOfClass(string sClassDp, string & sInitialRcpDp, dyn_string & exceptionInfo) {
  
  int iRetVal, iPos;
  string sRcpInstanceName, sClassName, sSystemName;
  dyn_dyn_anytype resultTable;
  
  _unRecipeFunctions_normalizeDp(sClassDp);
  dpGet(sClassDp + ".ProcessInput.ClassName", sClassName);
  sSystemName = unGenericDpFunctions_getSystemName(sClassDp);
    
  string sQuery = "SELECT '.RecipeName:_online.._value' FROM '*' REMOTE '" + sSystemName + "' WHERE "
                        + "('.RecipeName:_online.._value' LIKE \"" + sClassName + "/*\" OR '.RecipeName:_online.._value' LIKE \"*:" + sClassName + "/*\")"
                        + "AND '.MetaInfo.Predefined:_online.._value' = \"TRUE\" AND _DPT=\"_FwRecipeCache\" ";
                        
  // Get the initial recipe of the recipe class 'sClassName'
  iRetVal = dpQuery(sQuery, resultTable);   
 
  // Check if there isn't an initial recipe for the recipe class 
  if (dynlen(resultTable)<=1) {
    fwException_raise(exceptionInfo,"ERROR", "The recipe class '" + sClassName + "' doesn't have an initial recipe. ","");
	return RCP_INITIAL_NO_DEFINED;	   
  }

  // Check if there is more than one initial recipe instance
  if (dynlen(resultTable)>2) {
    fwException_raise(exceptionInfo, "ERROR", "The recipe class '" + sClassName + "' has more than one initial recipe instance. ","");
	return RCP_INITIAL_TOO_MANY;	   
  }
  
  // Get the recipe instance name (Recipe name: rcpClassName/rcpInstanceName)
  iPos = strpos(resultTable[2][2], "/");
  if (iPos<0) {
	fwException_raise(exceptionInfo, "ERROR", "The recipe name is wrong ('/' character is missing).");
	return RCP_INITIAL_UNDEFINED;
  }
  
  // Get the UNICOS recipe DP name
  sRcpInstanceName = unExportTree_removeSystemName(sRcpInstanceName);
  sRcpInstanceName = substr(resultTable[2][2], iPos+1);
  iRetVal = dpQuery("SELECT '.ProcessInput.ClassName:_online.._value' FROM '*' REMOTE '" + sSystemName + "' WHERE '.ProcessInput.ClassName:_online.._value' = \"" 
                    + sClassName + "\" AND '.ProcessInput.InstanceName:_online.._value' = \"" + sRcpInstanceName 
                    + "\" AND _DPT=\"" + UN_CONFIG_UNRCPINSTANCE_DPT_NAME + "\" ", resultTable);   
  
  if (dynlen(resultTable)!=2) {
    fwException_raise(exceptionInfo, "ERROR", "The recipe instance '" + sClassName + "/" + sRcpInstanceName + "' doesn't exist. ","");
	return RCP_INITIAL_UNDEFINED;	 
  }
  
  // Get the PVSS DP name of the recipe instance
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(resultTable[2][1], sInitialRcpDp);
  
  return RCP_INITIAL_OK;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get all the PCOs that are children of the sPco device (recursive method, get all the PCOs in the hierarchy).
 * @param sPcoDpName - [IN] Name of the PCO which children are requested.
 * @param children - [OUT] Contains the alias of all the PCO children of the sPcoDp device.
 * @param exceptionInfo - [OUT] Standard exception handling variable. 
 */
void _unRecipeFunctions_getPcoChildren(string sPcoDpName, dyn_string &children, dyn_string & exceptionInfo) {
  string sCurrentPcoAlias;
  dyn_string dsAllChildren;

  _unRecipeFunctions_normalizeDp(sPcoDpName);
  
  dsAllChildren = cpcGenericObject_getLinkedObjects(sPcoDpName, UN_CONFIG_CPC_PROCESSCONTROLOBJECT_DPT_NAME, exceptionInfo);
  
  for (int i = 1; i <= dynlen(dsAllChildren); i++) {
    sCurrentPcoAlias = dsAllChildren[i];
    dynAppend(children, sCurrentPcoAlias);
	_unRecipeFunctions_getPcoChildren(sCurrentPcoAlias, children, exceptionInfo);    
  }
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Open a new window to display the initial recipes of a PCO.
 * @param sPcoDp - [IN] DataPoint name of the PCO which initial recipes will be displayed.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_activateInitialRecipesFromPcoHierarchy(string sPcoDp, dyn_string & exceptionInfo) {
  dyn_string ds;
  dyn_float df;
  
  ChildPanelOnCentralReturn("vision/unRecipe/UnRcpInstance/unFaceplateUnRcpInstance.pnl",
			"Create Recipe Instance",
			makeDynString("$sAction:" + UN_RCP_ACTION_ACTIVATE_INITIAL_RECIPES, "$sDpName:" + sPcoDp),
			df,ds);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the DP name of the last activated recipe of the class 'sClassName'.
 * @param sClassDp - [IN] Datapoint name of the recipe class which last activated recipe is required.
 * @param sLastActivatedRcpDp - [OUT] DP Name of the last activated recipe of the specified class.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return 0 if the last activated recipe was found, otherwise returns a negative value
 */
int _unRecipeFunctions_getLastActivatedRecipeDp(string sClassDp, string & sLastActivatedRcpDp, dyn_string & exceptionInfo) {
  int iRetVal, iPos;
  string sRcpInstanceName, sClassName, sSystemName;
  dyn_dyn_anytype resultTable;
  time t0;
  
  sSystemName = unGenericDpFunctions_getSystemName(sClassDp);
  _unRecipeFunctions_normalizeDp(sClassDp);
  dpGet(sClassDp + ".ProcessInput.ClassName", sClassName);
  
  // Get the last activated recipe of the recipe class 'sClassName'
  iRetVal = dpQuery("SELECT '.RecipeName:_online.._value', '.MetaInfo.LastActivationTime:_online.._value' FROM '*' REMOTE '" + sSystemName + "' WHERE "
                + "('.RecipeName:_online.._value' LIKE \"" + sClassName + "/*\" OR '.RecipeName:_online.._value' LIKE \"*:" + sClassName + "/*\") "
                + "AND _DPT=\"_FwRecipeCache\" SORT BY 2 DESC", resultTable);   
 
  // Check if there aren't recipes matching the criteria
  if (dynlen(resultTable)<=1) {
    fwException_raise(exceptionInfo,"ERROR", "The recipe class '" + sClassName + "' doesn't have any recipe instance. ","");
	return -1;	   
  }
  
  // Check if the last activation time is the not valid (1970.01.01 01:00:00.000000000) 
  if (resultTable[2][3] == t0) {
    fwException_raise(exceptionInfo,"ERROR", "The recipe class '" + sClassName + "' doesn't have a last activated instance. ","");
    return -1;
  }
  
  // Get the recipe instance name (Recipe name: rcpClassName/rcpInstanceName)
  iPos = strpos(resultTable[2][2], "/");
  if (iPos<0) {
	fwException_raise(exceptionInfo, "ERROR", "The recipe name is wrong ('/' character is missing).");
	return -1;
  }
  
  // Get the UNICOS recipe DP name
  sRcpInstanceName = unExportTree_removeSystemName(sRcpInstanceName);
  sRcpInstanceName = substr(resultTable[2][2], iPos+1);
  iRetVal = dpQuery("SELECT '.ProcessInput.ClassName:_online.._value' FROM '*' REMOTE '" + sSystemName + "' WHERE '.ProcessInput.ClassName:_online.._value' = \"" 
                + sClassName + "\" AND '.ProcessInput.InstanceName:_online.._value' = \"" + sRcpInstanceName 
                + "\"  AND _DPT=\"" + UN_CONFIG_UNRCPINSTANCE_DPT_NAME + "\" ", resultTable);   
  
  if (dynlen(resultTable)!=2) {
    fwException_raise(exceptionInfo, "ERROR", "The recipe instance '" + sClassName + "/" + sRcpInstanceName + "' doesn't exist. ","");
	return -1;	 
  }
  
  // Get the PVSS DP name of the recipe instance
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(resultTable[2][1], sLastActivatedRcpDp);
  
  return 0;
}

//------------------------------------------------------------------------------------------------------------------------
/**
 * Open a new window to display the last activated recipes of a PCO.
 * @param sPcoDp - [IN] DataPoint name of the PCO which initial recipes will be displayed.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_activateLastActivatedRecipesFromPcoHierarchy(string sPcoDp, dyn_string & exceptionInfo) {
  dyn_string ds;
  dyn_float df;
  
  ChildPanelOnCentralReturn("vision/unRecipe/UnRcpInstance/unFaceplateUnRcpInstance.pnl",
			"Create Recipe Instance",
			makeDynString("$sAction:" + UN_RCP_ACTION_ACTIVATE_LAST_ACTIVATED_RECIPES, "$sDpName:" + sPcoDp),
			df,ds);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the last activated recipe of each recipe class for the current PCO.
 * @param sPcoDp - [IN] The PCO datapoint element which last activated recipes are required.
 * @param recipeList - [OUT] Datapoint elements of the last activated recipes of the PCO.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getLastActivatedPcoRecipes (string sPcoDp, dyn_string &recipeList, dyn_string &exceptionInfo) {
  int iRetVal, iLen, i;
  string sPcoAlias, sClassDp, sLastActivatedRcpDp;
  dyn_string dsRcpClasses, dsRcpClassDps;
  dyn_dyn_anytype resultTable;
  
  _unRecipeFunctions_normalizeDp(sPcoDp);
  sPcoDp = sPcoDp + ".";  
  
  // Check if the PCO DP exists
  if (!dpExists(sPcoDp)) {
	  fwException_raise(exceptionInfo, "ERROR", "The PCO DP '" + sPcoDp + "' doesn't exist", "");
	  return;
  }
  
  // Get the PCO UNICOS alias
  sPcoAlias = unGenericDpFunctions_getAlias(sPcoDp);
  
  // Get all the recipe classes linked to the PCO
  unRecipeFunctions_getRecipeClasses(dsRcpClasses, dsRcpClassDps, sPcoAlias);
  
  if (!dynlen(dsRcpClasses)) {
    // The selected PCO doesn't have any recipe class linked, return.
	  return;
  }
  
  iLen = dynlen(dsRcpClassDps);
  for (i=1; i<=iLen; i++) {
	  // Get the last activated recipe of the recipe class
	  sClassDp = dsRcpClassDps[i];	
	  iRetVal = _unRecipeFunctions_getLastActivatedRecipeDp(sClassDp, sLastActivatedRcpDp, exceptionInfo);	
	  if (iRetVal < 0) {
		  continue;
    }
	
	  dynAppend(recipeList, sLastActivatedRcpDp);
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the initial recipe of each recipe class for the current PCO.
 * @param sPcoDp - [IN] The PCO datapoint element which initial recipes are required.
 * @param recipeList - [OUT] Datapoint elements of the initial recipes of the PCO.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getInitialPcoRecipes (string sPcoDp, dyn_string &recipeList, dyn_string &exceptionInfo) {
  int iRetVal, iLen, i;
  string sPcoAlias, sClassDp, sInitialRcpDp;
  dyn_string dsRcpClasses, dsRcpClassDps;
  dyn_dyn_anytype resultTable;
  
  // Check if the PCO DP exists
  if (!dpExists(sPcoDp)) {
	  fwException_raise(exceptionInfo, "ERROR", "The PCO DP '" + sPcoDp + "' doesn't exist", "");
	  return;
  }
  
  // Get the PCO UNICOS alias
  sPcoAlias = unGenericDpFunctions_getAlias(sPcoDp);
  
  // Get all the recipe classes linked to the PCO
  unRecipeFunctions_getRecipeClasses(dsRcpClasses, dsRcpClassDps, sPcoAlias);
  if (!dynlen(dsRcpClasses)) {
    // The selected PCO doesn't have any recipe class linked, return.
	  return;
  }  
  
  iLen = dynlen(dsRcpClassDps);
  for (i=1; i<=iLen; i++) {
  	// Get the initial recipe of the recipe class
	  sClassDp = dsRcpClassDps[i];	
  	iRetVal = _unRecipeFunctions_getInitialRecipeOfClass(sClassDp, sInitialRcpDp, exceptionInfo);
	  if (iRetVal < 0) {
		  continue;
    }
	
	  dynAppend(recipeList, sInitialRcpDp);
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the last activated recipe of each recipe class for the PCO hierarchy.
 * @param sPcoDp - [IN] The PCO datapoint element which last activated recipes are required.
 * @param recipeList - [OUT] Datapoint elements of the last activated recipes of the PCO hierarchy.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_getLastActivatedRecipesFromPcoHierarchy(string sPcoDp, dyn_string &recipeList, dyn_string &exceptionInfo) {
  int i, iLen;
  dyn_string allPcos;

  dynAppend(allPcos, sPcoDp);
  
  // Get all the PCO that are children of the sPcoDp device
  _unRecipeFunctions_getPcoChildren(sPcoDp, allPcos, exceptionInfo);
  
  iLen = dynlen(allPcos);
  for (i=1; i<=iLen; i++) {
    string sCurrentPco = allPcos[i];
    _unRecipeFunctions_getLastActivatedPcoRecipes (sCurrentPco, recipeList, exceptionInfo);
  } 
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the initial recipe of each recipe class for the PCO hierarchy.
 * @param sPcoDp - [IN] The PCO datapoint element which initial recipes are required.
 * @param recipeList - [OUT] Datapoint elements of the initial recipes of the PCO hierarchy.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_getInitialRecipesFromPcoHierarchy(string sPcoDp, dyn_string &recipeList, dyn_string &exceptionInfo) {
  int i, iLen;
  dyn_string allPcos;
  
  dynAppend(allPcos, sPcoDp);
  
  // Get all the PCO that are children of the sPcoDp device
  _unRecipeFunctions_getPcoChildren(sPcoDp, allPcos, exceptionInfo);
  iLen = dynlen(allPcos);
  for (i=1; i<=iLen; i++) {
    string sCurrentPco = allPcos[i];
    _unRecipeFunctions_getInitialPcoRecipes (sCurrentPco, recipeList, exceptionInfo);
  } 
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Write a message in the recipe log.
 * @param sMessage 			 - [IN] The message to write in the recipe log.
 * @param bWriteInGeneralLog - [IN] Flag to specify if the message must be written in the UNICOS HMI log (false by default).
 * @param bWriteInLogViewer  - [IN] Flag to specify if the message must be written in the WinCC OA log viewer (false by default).
 */
void _unRecipeFunctions_writeInRecipeLog(string sMessage, bool bWriteInGeneralLog=false, bool bWriteInLogViewer=false) {
  int iElements;
  dyn_string exceptionInfo;
 
  // Check that all the necessary shapes exist
  if (shapeExists("RecipeLog")) {
	string sTime = formatTime("[%Y.%m.%d %H:%M:%S]", getCurrentTime());
	RecipeLog.appendItem(sTime + " " + sMessage);
	getValue("RecipeLog", "itemCount", iElements);
	setValue("RecipeLog", "bottomPos", iElements);
  } 

  if (bWriteInGeneralLog) {
    unSendMessage_toAllUsers(sMessage, exceptionInfo);
  }
  
  if (bWriteInLogViewer) {
    DebugTN(sMessage);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Cancel the recipe instance edition.
 * @param sRecipeDp - [IN] DP Name of the recipe instance that was being edited.
 */
void unRecipeFunctions_cancelEditRecipeInstance(string sRecipeDp) { 
  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeElements") || !shapeExists("RecipeClasses") || !shapeExists("RecipeInstances") ) {
	fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	return;
  } 
  
  dpSet(sRecipeDp+".ProcessInput.State", "");
  
  setMultiValue("PcoList", "enabled", true,
        "ApplicationList", "enabled", true,
				"RecipeClasses", "enabled", true,
				"RecipeInstances", "enabled", true,
				"RecipeElements", "namedColumnEditable", "Value", false);
  _unRecipeFunctions_resetRecipeElementsBgColor(); 
  unRecipeFunctions_loadRecipeInstanceData();
  
  g_editingRowNumber = -1;
  g_editingColName = "";
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the recipe instance object.
 * @param sRecipeDp     - [IN] Datapoint name of the recipe instance.
 * @param recipeObject  - [OUT] Recipe instance object from fwConfigDB.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_getRecipeObjectFromDp (string sRecipeDp, dyn_dyn_mixed &recipeObject, dyn_string &exceptionInfo)
{
  int pos;
  string recipeClass, recipeInstance, sRcpClassDp, sSystemName;
  dyn_string dsDeviceOrder;
  
  if ((pos=strpos(sRecipeDp, "."))>=0) {
	sRecipeDp = substr(sRecipeDp, 0, pos);
  }
  
  if (!dpExists(sRecipeDp+".ProcessInput.ClassName") || 
      !dpExists(sRecipeDp+".ProcessInput.InstanceName"))
  {
	fwException_raise(exceptionInfo, "ERROR", "The recipe DP '" + sRecipeDp + "' doesn't exist.","");
	return;
  }  
  
  dpGet(sRecipeDp+".ProcessInput.ClassName", recipeClass,
		sRecipeDp+".ProcessInput.InstanceName", recipeInstance);
  
  sSystemName = unGenericDpFunctions_getSystemName(sRecipeDp);
  fwConfigurationDB_loadRecipeFromCache(sSystemName + recipeClass+"/"+recipeInstance, makeDynString(), "", recipeObject, exceptionInfo);
  
  // Order the devices
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sSystemName + recipeClass, sRcpClassDp);
  dpGet(sRcpClassDp + "ProcessInput.DeviceOrder", dsDeviceOrder);
  _unRecipeFunctions_orderRecipeInstanceElements(dsDeviceOrder, recipeObject);
}
//------------------------------------------------------------------------------------------------------------------------
void _unRecipeFunctions_orderRecipeInstanceElements(dyn_string dsDeviceOrder, dyn_dyn_mixed &recipeObject) {
  dyn_mixed dmDpeName, dmDpName, dmDpType, dmElementName, dmElementType, dmValues;

  if (!dynlen(dsDeviceOrder)) {
    return;
  }

  // Check if the DeviceOrder and dsDeviceOrder lists contain the same elements
  dyn_mixed dmOriginalDevices = recipeObject[fwConfigurationDB_RO_DP_NAME];
  dynUnique(dmOriginalDevices);  
  if (!_unRecipeFunctions_areListsEqual(dmOriginalDevices, dsDeviceOrder)) {
    DebugTN("Error: The device lists are not synchronized. The recipe elements can not be ordered.");
    DebugTN("OriginalDevices: ", dmOriginalDevices, "Device Order: ", dsDeviceOrder);
    DebugTN(getStackTrace());
    return;
  }
  
  for (int i=1; i<=dynlen(dsDeviceOrder); i++) {
    string sDevice = dsDeviceOrder[i];
    int j = dynContains(recipeObject[fwConfigurationDB_RO_DP_NAME], sDevice);
    if (j<=0) {
      // Error, the device lists are not synchronized
      DebugTN("Error: The device lists are not synchronized. Device " + sDevice + " does not exist in the recipeObject.");      
      return;
    }
    
    do {      
      dynAppend(dmDpeName, recipeObject[fwConfigurationDB_RO_DPE_NAME][j]);
      dynAppend(dmDpName,  recipeObject[fwConfigurationDB_RO_DP_NAME][j]);
      dynAppend(dmDpType,  recipeObject[fwConfigurationDB_RO_DP_TYPE][j]);
      dynAppend(dmElementName, recipeObject[fwConfigurationDB_RO_ELEMENT_NAME][j]);
      dynAppend(dmElementType, recipeObject[fwConfigurationDB_RO_ELEMENT_TYPE][j]);
      dynAppend(dmValues,  recipeObject[fwConfigurationDB_RO_VALUE][j]);
      j++;
    } while (j<=dynlen(recipeObject[fwConfigurationDB_RO_DP_NAME]) && recipeObject[fwConfigurationDB_RO_DP_NAME][j] == sDevice);
  }

  recipeObject[fwConfigurationDB_RO_DPE_NAME]     = dmDpeName;
  recipeObject[fwConfigurationDB_RO_DP_NAME]      = dmDpName;
  recipeObject[fwConfigurationDB_RO_DP_TYPE]      = dmDpType;
  recipeObject[fwConfigurationDB_RO_ELEMENT_NAME] = dmElementName;
  recipeObject[fwConfigurationDB_RO_ELEMENT_TYPE] = dmElementType;
  recipeObject[fwConfigurationDB_RO_VALUE]        = dmValues;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Get the recipe class object from the JCOP library.
 * @param sRcpClassDp	  - [IN] Recipe class name.
 * @param recipeClassInfo - [OUT] Recipe class object returned from the JCOP library.
 * @param exceptionInfo   - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_getRecipeClassObject(string sRcpClassDp, dyn_mixed &recipeClassInfo, dyn_string &exceptionInfo) {
  string sRcpClassName, sSystemName;
  _unRecipeFunctions_normalizeDp(sRcpClassDp);
  sSystemName = unGenericDpFunctions_getSystemName(sRcpClassDp);
  dpGet(sRcpClassDp + ".ProcessInput.ClassName", sRcpClassName);
  fwConfigurationDB_getRecipeClassInfo(sSystemName + sRcpClassName, recipeClassInfo, exceptionInfo);
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Add new devices to the recipe class that is being displayed in the recipe class panel.
 * @param sRcpClassDp - [IN] Datapoint element of the recipe class that will be modified.
 */
void unRecipeFunctions_addRecipeClassDevices(string sRcpClassDp) {
  dyn_string exceptionInfo;
  dyn_string types;
  dyn_dyn_string fwDevices;
  int i,n;
  string sPreviousState;
  dyn_string deviceList, dsAppl, existingDevices;
  dyn_dyn_string rcpTypeData;
  dyn_float df;
  dyn_string ds;
  mapping mRcpTypeMap, mRcpTypeMapString;
  

  _fwConfigurationDB_getRecipeTypeData(UN_CONFIG_UNRCPTYPE_DPT_NAME, rcpTypeData, exceptionInfo);  
  if (dynlen(exceptionInfo)) {
	fwExceptionHandling_display(exceptionInfo);
    DebugTN(exceptionInfo);
    return;
  }
  
  _unRecipeFunctions_getTypeMapping(UN_CONFIG_UNRCPTYPE_DPT_NAME, mRcpTypeMap, exceptionInfo);
  dyn_string mappingKeys = mappingKeys(mRcpTypeMap);
  for (i=1; i<=dynlen(mappingKeys); i++) {
    string sCurrentType = mappingKeys[i];
    dyn_string dsTypeDpes = mRcpTypeMap[sCurrentType];
    dynSortAsc(dsTypeDpes);	
    string sDeviceTypeDpes = "";
    fwGeneral_dynStringToString(dsTypeDpes, sDeviceTypeDpes, ", ");
    mRcpTypeMapString[sCurrentType] = sDeviceTypeDpes;
  }

  dyn_dyn_string allfwDeviceTypeInfo;
  dyn_string allfwDpTypes,allfwDevTypes;
  fwDevice_getAllTypes(allfwDeviceTypeInfo,exceptionInfo);
  if (dynlen(exceptionInfo)) {
    fwExceptionHandling_display(exceptionInfo);
    return;
  }
  allfwDevTypes=allfwDeviceTypeInfo[1];
  allfwDpTypes=allfwDeviceTypeInfo[2];
  
  bool allowDevsWithNoAlias=true;
  if (isDollarDefined("$allowDevsWithNoAlias")) allowDevsWithNoAlias=$allowDevsWithNoAlias;
  
  dyn_string allDevDps,allDevAliases, rcpDeviceTypes;

  for (int i=1; i<=dynlen(rcpTypeData); i++) {
    string sCurrTypeName = rcpTypeData[i][2];
    dynAppend(rcpDeviceTypes, sCurrTypeName);
  }
  _unRecipeFunctions_getFilteredDevices(rcpDeviceTypes, allDevDps, deviceList, dsAppl, TRUE);
  
  // Get the current recipe state
  dpGet(sRcpClassDp + ".ProcessInput.State", sPreviousState);
  if (sPreviousState == "") {
    // Set the recipe State to Edit mode
    dpSet(sRcpClassDp + ".ProcessInput.State", UN_RCP_CLASS_STATE_EDIT);
  }
  
  // Get the list of existing devices from the RecipeElements table
  getValue("RCCElements.RecipeElementsTable", "getColumnN", 2, existingDevices);
  
  // Create a String with the allowed device types for the recipe type
  string sRcpTypeList;
  dynSortAsc(mappingKeys);
  fwGeneral_dynStringToString(mappingKeys, sRcpTypeList, ",");
  
  ChildPanelOnCentralModalReturn("vision/unRecipe/UnRcpClass/unUnRcpClass_SelectionList.pnl",
		"Select Devices",
		makeDynString("$sDpName:" + sRcpClassDp,
					  "$text:Select devices:",
					  "$items:"+deviceList,
                      "$itemDps:"+allDevDps,
                      "$appl:"+dsAppl,
					  "$types:"+sRcpTypeList,
        			  "$AliasSelector:FALSE",
        			  "$DPSelector:FALSE",
					  "$selectMultiple:"+TRUE,
					  "$disabledItems:"+existingDevices),
		df,ds);
		
  // Insert the selected elements in the table
  for (int i=1; i<=dynlen(ds); i++) {
    string dpName,devName,alias;
    // firstly check if this is h/w or logical device
    if (strpos(ds[i],":")>0){
      alias="";
      dpName=ds[i];
    } else {
      alias=ds[i];
	  int idx=dynContains(deviceList,alias);
      if (idx<1) {
        continue;
      }
      dpName=dpSubStr(allDevDps[idx],DPSUB_SYS_DP);
    }
    if (!dpExists(dpName)) {
      continue;
    }

    string deviceDpType = dpTypeName(dpName);
    string deviceType=deviceDpType;
    dyn_string deviceElements = mRcpTypeMapString[deviceType];
    setValue("RCCElements.RecipeElementsTable", "appendLine",
             "DeviceDpType",deviceDpType,
             "DeviceType", deviceType,
             "DeviceName", ds[i],
			 "DeviceElements", deviceElements);             
  }
  
  if (dynlen(ds)) {
    _unRecipeFunctions_writeInRecipeLog("The new devices have been added. Please apply the changes to the recipe class.");
    setValue("applyChangesFrame", "visible", true);
  } else if (sPreviousState=="") {
    dpSet(sRcpClassDp + ".ProcessInput.State", "");
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Apply the modifications done in the recipe class elements.
 * @param sRcpClassDp - [IN] DataPoint name of the recipe class where the changes must be applied.
 */
void unRecipeFunctions_applyRecipeClassModifications(string sRcpClassDp) {
  int i, j, iCol;
  string sClassName, sTypeName, sDesc/*, sSystemName*/;
  dyn_string deviceAliases, deviceDpes, rcpDpes, exceptionInfo;
  dyn_mixed dmRcpClassInfo;

  dpSet(sRcpClassDp + ".ProcessInput.State", UN_RCP_CLASS_STATE_APPLY_CHANGES);
  // Get the data from the recipe class panel
  getMultiValue("RecipeClass", "text", sClassName,
				"RecipeType", "text", sTypeName,
				"RecipeClassComment", "text", sDesc,
				"RCCElements.RecipeElementsTable", "nameToColumn", "DeviceName", iCol,
				"RCCElements.RecipeElementsTable", "getColumnN", iCol, deviceAliases,
				"RCCElements.RecipeElementsTable", "nameToColumn", "DeviceElements", iCol,
				"RCCElements.RecipeElementsTable", "getColumnN", iCol, deviceDpes);
  
  for (i=1; i<=dynlen(deviceAliases); i++) {
    dyn_string deviceDpesSplit = strsplit(deviceDpes[i], ",");
    for (j=1; j<=dynlen(deviceDpesSplit); j++) {		
        dynAppend(rcpDpes, deviceAliases[i] + ".ProcessOutput." + strltrim(deviceDpesSplit[j]));
    }
  }
    
  // Get the recipe class info before applying any modifications
  unRecipeFunctions_getRecipeClassObject(sRcpClassDp, dmRcpClassInfo, exceptionInfo);
  
  // Updates the JCOP Recipe class DP
  fwConfigurationDB_modifyRecipeClass(sClassName, rcpDpes, sTypeName, sDesc, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error: The recipe class could not be modified:");
    _unRecipeFunctions_writeInRecipeLog(exceptionInfo[2]);
    dpSet(sRcpClassDp + ".ProcessInput.State", UN_RCP_CLASS_STATE_EDIT);
    return;
  }
  
  // Set the online values for the new DPEs added to the recipe 
  _unRecipeFunctions_setOnlineValuesForNewDpes(
    dmRcpClassInfo[fwConfigurationDB_RCL_ELEMENTS], 
    rcpDpes, 
    dmRcpClassInfo[fwConfigurationDB_RCL_INSTANCES]);
  
  // Remove the Edit state of the recipe class
  dpSet(sRcpClassDp + ".ProcessInput.State", "",
        sRcpClassDp + ".ProcessInput.DeviceOrder", deviceAliases);
  //sSystemName = unGenericDpFunctions_getSystemName(sRcpClassDp);
  unRecipeFunctions_loadRecipeClassData(sRcpClassDp);
  _unRecipeFunctions_writeInRecipeLog("The changes have been applied to the recipe class " + sClassName + " and all its instances.", TRUE);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * When a recipe class is modified to include new DPEs the method sets the online values in all the recipe instances
 * for the new DPEs.
 * @param dsOldDpes         - [IN] Old DPEs of the recipe class.
 * @param dsNewDpes         - [IN] New DPEs of the recipe class.
 * @param dsRcpInstNames    - [IN] Recipe instance names.
 */
void _unRecipeFunctions_setOnlineValuesForNewDpes(dyn_string dsOldDpes, dyn_string dsNewDpes, dyn_string dsRcpInstNames) {
  int i, j, pos;
  string sPvssDpe;
  dyn_int diPos;
  dyn_string dsDpeDiff, dsPvssDpes, exceptionInfo;
  dyn_anytype daOnlineValues;
  dyn_dyn_mixed recipeObject;
  
  // Get the list of new DPEs added to the recipe class
  _unRecipeFunctions_dynDiff(dsNewDpes, dsOldDpes, dsDpeDiff);
  
  if (dynlen(dsDpeDiff) == 0) {
    return;
  }

  // Convert the UNICOS DPEs to PVSS DPEs
  for (i=1; i<=dynlen(dsDpeDiff); i++) {
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(dsDpeDiff[i], sPvssDpe);
    dynAppend(dsPvssDpes, sPvssDpe);
  }
  
  for (i=1; i<=dynlen(dsRcpInstNames); i++) {
    fwConfigurationDB_loadRecipeFromCache(dsRcpInstNames[i], makeDynString(), "", recipeObject, exceptionInfo);
    
    if (i == 1) {
        // Find the positions where the online values must be modified in the recipeObject
        _unRecipeFunctions_getDynPos(dsPvssDpes, recipeObject[fwConfigurationDB_RO_DPE_NAME], diPos);
        unRecipeFunctions_getOnlineValues(dsPvssDpes, daOnlineValues, exceptionInfo);
    }
    
    // Set the online values
    for (j=1; j<=dynlen(daOnlineValues); j++) {
        pos = diPos[j];
        if (pos > 0) {
            recipeObject[fwConfigurationDB_RO_VALUE][pos] = daOnlineValues[j];
        }
    }
    
    fwConfigurationDB_saveRecipeToCache(recipeObject, "", dsRcpInstNames[i], exceptionInfo);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the position of the dsSearch elements in the dsItems list.
 * @param dsSearch  - [IN] List of elements to search in the dsItems list.
 * @param dsItems   - [IN] List of items.
 * @param diResult  - [OUT] List containing the positions of the dsSearch elements in the dsItems list.
 */
void _unRecipeFunctions_getDynPos(dyn_string dsSearch, dyn_string dsItems, dyn_int &diResult) {
  int iPos;
  
  for (int i=1; i<=dynlen(dsSearch); i++) {
    iPos = dynContains(dsItems, dsSearch[i]);
    dynAppend(diResult, iPos);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the difference of the daList1 with the daList2.
 * @param daList1   - [IN] First list of elements.
 * @param daList2   - [IN] Second list of elements.
 * @param daResult  - [OUT] List containing the elements of daList1 not contained in daList2.
 */
void _unRecipeFunctions_dynDiff(dyn_anytype daList1, dyn_anytype daList2, dyn_anytype &daResult) {
    for (int i=1; i<=dynlen(daList1); i++) {
        if (dynContains(daList2, daList1[i]) == 0) {
            dynAppend(daResult, daList1[i]);
        }
    }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Remove the selected recipe class devices from the table.
 * @param sRcpClassDp - [IN] Datapoint element of the recipe class to be modified.
 */
void unRecipeFunctions_removeSelectedRecipeClassDevices(string sRcpClassDp) {
  int i, iLen;
  dyn_int diLines;
  string sRcpState;

  getValue("RCCElements.RecipeElementsTable", "getSelectedLines", diLines);
  iLen = dynlen(diLines);
  for (i=iLen; i>=1; i--) {
    setValue("RCCElements.RecipeElementsTable", "deleteLineN", diLines[i]);
  }
  
  if (dynlen(diLines)) {
    dpGet(sRcpClassDp + ".ProcessInput.State", sRcpState);
    if (sRcpState == UN_RCP_CLASS_STATE_INVALID || sRcpState == UN_RCP_CLASS_STATE_INVALID_EDIT) {
        dpSet(sRcpClassDp + ".ProcessInput.State", UN_RCP_CLASS_STATE_INVALID_EDIT);
    } else {
        dpSet(sRcpClassDp + ".ProcessInput.State", UN_RCP_CLASS_STATE_EDIT);
    }
    setValue("applyChangesFrame", "visible", true);
    _unRecipeFunctions_writeInRecipeLog("The selected devices have been removed. Please apply the changes to the recipe class.");
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Create the recipe class device.
 * @param sClassName          - [IN]  Recipe class name.
 * @param sDescription        - [IN]  Recipe class description.
 * @param sPcoLink            - [IN]  Alias of the PCO where the recipe class will be linked.
 * @param sRcpClassDp         - [OUT] Datapoint of the new recipe class.
 * @param exceptionInfo       - [OUT] Standard exception handling variable.
 * @param dsDeviceList        - [IN]  List of recipe elements (optional).
 * @param sAcDomain           - [IN]  Access control domain (optional).
 * @param sOperatorPrivileges - [IN]  Operator privileges (optional).
 * @param sExpertPrivileges   - [IN]  Expert privileges (optional).
 * @param sAdminPrivileges    - [IN]  Admin privileges (optional).
 */
public void _unRecipeFunctions_createRecipeClass(string sClassName, string sDescription, string sPcoLink, string &sRcpClassDp, dyn_string &exceptionInfo, 
        dyn_string dsDeviceList=makeDynString(), string sAcDomain="", string sOperatorPrivileges="", string sExpertPrivileges="", string sAdminPrivileges="")
{
  int iDeviceNumber, iDriverNumber, iLen, iNumInstances;
  string sPrefix, sDeviceList = "", sPcoLinkDpe, sPvssDpe, sFrontEndName, sFrontEndDp, sAppName;
  dyn_mixed recipeClass;
  dyn_string dsConfig, dsRunningArchive, dsSplitDp, dsOnlineDpes, dsRcpNames, dsArchBool, dsArchAnalog, dsArchEvent, dsParameters;
  bool bEvent;
  
  _unRecipeFunctions_writeInRecipeLog("Creating recipe class: '" + sClassName + "' ...");
  
  // Check if the recipe class already exists in any system
  if(_unRecipeFunctions_doesRecipeClassExist(sClassName)) {
    fwException_raise(exceptionInfo,"ERROR", "The recipe class '" + sClassName + "' already exists. ","");
	return;	   
  }

  // Get the application parameters from the PCO Link
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sPcoLink, sPcoLinkDpe);
  sAppName = unGenericDpFunctions_getApplication(sPcoLinkDpe);
  
  // Initialize the fwDevice and value archive global variables
  unImportDevice_initialize();
  unImportDevice_getApplicationCharacteristics(bEvent, sPrefix);
  
  // Get necessary data to create a new recipe instance
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(UN_RECIPE_FE_ALIAS, sPvssDpe);
  if (sPvssDpe==UN_RECIPE_FE_ALIAS) {
    // The RCP_FE doesn't exist, create it
    // Get the Front End where the PCO link is defined to get the archive files
    string sPcoLinkDpName;
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sPcoLink, sPcoLinkDpName);
    string sPcoFeName = unGenericObject_GetPLCNameFromDpName(sPcoLinkDpName);
    dyn_string dsAnalog, dsBoolean, dsEvent;
    dpGet(sPcoFeName + ".configuration.archive_analog", dsAnalog,
		  sPcoFeName + ".configuration.archive_bool", dsBoolean,
		  sPcoFeName + ".configuration.archive_event", dsEvent);
	
    bool bOk = unRecipeFunctions_createRecipeFrontEnd(sAppName, sPrefix, bEvent, dsAnalog[1], dsBoolean[1], dsEvent[1], exceptionInfo);
    if (bOk == false) {
      fwException_raise(exceptionInfo, "ERROR", "The " + UN_RECIPE_FE_ALIAS + " SOFT_FE couldn't be created.", "");
		  return;
	  }
  
	  if (dynlen(exceptionInfo)) { 
      DebugTN(exceptionInfo); 
      return; 
    }
  
	  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(UN_RECIPE_FE_ALIAS, sPvssDpe);
  }

  sFrontEndName = unGenericObject_GetPLCNameFromDpName(sPvssDpe);
  strreplace(sFrontEndName, SOFT_FE_DPTYPE+"_", "");
  
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sFrontEndName, sFrontEndDp);
  dpGet(sFrontEndDp + "configuration.archive_bool", dsArchBool,
		sFrontEndDp + "configuration.archive_analog", dsArchAnalog,
		sFrontEndDp + "configuration.archive_event", dsArchEvent,
		sFrontEndDp + "communication.driver_num", iDriverNumber);
  
  // [UCPC-1467] Setting the application in the SOFT FE
  dyn_string dsSoftFeConfig;
  dynAppend(dsSoftFeConfig, SOFT_FE_DPTYPE);
  dynAppend(dsSoftFeConfig, UN_RECIPE_FE_ALIAS);
  dynAppend(dsSoftFeConfig, sAppName);
  dynAppend(dsSoftFeConfig, "N");
  dynAppend(dsSoftFeConfig, iDriverNumber);
  dynAppend(dsSoftFeConfig, dsArchBool[1]);
  dynAppend(dsSoftFeConfig, dsArchAnalog[1]);
  dynAppend(dsSoftFeConfig, "FALSE");
  dynAppend(dsSoftFeConfig, dsArchEvent[1]);
  unConfigSOFT_FE_setApplication(dsSoftFeConfig, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    DebugTN(exceptionInfo); 
    return; 
  }

  // Create the JCOP recipe class with the 'editable' flag to TRUE.
  fwConfigurationDB_createRecipeClass(sClassName, UN_CONFIG_UNRCPTYPE_DPT_NAME, sDescription, true, makeDynString(), exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    DebugTN(exceptionInfo); 
    return; 
  }
  
  // If there is a device list specified, converts the dyn_string to a string
  if (dynlen(dsDeviceList)) {
    fwGeneral_dynStringToString(dsDeviceList, sDeviceList, ",");
  }
  
  // Create a unique recipe class DP that is not existing yet
  _unRecipeFunctions_createNewRecipeDp(UN_CONFIG_UNRCPCLASS_DPT_NAME, sFrontEndName, sAppName, sRcpClassDp, iDeviceNumber);
  
  // Create the dsConfig for the device
  dynAppend(dsConfig, UN_CONFIG_UNRCPCLASS_DPT_NAME);	// Recipe instance DPType name
  dynAppend(dsConfig, iDeviceNumber);					// Device number
  dynAppend(dsConfig, sClassName);						// Device alias
  dynAppend(dsConfig, sDescription);					// Description
  dynAppend(dsConfig, "");								// Diagnostics
  dynAppend(dsConfig, "");								// WWWLink
  dynAppend(dsConfig, "");								// Synoptic
  dynAppend(dsConfig, "");								// Domain
  dynAppend(dsConfig, "");								// Nature
  dynAppend(dsConfig, "UnRcpClassState");				// Widget Type
  dynAppend(dsConfig, UN_CONFIG_UNRCPTYPE_DPT_NAME);	// Recipe Type
  dynAppend(dsConfig, sDeviceList);						// Device List
    
  unImportDevice_import(dsConfig, exceptionInfo, SOFT_FE_DPTYPE, sFrontEndName,
					    sAppName, sPrefix, iDriverNumber,
						dsArchBool[1], dsArchAnalog[1], dsArchEvent[1], bEvent,
						dsRunningArchive);
						
  if (dynlen(exceptionInfo)) {
    DebugTN(exceptionInfo);
    return; 
  }
  
  unImportDevice_triggerUpdate();
  
  // Get the default values of the privileged actions
  if(sOperatorPrivileges=="" && sExpertPrivileges=="" && sAdminPrivileges=="") {
    _unRecipeFunctions_getDeviceDefaultPrivileges(UN_CONFIG_UNRCPCLASS_DPT_NAME, sOperatorPrivileges, sExpertPrivileges, sAdminPrivileges, exceptionInfo);
    if (dynlen(exceptionInfo)) {
      DebugTN(exceptionInfo);
      return; 
    }
  } 
  
  // Set the recipe class data
  dpSet(sRcpClassDp + ".statusInformation.deviceLink", sPcoLink,
		sRcpClassDp + ".ProcessInput.ClassName", sClassName,
		sRcpClassDp + ".ProcessInput.RecipeType", UN_CONFIG_UNRCPTYPE_DPT_NAME,
        sRcpClassDp + ".statusInformation.accessControl.accessControlDomain", sAcDomain,
		sRcpClassDp + ".statusInformation.accessControl.operator", sOperatorPrivileges,
		sRcpClassDp + ".statusInformation.accessControl.expert", sExpertPrivileges,
		sRcpClassDp + ".statusInformation.accessControl.admin", sAdminPrivileges);
	
  // Update the recipe class privileges. This is necessary when the privileges are specified in the function parameters
  // and there are missing or obsolete parameters.
  unRecipeFunctions_updateRecipeClassPrivileges(sRcpClassDp);  
    
  // Set the value of the recipe class created, it will be reloaded when
  // the callback method unRecipeFunctions_deviceUpdatedRcpClassCB(..) is executed
  if (shapeExists("LoadedRcpDp")) {
    setValue("LoadedRcpDp", "text", sRcpClassDp);
  }

  _unRecipeFunctions_writeInRecipeLog("The recipe class '" + sClassName + "' has been created.", TRUE);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if a distributed system is connected.
 * @param sSystemName   - [IN] System name.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return TRUE if the system is not remote or if it is a remote and connected, otherwise FALSE.
 */
bool unRecipeFunctions_isRemoteSystemConnected(string sSystemName, dyn_string &exceptionInfo) {
  bool bIsRemote, bIsConnected;
  
  unDistributedControl_isRemote(bIsRemote, sSystemName);
  if (bIsRemote) {
	unDistributedControl_isConnected(bIsConnected, sSystemName);
	if (!bIsConnected){
    string sError = "The distributed system '" + sSystemName + "' isn't connected.";
		fwException_raise(exceptionInfo, "ERROR", sError, "");
		return false;
	}
  }
  
  return true;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Function used to know if a recipe class already exists in any system.
 * @param sClassName - [IN] Recipe class name.
 * @return TRUE if the recipe class already exists, otherwise FALSE.
 */
bool _unRecipeFunctions_doesRecipeClassExist(string sClassName) {
  dyn_string dsSystemNames, exceptionInfo;
  dyn_dyn_anytype resultTable;
  
  // Check if the recipe class already exists in any system
  dsSystemNames = mappingKeys(g_m_ddsTreeDeviceOverviewDeviceList);
  
  int iLen = dynlen(dsSystemNames);
  for (int i=1; i<=iLen; i++) {
    if (!unRecipeFunctions_isRemoteSystemConnected(dsSystemNames[i], exceptionInfo)) {
      _unRecipeFunctions_writeInRecipeLog("Warning: The remote system '" + dsSystemNames[i] + "' is not connected.", true, true);
      dynClear(exceptionInfo);
      continue;
    }
    
    dpQuery("SELECT '.ProcessInput.ClassName:_online.._value' FROM '*' REMOTE '" + dsSystemNames[i] + "' "
             + "WHERE '.ProcessInput.ClassName:_online.._value' = \""+sClassName
             +"\" AND _DPT=\""+UN_CONFIG_UNRCPCLASS_DPT_NAME+"\"", resultTable); 
             
    if (dynlen(resultTable)>1) {
      return TRUE;
    }
  }
  
  return FALSE;
}  
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Callback function from the recipe class panel. The function is called when the devices list has been updated.
 * @param sCommandDp 	- [IN] DP Name of the command over the list of devices.
 * @param iCommand   	- [IN] Command value over the list of devices.
 * @param sSystemNameDp - [IN] DP Name of the system.
 * @param sSystemName   - [IN] System name.
 */
unRecipeFunctions_deviceUpdatedRcpClassCB(string sCommandDp, int iCommand, string sSystemNameDp, string sSystemName)
{
  int iCol;
  bool bAreEqual;
  string sPcoLink, sDpName, sState, sPanelName;
  dyn_float dfReturn;
  dyn_string dsRcpClasses, dsRcpClassesRet, dsRcpClassesDpRet, exceptionInfo, dsReturn;
  
  switch(iCommand) 
  {
    case DEVUN_ADD_COMMAND:
    case DEVUN_DELETE_COMMAND:
    {
      synchronized(g_dsTreeApplication) 
      {
	    // Get the PCO link selected in the panel
		getValue("PcoList", "selectedText", sPcoLink);
		
		// Get the List of recipe classes loaded in the panel
		getValue("RecipeClasses", "nameToColumn", "RecipeClass", iCol);
		getValue("RecipeClasses", "getColumnN", iCol, dsRcpClasses);
		
		// Get the list of recipe classes linked to the PCO
		unRecipeFunctions_getRecipeClasses(dsRcpClassesRet, dsRcpClassesDpRet, sPcoLink);
		
		// Compare the two recipe classes list
		bAreEqual = _unRecipeFunctions_areListsEqual(dsRcpClasses, dsRcpClassesRet);
		
		if (!bAreEqual) {
		  // The list of recipe classes is different and it should be reloaded to avoid inconsistencies
		  getValue("LoadedRcpDp", "text", sDpName);

		  if (sDpName != "" && sDpName != "_UnRcpClass") {
		    // Check if the current recipe class is being edited
		    dpGet(sDpName + ".ProcessInput.State", sState);
		    if (sState != "") {
			    // The recipe class is being edited, display a confirmation message
			    unGraphicalFrame_ChildPanelOnCentralModalReturn(
				  "vision/MessageInfo",
				  sPanelName,
				  makeDynString("$1:The list of recipe classes has been modified.\nDo you want update the recipe class list?\n" +
						+ "(The changes of the recipe class will be applied)", "$2:Yes","$3:No"), 
				  dfReturn, 
				  dsReturn);
			    if(dynlen(dfReturn)>0 && dfReturn[1]==1)
			    { // Apply the changes to the recipe class
			      unRecipeFunctions_applyRecipeClassModifications(sDpName);
			    } else {
			      return;
			    }
		    }
        unRecipeFunctions_loadRecipeClasses(sPcoLink, exceptionInfo, FALSE);
		    unRecipeFunctions_selectRecipeClass(sDpName, exceptionInfo, TRUE);
		  } else {
		    // There is no recipe class selected, load the recipe classes
		    unRecipeFunctions_loadRecipeClasses(sPcoLink, exceptionInfo);
		  }
		}
      }
      break;
    }
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Callback function from the recipe instance panel. The function is called when the devices list has been updated.
 * @param sCommandDp 	- [IN] DP Name of the command over the list of devices.
 * @param iCommand   	- [IN] Command value over the list of devices.
 * @param sSystemNameDp - [IN] DP Name of the system.
 * @param sSystemName   - [IN] System name.
 */
unRecipeFunctions_deviceUpdatedRcpInstanceCB(string sCommandDp, int iCommand, string sSystemNameDp, string sSystemName)
{
  int iCol;
  bool bAreEqual;
  string sClassName, sLoadedRecipeInstance, sState, sPanelName;
  dyn_string dsRcpInstances, dsRcpInstanceNamesRet, dsRcpInstanceDpsRet, exceptionInfo, dsReturn;
  dyn_float dfReturn;
  
  if(g_bCreatingRecipeInstance) {
    return;
  }
  
  switch(iCommand) 
  {
    case DEVUN_ADD_COMMAND:
    case DEVUN_DELETE_COMMAND:
    {
      synchronized(g_dsTreeApplication) 
      {	
		// Get the recipe class selected in the panel
		getValue("RecipeClasses", "selectedText", sClassName);

		// Get the list of recipe instances displayed in the panel
		getValue("RecipeInstances", "nameToColumn", "RecipeInstances", iCol);
		getValue("RecipeInstances", "getColumnN", iCol, dsRcpInstances);
		
		// Get the list of recipe instances of the selected class
		_unRecipeFunctions_getRecipeInstancesOfClass(sClassName, dsRcpInstanceNamesRet, dsRcpInstanceDpsRet);

		// Compare the two recipe classes list
		bAreEqual = _unRecipeFunctions_areListsEqual(dsRcpInstances, dsRcpInstanceNamesRet);
		
		if (!bAreEqual) {
		  // The list of recipe instances is different and it should be reloaded to avoid inconsistencies
		  
		  getValue("LoadedRecipeDp", "text", sLoadedRecipeInstance);
		  if (sLoadedRecipeInstance!="") {
  		    // Check if the current recipe has been removed
			  
			if (!dpExists(sLoadedRecipeInstance + ".ProcessInput.State")) {
			  _unRecipeFunctions_writeInRecipeLog("The recipe has been removed.");
			  _unRecipeFunctions_clearInstanceTables();
			  _unRecipeFunctions_loadRecipeInstances(sClassName);
			  return;
			} 
			  
			// Check if the current recipe is being edited
			dpGet(sLoadedRecipeInstance + ".ProcessInput.State", sState);
			if (sState == UN_RCP_INSTANCE_STATE_EDIT) {
			  // The recipe is being edited, display a confirmation message
			  unGraphicalFrame_ChildPanelOnCentralModalReturn(
					"vision/MessageInfo",
					sPanelName,
					makeDynString("$1:The list of recipe instances has been modified.\nDo you want update the recipe instance list?\n" +
							+ "(The changes of the recipe will be applied)", "$2:Yes","$3:No"), 
					dfReturn, 
					dsReturn);
			  if(dynlen(dfReturn)>0 && dfReturn[1]==1)
			  { // Apply the changes to the recipe 
				unRecipeFunctions_saveRecipeInstance($sDpName, exceptionInfo, FALSE);
			  } else {
				return;
			  }
			}
		  }
		  _unRecipeFunctions_loadRecipeInstances(sClassName);
		  unRecipeFunctions_selectRecipeInstance(sLoadedRecipeInstance, exceptionInfo);
		}
      }
      break;
    }
  }
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Function to check if the items of two lists are equal.
 * @param list1 - [IN] The first list of elements.
 * @param list2 - [IN] The second list of elements.
 * @return TRUE if the lists contain the same elements, otherwise FALSE.
 */
bool _unRecipeFunctions_areListsEqual(dyn_string list1, dyn_string list2) {
  if (dynlen(list1) != dynlen(list2)) {
    return false;
  }
	
  dynSortAsc(list1);
  dynSortAsc(list2);
  
  int iLen = dynlen(list1);
  for (int i=1; i<=iLen; i++) {
    if (list1[i] != list2[i]) {
	  return false;
    }
  }
  
  return true;
}

//------------------------------------------------------------------------------------------------------------------------
public void unRecipeFunctions_getRecipeInstancePredefinedPrivileges
    (string sClassName, string &sRcpInstanceDp, string &sAcDomain, string &sOperatorAction, string &sExpertAction, string &sAdminAction) {
  
  // Get the recipe instance privileges and ACDomain.
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sClassName + "/" + UN_RCP_PREDEFINED_NAME, sRcpInstanceDp);

  dpGet(sRcpInstanceDp + "accessControl.accessControlDomain", sAcDomain,
        sRcpInstanceDp + "accessControl.operator", sOperatorAction,
        sRcpInstanceDp + "accessControl.expert", sExpertAction,
        sRcpInstanceDp + "accessControl.admin", sAdminAction);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Open a new window to edit the recipe class privileges and parameters.
 * @param sRcpClassDp - [IN] The recipe class DP.
 */
void unRecipeFunctions_editRecipeClassPrivileges(string sRcpClassDp) {
  int i, iLen;
  string sClassName, sDefinedACDomains, sClassACDomain, sClassOperatorAction, sClassExpertAction, sClassAdminAction;
  string sRcpInstanceDp, sInstanceACDomain, sInstanceOperatorAction, sInstanceExpertAction, sInstanceAdminAction;
  dyn_string ds, dsAccessControlDomains, dsFullAccessControlDomain, exceptionInfo, dsRcpInstanceNames, dsRcpInstanceDps;
  dyn_float df;
  
  // Get the recipe class privileges and ACDomain.
  sClassName = dpGetAlias(sRcpClassDp + ".");
  _unRecipeFunctions_getDevicePrivileges(sRcpClassDp, sClassACDomain, sClassOperatorAction, 
			sClassExpertAction, sClassAdminAction);
			
  // Get the recipe instance privileges and ACDomain.
  unRecipeFunctions_getRecipeInstancePredefinedPrivileges(sClassName, sRcpInstanceDp, sInstanceACDomain, sInstanceOperatorAction, sInstanceExpertAction, sInstanceAdminAction);
		
  // Get the list of access control domains
  fwAccessControl_getAllDomains(dsAccessControlDomains, dsFullAccessControlDomain, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    string errMsg = "Error getting the list of access control domains.";
    _unRecipeFunctions_writeInRecipeLog(errMsg);
    DebugTN(errMsg, exceptionInfo);
    return;
  }
  
  // Remove the SYSTEM DOMAIN if it exists
  if ( (i=dynContains(dsAccessControlDomains, "SYSTEM")) > 0) {
    dynRemove(dsAccessControlDomains, i);
  }
  fwGeneral_dynStringToString(dsAccessControlDomains, sDefinedACDomains, ",");

  // Display the panel with the recipe class and recipe instance privileges  
  ChildPanelOnCentralModalReturn(
		"vision/unRecipe/UnRcpClass/unUnRcpClass_Privileges.pnl",
		sClassName + " - Edit recipe privileges",
		makeDynString(
			"$sDpName:" + sRcpClassDp,
			"$sDefinedACDomains:" + sDefinedACDomains,
			"$sClassACDomain:" + sClassACDomain,
			"$sClassOperatorAction:" + sClassOperatorAction,
			"$sClassExpertAction:" + sClassExpertAction,
			"$sClassAdminAction:" + sClassAdminAction,
			"$sInstanceACDomain:" + sInstanceACDomain,
			"$sInstanceOperatorAction:" + sInstanceOperatorAction,
			"$sInstanceExpertAction:" + sInstanceExpertAction,
			"$sInstanceAdminAction:" + sInstanceAdminAction
			),
		df, ds);
		
  // Check if the Ok button has been pressed
  if (dynlen(df) && df[1]>0) {
    // Set the recipe class privileges
	sClassACDomain = ds[1];
	sClassOperatorAction = ds[2];
	sClassExpertAction = ds[3];
	sClassAdminAction = ds[4];	
	_unRecipeFunctions_setDevicePrivileges(sRcpClassDp, sClassACDomain, sClassOperatorAction, 
			sClassExpertAction, sClassAdminAction, exceptionInfo);
	
	// Get the recipe instance privileges
	sInstanceACDomain = ds[5];
	sInstanceOperatorAction = ds[6];
	sInstanceExpertAction = ds[7];
	sInstanceAdminAction = ds[8];

	// Apply the privileges to the default recipe instance
	dpSet(sRcpInstanceDp + "accessControl.accessControlDomain", sInstanceACDomain,
        sRcpInstanceDp + "accessControl.operator", sInstanceOperatorAction,
		sRcpInstanceDp + "accessControl.expert", sInstanceExpertAction,
		sRcpInstanceDp + "accessControl.admin", sInstanceAdminAction);
	
	// Get all the recipe instances of the class
	_unRecipeFunctions_getRecipeInstancesOfClass(sClassName, dsRcpInstanceNames, dsRcpInstanceDps);
	// Set the recipe instances privileges
	iLen = dynlen(dsRcpInstanceDps);
	for (i=1; i<=iLen; i++) {
		_unRecipeFunctions_setDevicePrivileges(dsRcpInstanceDps[i], sInstanceACDomain, sInstanceOperatorAction, 
			sInstanceExpertAction, sInstanceAdminAction, exceptionInfo);
	}
	
	_unRecipeFunctions_writeInRecipeLog("The privileges of the recipe class " + sClassName + 
			" and all its instances have been modified.", TRUE);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Auxiliary function to get the device privileges.
 * @param sDpName 	   - [IN] Datapoint name of the device.
 * @param sAcDomain	   - [OUT] Access control domain for the device.
 * @param sOperatorAction - [OUT] Privileged actions for the operator.
 * @param sExpertAction   - [OUT] Privileged actions for the expert.
 * @param sAdminAction	  - [OUT] Privileged actions for the admin.
 */
void _unRecipeFunctions_getDevicePrivileges(string sDpName, string & sAcDomain, string & sOperatorAction, 
			string & sExpertAction, string & sAdminAction) 
{
	_unRecipeFunctions_normalizeDp(sDpName);

	dpGet(sDpName + ".statusInformation.accessControl.accessControlDomain", sAcDomain,
 		  sDpName + ".statusInformation.accessControl.operator", sOperatorAction,
		  sDpName + ".statusInformation.accessControl.expert", sExpertAction,
		  sDpName + ".statusInformation.accessControl.admin", sAdminAction);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Auxiliary function to set the device privileges.
 * @param sDpName 	      - [IN] Datapoint name of the device.
 * @param sAcDomain	      - [IN] Access control domain for the device.
 * @param sOperatorAction - [IN] Privileged actions for the operator.
 * @param sExpertAction   - [IN] Privileged actions for the expert.
 * @param sAdminAction	  - [IN] Privileged actions for the admin.
 * @param exceptionInfo   - [OUT] Standard exception handling variable.
 */
void _unRecipeFunctions_setDevicePrivileges(string sDpName, string sAcDomain, string sOperatorAction, 
			string sExpertAction, string sAdminAction, dyn_string & exceptionInfo) 
{
    _unRecipeFunctions_normalizeDp(sDpName);
	
	dpSet(sDpName + ".statusInformation.accessControl.accessControlDomain", sAcDomain,
 		  sDpName + ".statusInformation.accessControl.operator", sOperatorAction,
		  sDpName + ".statusInformation.accessControl.expert", sExpertAction,
		  sDpName + ".statusInformation.accessControl.admin", sAdminAction);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Function used to get the default privileges for the recipe class actions.
 * @param sRoleType    - [IN] Role type (operator, expert or admin).
 * @param dsPrivileges - [OUT] List of default privileged actions for the specified role.
 */
void _unRecipeFunctions_getDefaultClassPrivileges(string sRoleType, dyn_string & dsPrivileges) 
{
  string sRole = strtolower(sRoleType);
  dynClear(dsPrivileges);
  
  if (sRole=="operator") {
    return;
  } else if (sRole=="expert") {
    return;
  } else if (sRole=="admin") {
    dynAppend(dsPrivileges, makeDynString("AddDevices", "ApplyChanges", "Cancel", "Duplicate", "EditElements", "NewClass", 
			"NewInstance", "Privileges", "RemoveClass", "RemoveSelected", "SaveToDb", "LoadFromDb", UN_FACEPLATE_BUTTON_SELECT));
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Function used to get the default privileges for the recipe instance actions.
 * @param sRoleType    - [IN] Role type (operator, expert or admin).
 * @param dsPrivileges - [OUT] List of default privileged actions for the specified role.
 */
void _unRecipeFunctions_getDefaultInstancePrivileges(string sRoleType, dyn_string & dsPrivileges) 
{
  string sRole = strtolower(sRoleType);
  dynClear(dsPrivileges);
  
  if (sRole=="operator") {
    dynAppend(dsPrivileges, makeDynString("Activate", UN_FACEPLATE_BUTTON_SELECT));
  } else if (sRole=="expert"){
    dynAppend(dsPrivileges, makeDynString("Cancel", "Delete", "Duplicate", "Edit", "InitialRecipes", "NewInstance", 
				"LastActivatedRecipes", "OnlineValues", "DbValues", "Save", "SaveToDb",  "SaveAsInitial"));
  } else if (sRole=="admin") {
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Auxiliary function to add values to a mapping where the mapping value is a dynamic list.
 * @param m   		   - [OUT] Mapping where to insert the values.
 * @param key 		   - [IN] Key for the new value.
 * @param value        - [IN] The value to insert.
 * @param bUniqueValue - [IN] If TRUE the values can not be repeated for the same key.
 */
public void _unRecipeFunctions_mappingInsertValue(mapping & m, string key, string value, bool bUniqueValue=TRUE) 
{
  dyn_string dsValues;
  
  if (mappingHasKey(m, key)) {
    dsValues = m[key];
  }
  
  if (bUniqueValue==FALSE || FALSE==dynContains(dsValues, value)) {
    dynAppend(dsValues, value);
  }

  m[key] = dsValues;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Create the SOFT_FE for the recipes.
 * @param sAppName - [IN] Application name.
 * @param sPrefix  - [IN] Application prefix.
 * @param bEvent   - [IN] TRUE=16 bit event type, FALSE=32 bit event type
 * @param sAnalog  - [IN] Name of the analog archive file.
 * @param sBool    - [IN] Name of the boolean archive file.
 * @param sEvent   - [IN] Name of the event archive file.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */ 
bool unRecipeFunctions_createRecipeFrontEnd(string sAppName, string sPrefix, bool bEvent, string sAnalog, string sBool, string sEvent, dyn_string & exceptionInfo)
{
  string sPvssDpe;
  
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(UN_RECIPE_FE_ALIAS, sPvssDpe);
  if (sPvssDpe==UN_RECIPE_FE_ALIAS) {

	dyn_float df;
	dyn_string ds;
	ChildPanelOnCentralModalReturn(
		"vision/unRecipe/UnRcpFrontEnd/unUnRcpFrontEnd_Create.pnl",
		"Recipe Front End Creation",
		makeDynString(
			"$sAppName:" + sAppName,
			"$sPrefix:" + sPrefix,
			"$bEvent:" + bEvent,
			"$sAnalog:" + sAnalog,
			"$sBool:" + sBool,
			"$sEvent:" + sEvent),
		df, ds);

	if (dynlen(df)==1 && df[1]==1) {
		return true;
    }
	return false;
  } 
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Create a new recipe instance of the selected recipe class.
 */
void unRecipeFunctions_createRecipeInstance()
{
  int iLen;
  string sAllowInitialRecipe = "FALSE";
  string sRecipeClass, sDpName, sRcpName, sRcpDesc, sInitial, sRecipeClassesType;
  string sDefaultRcpInstDp, sAcDomain, sOperatorAction, sExpertAction, sAdminAction;
  dyn_int lines;
  dyn_string exceptionInfo, ds, dsSelectedClassNames;
  dyn_float df;
  
  // Check that all the necessary shapes exist
  if (!shapeExists("RecipeClasses")) {
	_unRecipeFunctions_writeInRecipeLog("Error: Some graphical shapes are missing.");
	return;
  } 
  
  // TODO: This could be a function ... check if it's repeated in the code.
  // Get the type of the RecipeClasses shape (it can be "Table" or "ComboBox")
  getValue("RecipeClasses", "type", sRecipeClassesType);  
  if (sRecipeClassesType=="Table") {
	// Get the selected recipe class from the table
	getValue("RecipeClasses", "getSelectedLines", lines); 
	if (!dynlen(lines)>0) {
		_unRecipeFunctions_writeInRecipeLog("Error: Select a recipe class before creating a recipe.");
		return;
	}    
	getValue("RecipeClasses", "cellValueRC", lines[1], "RecipeClass", sRecipeClass);  
  } else if (sRecipeClassesType=="ComboBox") {
    // Get the selected recipe class from the combo box
	getValue("RecipeClasses", "selectedText", sRecipeClass);
  } else {
    // The type of the RecipeClasses shape is unknown
	_unRecipeFunctions_writeInRecipeLog("Error: The type of the RecipeClasses shape is unknown.");
	return;
  }
  
  // Check if the recipe class name is empty
  if (sRecipeClass=="" || sRecipeClass=="-None-") {
    _unRecipeFunctions_writeInRecipeLog("Error: Please, select a recipe class before creating a recipe.");
    return;
  }
  
  // Check if the recipe class doesn't exist
  sDpName = _unGenericDpFunctions_dpAliasToName(sRecipeClass);
  if (sDpName == "") {
	_unRecipeFunctions_writeInRecipeLog("Error: The recipe class '" + sRecipeClass + "' doesn't exist.");
	return;
  }

  // Check if an initial recipe of the class already exists  
  if (!_unRecipeFunctions_hasInitialRecipe(sRecipeClass)) {
    sAllowInitialRecipe = "TRUE"; 
  }
  
  ChildPanelOnCentralModalReturn("vision/unRecipe/UnRcpInstance/unUnRcpInstance_Duplicate.pnl",
			"Create Recipe Instance",
			makeDynString("$sTitle:Create Recipe of class: " + sRecipeClass,
                "$sRecipeClass:" + sRecipeClass,
                "$sDescription: Input the name for the new recipe instance",
                "$sAllowInitialRecipe:"+sAllowInitialRecipe),
			df,ds);
  
  if (!dynlen(df) || df[1]<0) {
    return;
  }
  
  if (!dynlen(ds) || ds[1]=="") {
    _unRecipeFunctions_writeInRecipeLog("Error: The recipe name can't be empty. ");
    DebugTN(exceptionInfo);
    return;     
  }  			
  
  sRcpName = ds[1];
  sRcpDesc = ds[2];
  sInitial = ds[3];
  dsSelectedClassNames = strsplit(ds[4], "|");

  if (!_unRecipeFunctions_isLegalRecipeName(sRcpName)) {
	_unRecipeFunctions_writeInRecipeLog("Error: Wrong recipe name: " + sRcpName);
    return;  
  }

  iLen = dynlen(dsSelectedClassNames);
  for (int i=1; i<=iLen; i++) {
    _unRecipeFunctions_writeInRecipeLog("Creating recipe '" + sRcpName + "' of class '" + dsSelectedClassNames[i] + "' ...");
    unRecipeFunctions_getRecipeInstancePredefinedPrivileges(dsSelectedClassNames[i], sDefaultRcpInstDp, sAcDomain, sOperatorAction, sExpertAction, sAdminAction);
    _unRecipeFunctions_createRecipeInstance(dsSelectedClassNames[i], sRcpName, sRcpDesc, exceptionInfo, sInitial, "", sAcDomain, sOperatorAction, sExpertAction, sAdminAction);
    if (dynlen(exceptionInfo)) { 
      _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
      DebugTN(exceptionInfo); 
      dynClear(exceptionInfo);
      continue; 
    }
    _unRecipeFunctions_writeInRecipeLog("The recipe '" + sRcpName + "' of class '" + dsSelectedClassNames[i] + "' has been created.", TRUE);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if a recipe name is legal (for a recipe class or a recipe instance).
 * @param sName - [IN] Recipe name.
 * @return TRUE if the recipe name is legal, otherwise FALSE.
 */
bool _unRecipeFunctions_isLegalRecipeName(string sName) {
  if (!dpIsLegalName(sName) || strpos(sName, "/") >= 0 || strpos(sName, "\\") >=0 || strpos(sName, ".") >= 0 ) {
    return FALSE;
  }
  return TRUE;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Check if an initial recipe instance exists for the specified class.
 * @param sRecipeClass - [IN] Recipe class name.
 * @return TRUE if the specified recipe class has an initial recipe, otherwise FALSE.
 */
bool _unRecipeFunctions_hasInitialRecipe(string sRecipeClass) {
  string sQuery;
  dyn_dyn_anytype resultTable;

  sQuery = "SELECT '.RecipeName:_online.._value' FROM '*' WHERE "
              + "('.RecipeName:_online.._value' LIKE \"" + sRecipeClass + "/*\" OR '.RecipeName:_online.._value' LIKE \"*:" + sRecipeClass + "/*\") "
              + " AND '.MetaInfo.Predefined:_online.._value' = \"TRUE\"  AND _DPT=\"_FwRecipeCache\" ";
  
  // Check if an initial recipe of the class already exists
  dpQuery(sQuery, resultTable);   
  
  // Check if there isn't an initial recipe for the recipe class 
  if (dynlen(resultTable)<=1) {
    return FALSE;
  }
  
  return TRUE;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Check if a recipe class has an instance with the specified name.
 * @param sClassName    - [IN] Recipe class name.
 * @param sInstanceName - [IN] Recipe instance name.
 * @return TRUE if the recipe class has an instance with the specified name, otherwise FALSE.
 */
bool _unRecipeFunctions_hasRecipeInstance(string sClassName, string sInstanceName) {
  dyn_dyn_anytype resultTable;
  
  dpQuery("SELECT '.ProcessInput.InstanceName:_online.._value' FROM '*' WHERE '.ProcessInput.ClassName:_online.._value' = \""+sClassName+
				"\" AND _DPT=\""+UN_CONFIG_UNRCPINSTANCE_DPT_NAME+"\"", resultTable); 
  
  if (dynlen(resultTable)>1) {
    dynRemove(resultTable, 1);
    for (int i=1; i<=dynlen(resultTable); i++) {
		// If the recipe instance already exists, show an error message and return
		if (resultTable[i][2] == sInstanceName) {
			return TRUE;	
		}
    }    
  }
  
  return FALSE;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Create a new recipe instance.
 * @param sClassName 	      - [IN]  Name of the recipe class.
 * @param sInstanceName       - [IN]  Name of the new recipe instance.
 * @param sDescription        - [IN]  Description for the new recipe instance.
 * @param exceptionInfo       - [OUT] Standard exception handling variable.
 * @param sInitial		      - [IN]  Flag to define if the new recipe is the initial recipe of the class 
 * @param sDpeValues 	      - [IN]  Recipe DPE|Values string. If the parameter is not specified the recipe elements will have the online values.
 * @param sAcDomain           - [IN]  Access control domain (optional).
 * @param sOperatorPrivileges - [IN]  Operator privileges (optional).
 * @param sExpertPrivileges   - [IN]  Expert privileges (optional).
 * @param sAdminPrivileges    - [IN]  Admin privileges (optional).
 */ 
synchronized void _unRecipeFunctions_createRecipeInstance(string sClassName, string sInstanceName, string sDescription, 
        dyn_string &exceptionInfo, string sInitial="FALSE", string sDpeValues="", string sAcDomain="",
        string sOperatorPrivileges="", string sExpertPrivileges="", string sAdminPrivileges="")
{
  int iDeviceNumber, iDriverNumber, iLen, iNumInstances;
  string sDpName, sPrefix, sDpeValue = "", sRcpClassDp, sFrontEndName, sFrontEndDp, sAppName, sDomain, sNature;
  dyn_mixed recipeClass;
  dyn_string dsConfig, dsRunningArchive, dsSplitDp, dsOnlineDpes, dsRcpNames, dsArchBool, dsArchAnalog, dsArchEvent, dsParameters;
  dyn_string dsOperatorPrivileges, dsExpertPrivileges, dsAdminPrivileges, dsAllInstanceNames, dsAllInstanceDps;
  bool bEvent;

  g_bCreatingRecipeInstance = true;
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sClassName, sRcpClassDp);
  if (!_unRecipeFunctions_isRecipeClassDefinitionValid(sRcpClassDp, exceptionInfo)) {
    g_bCreatingRecipeInstance = false;
    return;
  }

  // Check if the recipe instance already exists
  if (_unRecipeFunctions_hasRecipeInstance(sClassName, sInstanceName)) {
    fwException_raise(exceptionInfo, "ERROR", "The recipe " + sClassName + "/" + sInstanceName +" already exists. ","");
    DebugTN(exceptionInfo);
    g_bCreatingRecipeInstance = false;
    return;	
  }

  // Check if the initial recipe already exists
  if (sInitial == "TRUE" && _unRecipeFunctions_hasInitialRecipe(sClassName)) {
    fwException_raise(exceptionInfo, "ERROR", "The recipe class" + sClassName + " already has an initial recipe. ","");
    DebugTN(exceptionInfo);
    g_bCreatingRecipeInstance = false;
    return;	
  }

  // Get necessary data to create a new recipe instance
  sFrontEndName = unGenericObject_GetPLCNameFromDpName(sRcpClassDp);
  strreplace(sFrontEndName, SOFT_FE_DPTYPE+"_", "");
  sAppName = unGenericDpFunctions_getApplication(sRcpClassDp);
  unGenericDpFunctions_getParameters(sRcpClassDp, dsParameters);
  sDomain = dsParameters[4];
  sNature = dsParameters[5];

  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sFrontEndName, sFrontEndDp);
  dpGet(sFrontEndDp + "configuration.archive_bool", dsArchBool,
		sFrontEndDp + "configuration.archive_analog", dsArchAnalog,
		sFrontEndDp + "configuration.archive_event", dsArchEvent,
		sFrontEndDp + "communication.driver_num", iDriverNumber);

  unRecipeFunctions_getRecipeClassObject(sRcpClassDp, recipeClass, exceptionInfo);
  if (dynlen(exceptionInfo)) {	// || dynlen(recipeClass[fwConfigurationDB_RCL_ELEMENTS])==0 ) 
    fwException_raise(exceptionInfo,"ERROR", "Error getting the recipe class data. The recipe can't be created.", "");
    DebugTN(exceptionInfo);
    g_bCreatingRecipeInstance = false;
    return;
  }

  // Creates a unique recipe DP that is not existing yet
  _unRecipeFunctions_createNewRecipeDp(UN_CONFIG_UNRCPINSTANCE_DPT_NAME, sFrontEndName, sAppName, sDpName, iDeviceNumber);
 
  // Initialize the fwDevice and value archive global variables
  unImportDevice_initialize();
  unImportDevice_getApplicationCharacteristics(bEvent, sPrefix);

  if (sDpeValues == "") {
    // If the list of DPE|Value is not specified, get the online values for the recipe elements.
    sDpeValues = _unRecipeFunctions_getDpeValueString(recipeClass[fwConfigurationDB_RCL_ELEMENTS], sDpeValues);
  }
  
  // Creates the dsConfig for the device
  dynAppend(dsConfig, UN_CONFIG_UNRCPINSTANCE_DPT_NAME);	// Recipe instance DPType name
  dynAppend(dsConfig, iDeviceNumber);						// Device number
  dynAppend(dsConfig, sClassName+"/"+sInstanceName);		// Device alias
  dynAppend(dsConfig, sDescription);						// Description
  dynAppend(dsConfig, "");									// Diagnostics
  dynAppend(dsConfig, "");									// WWWLink
  dynAppend(dsConfig, "");									// Synoptic
  dynAppend(dsConfig, sDomain);								// Domain
  dynAppend(dsConfig, sNature);								// Nature
  dynAppend(dsConfig, "UnRcpInstanceState");				// Widget Type
  dynAppend(dsConfig, sClassName);							// Recipe class
  dynAppend(dsConfig, sInitial);							// Initial recipe
  dynAppend(dsConfig, sDpeValues);							// Recipe DPE-Value definitions

  unImportDevice_import(dsConfig, exceptionInfo, SOFT_FE_DPTYPE, sFrontEndName,
					    sAppName, sPrefix, iDriverNumber,
						dsArchBool[1], dsArchAnalog[1], dsArchEvent[1], bEvent,
						dsRunningArchive);
                        
  if (dynlen(exceptionInfo)) {
    DebugTN(exceptionInfo);
	if (dpExists(sDpName)) {
		dpDelete(sDpName);
	}
    g_bCreatingRecipeInstance = false;
	return; 
  }

  unImportDevice_triggerUpdate();
  
  // Get the default values of the privileged actions
  if (sOperatorPrivileges=="" && sExpertPrivileges=="" && sAdminPrivileges=="") {
    string sPredefinedRcpInstance;
    unRecipeFunctions_getRecipeInstancePredefinedPrivileges(sClassName, sPredefinedRcpInstancep, sAcDomain, sOperatorPrivileges, sExpertPrivileges, sAdminPrivileges);
  } 

  // Set the recipe instance privileges
  dpSetWait(sDpName + ".statusInformation.accessControl.accessControlDomain", sAcDomain,
        sDpName + ".statusInformation.accessControl.operator", sOperatorPrivileges,
		sDpName + ".statusInformation.accessControl.expert", sExpertPrivileges,
		sDpName + ".statusInformation.accessControl.admin", sAdminPrivileges);

  unRecipeFunctions_updateRecipeInstancePrivileges(sDpName, sClassName);
  
  g_bCreatingRecipeInstance = false;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get a string containing the list of DPEs and values with the format DPName|Value[,DPName|Value]*
 * @param dsRcpDpes   - [IN] List of recipe elements (e.g. Device1.ProcessOutput.ManReg01);
 * @param dsRcpValues - [IN] List of recipe values for the dsRecipeDpes.
 * @return String containing the list of DPEs and values with the format DPName|Value[,DPName|Value]*
 */
string _unRecipeFunctions_getDpeValueString(dyn_string dsRcpDpes, dyn_string dsRcpValues) {
  int iLen;
  string sPvssDpe, sDpeValue;
  dyn_string dsSplitDp, dsOnlineDpes, dsTmpDpes, dsDpNameValue, exceptionInfo;

  iLen = dynlen(dsRcpDpes);
  if (iLen<=0) {
    return "";
  }
  
  if (dynlen(dsRcpValues)==0 || dynlen(dsRcpValues)!=iLen) { 
    // Get the online values
    dynClear(dsRcpValues);
    for (int i=1; i<=iLen; i++) {
      dsSplitDp = strsplit(dsRcpDpes[i], ".");
      unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(dsSplitDp[1] + "." + dsSplitDp[dynlen(dsSplitDp)], sPvssDpe);
      _unRecipeFunctions_getOnlineValueDpes (makeDynString(sPvssDpe), dsTmpDpes, exceptionInfo);
      dynAppend(dsOnlineDpes, dsTmpDpes);
    }
    dpGet(dsOnlineDpes, dsRcpValues);
  } 
  
  // Create the string containing the DPName|Value[,DPName|Value]*
  for (int i=1; i<=iLen; i++) {
    dsSplitDp = strsplit(dsRcpDpes[i], ".");
    dynAppend(dsDpNameValue, dsSplitDp[1] + "." + dsSplitDp[dynlen(dsSplitDp)] + "|" + dsRcpValues[i]);
  }

  fwGeneral_dynStringToString(dsDpNameValue, sDpeValue, ",");
  
  return sDpeValue;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Join the content of two dyn_string variables if they have the same number of elements.
 * @param ds1 - [IN] The first dyn_string variable.
 * @param ds2 - [IN] The second dyn_string variable.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return New dyn_string with the concatenated elements of the variables ds1 and ds2.
 */
dyn_string _unRecipeFunctions_joinDynString(dyn_string ds1, dyn_string ds2, dyn_string &exceptionInfo) {
  dyn_string dsResult;
  
  if (dynlen(ds1) != dynlen(ds2)) {
    fwException_raise(exceptionInfo, "ERROR", "The number of elements in the lists is different.","");
    return dsResult;
  }
  
  int iLen = dynlen(ds1);
  for (int i=1; i<=iLen; i++) {
    dynAppend(dsResult, ds1[i] + ds2[i]);
  }
  
  return dsResult;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Modify the font type of the edited rows to Bold.
 * @param sDpName - [IN] The data point name of the edited row.
 */
void _unRecipeFunctions_highlightEditedValue(string sDpName)
{
	int retVal;
	string newValue;
	dyn_string exceptionInfo, formattedValues;
	dyn_dyn_mixed recipeObject;
	
	if (g_editingRowNumber < 0 || g_editingColName=="") {
		return;
	}
	
	_unRecipeFunctions_getRecipeObjectFromDp (sDpName, recipeObject, exceptionInfo);
	getValue("RecipeElements", "cellValueRC", g_editingRowNumber, g_editingColName, newValue);
					
	_unRecipeFunctions_formatValues(makeDynString(recipeObject[fwConfigurationDB_RO_DPE_NAME][g_editingRowNumber+1]), 
					makeDynString(newValue), formattedValues, exceptionInfo);

	setValue ("RecipeElements", "cellValueRC", g_editingRowNumber, g_editingColName, formattedValues[1]);
	if (formattedValues[1] != recipeObject[fwConfigurationDB_RO_VALUE][g_editingRowNumber+1]) {
		RecipeElements.rowFontType (g_editingRowNumber, makeDynInt(1));
	}
	
	g_editingRowNumber = -1;
	g_editingColName = "";	
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * The edition of a cell value has been completed.
 * @param sDpName 	- [IN] The data point name of the recipe instance.
 * @param row	  	- [IN] Row number of the edited cell.
 * @param column  	- [IN] Column name of the edited cell.
 * @param iRowIndex - [IN] Row index in the "Index" column of the table (proper order in the JCOP recipe definition).
 */
void _unRecipeFunctions_cellEditionFinished(string sDpName, int row, string column, int iRowIndex)
{
	int retVal;
	string newValue;
	dyn_string exceptionInfo, formattedValues;
	dyn_dyn_mixed recipeObject;
	
	_unRecipeFunctions_getRecipeObjectFromDp (sDpName, recipeObject, exceptionInfo);
	getValue("RecipeElements", "cellValueRC", row, column, newValue);
	
	// Check if the recipe value is in the defined range.
	retVal = unRecipeFunctions_checkRecipeValue (
					recipeObject[fwConfigurationDB_RO_DPE_NAME][iRowIndex], 
					newValue, 
					exceptionInfo);
	
	// If the value is out of range shows an error message and modifies the cell color
	if (retVal == -1) {
	  _unRecipeFunctions_writeInRecipeLog("The new value is out of range");
	  setValue("RecipeElements", "cellBackColRC", (row), "Value", CELL_VALUE_OUT_OF_RANGE_COLOR);
	  return;
	} else {
		// Restore the cell color
		setValue("RecipeElements", "cellBackColRC", (row), "Value", CELL_EDIT_COLOR);
	}
	
	_unRecipeFunctions_formatValues(makeDynString(recipeObject[fwConfigurationDB_RO_DPE_NAME][iRowIndex]), 
								    makeDynString(newValue), formattedValues, exceptionInfo);

	if (dynlen(exceptionInfo)) {
		// The value introduced is wrong
		_unRecipeFunctions_writeInRecipeLog("The introduced value is not valid: '" + newValue + "'.");
		setValue ("RecipeElements", "cellValueRC", row, column, recipeObject[fwConfigurationDB_RO_VALUE][iRowIndex]);	
		RecipeElements.rowFontType (row, makeDynInt(0));
	} else {
		setValue ("RecipeElements", "cellValueRC", row, column, formattedValues[1]);
		if (formattedValues[1] != recipeObject[fwConfigurationDB_RO_VALUE][iRowIndex]) {
			RecipeElements.rowFontType (row, makeDynInt(1));
		} else {
			RecipeElements.rowFontType (row, makeDynInt(0));
		}
	}
	g_editingRowNumber = -1;
	g_editingColName = "";	
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Duplicate the selected recipe instance.
 * @param sRecipeDp - [IN] DP name of the recipe instance to duplicate.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_duplicateRecipeInstance(string sRecipeDp, dyn_string &exceptionInfo)
{
  string sRecipeInstance, sRecipeClass, sRecipeDesc, sInitial;
  dyn_float df;
  dyn_string ds;
  dyn_dyn_mixed recipeObject;  
  
  _unRecipeFunctions_normalizeDp(sRecipeDp);
  
  // Recipe instance in DUPLICATE mode
  dpSet(sRecipeDp+".ProcessInput.State", UN_RCP_INSTANCE_STATE_DUPLICATE);
  if (!shapeExists("RecipeClasses") || !shapeExists("RecipeInstances")) {
    dpSet(sRecipeDp+".ProcessInput.State", "");
	fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	return;
  }
  
  _unRecipeFunctions_getRecipeObjectFromDp(sRecipeDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    dpSet(sRecipeDp+".ProcessInput.State", "");
    DebugTN(exceptionInfo);	
    return; 
  }
  
  ChildPanelOnCentralReturn("vision/unRecipe/UnRcpInstance/unUnRcpInstance_Duplicate.pnl",
			"Duplicate Recipe Instance",
			makeDynString("$sTitle:Duplicate Recipe Instance",
				      "$sDescription: Input the name for the new recipe instance"),
			df,ds);
  
  if (!dynlen(df) || df[1]<0) {
    dpSet(sRecipeDp+".ProcessInput.State", "");
    return;
  }
  
  if (!dynlen(ds) || ds[1]=="") {
    dpSet(sRecipeDp+".ProcessInput.State", "");
    fwException_raise(exceptionInfo,"ERROR", "The recipe name can't be empty. ","");
    return;     
  }  
  
  sRecipeInstance = ds[1];
  sRecipeDesc = ds[2];
  sInitial = "FALSE";
  sRecipeClass = recipeObject[fwConfigurationDB_RO_META_CLASSNAME];   
  
  if (!_unRecipeFunctions_isLegalRecipeName(sRecipeInstance)) {
	_unRecipeFunctions_writeInRecipeLog("Error: Wrong recipe name: " + sRecipeInstance);
    dpSet(sRecipeDp+".ProcessInput.State", "");
    return;  
  }
  
  _unRecipeFunctions_writeInRecipeLog("Creating the recipe '" + sRecipeInstance + "' of class '" + sRecipeClass + "' ...");
  // Create a string containing the list of DPE|Value for all the recipe elements
  dyn_string dsDpes = _unRecipeFunctions_joinDynString(
        recipeObject[fwConfigurationDB_RO_DP_NAME], 
        recipeObject[fwConfigurationDB_RO_ELEMENT_NAME], exceptionInfo);
  string sDpeValue = _unRecipeFunctions_getDpeValueString(dsDpes, recipeObject[fwConfigurationDB_RO_VALUE]);
  _unRecipeFunctions_createRecipeInstance(sRecipeClass, sRecipeInstance, sRecipeDesc, 
										  exceptionInfo, sInitial, sDpeValue);
  
  if (dynlen(exceptionInfo)) { 
    dpSet(sRecipeDp+".ProcessInput.State", "");
    DebugTN(exceptionInfo); 
    _unRecipeFunctions_writeInRecipeLog("Error: " + exceptionInfo[2]);
    return; 
  }
  _unRecipeFunctions_writeInRecipeLog("The recipe '" + sRecipeInstance + "' of class '" + sRecipeClass + "' has been created.", TRUE);
  _unRecipeFunctions_loadRecipeInstances(sRecipeClass);
  dpSet(sRecipeDp+".ProcessInput.State", "");
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Delete a recipe instance.
 * @param sRecipeDp - [IN] Data point name of the recipe instance to delete.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_deleteRecipeInstance(string sRecipeDp, dyn_string &exceptionInfo)
{
  int iPos;
  dyn_dyn_mixed recipeObject;
  dyn_string ds;
  dyn_float df;
  string sClassName, sInstanceName;
  bool bInitial;
  
  // Set the recipe State to "Delete"
  dpSet($sDpName+".ProcessInput.State", UN_RCP_INSTANCE_STATE_DELETE);
  
  if (!shapeExists("RecipeClasses") || !shapeExists("RecipeInstances") || 
	  !shapeExists("SelectionProgressBar") || !shapeExists("PreviousRecipeDp") ) 
  {
	dpSet($sDpName+".ProcessInput.State", "");
	fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	return;
  }
  
  _unRecipeFunctions_getRecipeObjectFromDp(sRecipeDp, recipeObject, exceptionInfo);
  if (dynlen(exceptionInfo)) { 
    dpSet($sDpName+".ProcessInput.State", "");
    DebugTN(exceptionInfo);	
    return; 
  }
  
  // Get the inital recipe value of the recipe instance
  bInitial = recipeObject[fwConfigurationDB_RO_META_PREDEFINED][1];
  if (bInitial) {
    _unRecipeFunctions_writeInRecipeLog("Error: The initial recipes can't be deleted.");
    dpSet(sRecipeDp+".ProcessInput.State", "");
    return;
  }
  
  sClassName = recipeObject[fwConfigurationDB_RO_META_CLASSNAME];
  sInstanceName = recipeObject[fwConfigurationDB_RO_META_ORIGNAME];
  if ((iPos=strpos(sInstanceName, "/")) >=0 ) {
    sInstanceName = substr(sInstanceName, iPos+1);
  }
  
  // Asks for user confirmation
  ChildPanelOnCentralReturn("fwGeneral/fwOkCancel.pnl",
							"Delete Recipe",
							makeDynString("$sTitle:Delete recipe ",
										  "$text: Do you want to delete the recipe " + sClassName + "/" + sInstanceName + "?"  ),
							df,
							ds);
  
  if (df[1]<=0)  {
    dpSet(sRecipeDp+".ProcessInput.State", "");
    return;
  }
  
  setValue("LoadedRecipeDp", "text", "");
  unSelectDeselectHMI_select(sRecipeDp, false, exceptionInfo);
  unProgressBar_setPosition("SelectionProgressBar", 0);    
  
  _unRecipeFunctions_deleteRecipeInstances(makeDynString(sRecipeDp));
  _unRecipeFunctions_clearInstanceTables();
  _unRecipeFunctions_loadRecipeInstances(sClassName);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Delete a list of recipe instances.
 * @param dsRecipeDps   - [IN] Data point names of the recipes to delete.
 */
bool _unRecipeFunctions_deleteRecipeInstances(dyn_string dsRecipeDps) {
  int iLen, iPos, rc;
  bool bInitial, bDeleteOk = TRUE;
  string sRecipeDp, sClassName, sInstanceName, sPvssDpe, sSystemName;
  dyn_string recipeInstances, exceptionInfo;
  dyn_dyn_mixed recipeObject;
  
  unImportDevice_initialize();
  
  iLen = dynlen(dsRecipeDps);
  for (int i=1; i<=iLen; i++) {
    sRecipeDp = dsRecipeDps[i];
    sSystemName = unGenericDpFunctions_getSystemName(sRecipeDp);
    
    // Set the recipe State to "Delete"
    dpSet(sRecipeDp+".ProcessInput.State", UN_RCP_INSTANCE_STATE_DELETE);

    _unRecipeFunctions_getRecipeObjectFromDp(sRecipeDp, recipeObject, exceptionInfo);
    if (dynlen(exceptionInfo)) { 
      dpSet(sRecipeDp+".ProcessInput.State", "");
      _unRecipeFunctions_writeInRecipeLog("Exception getting the recipe instance data: " + exceptionInfo[2], true, true);
      dynClear(exceptionInfo);
      bDeleteOk = FALSE;
      continue; 
    }
    
    sClassName = recipeObject[fwConfigurationDB_RO_META_CLASSNAME];
    sInstanceName = recipeObject[fwConfigurationDB_RO_META_ORIGNAME];
    if ((iPos=strpos(sInstanceName, "/")) >=0 ) {
	  sInstanceName = substr(sInstanceName, iPos+1);
    }

    if (sSystemName != getSystemName()) {
      _unRecipeFunctions_writeInRecipeLog("Error: The recipe '" + sClassName + " / " + sInstanceName 
            + "' is defined in a remote system and it can not be deleted.", true, true);
      dpSet(sRecipeDp+".ProcessInput.State", "");
      continue;
    }
    
    // Get the initial recipe value of the recipe instance
    bInitial = recipeObject[fwConfigurationDB_RO_META_PREDEFINED][1];
    if (bInitial) {
      _unRecipeFunctions_writeInRecipeLog("Error: The initial recipe '" + sClassName + " / " + sInstanceName + "' can't be deleted.");
      dpSet(sRecipeDp+".ProcessInput.State", "");
      continue;
    }
    
    _unRecipeFunctions_writeInRecipeLog("Deleting recipe '" + sClassName + " / " + sInstanceName + "' ...");
    // Delete the recipe instance from the fwConfigurationDB
    fwConfigurationDB_deleteRecipeOfClass(sInstanceName, sClassName, exceptionInfo);
    if (dynlen(exceptionInfo)) { 	
      _unRecipeFunctions_writeInRecipeLog("Error deleting the recipe '" + sClassName + " / " + sInstanceName + "' : " + exceptionInfo[2], true, true);
      dpSet(sRecipeDp+".ProcessInput.State", "");
      dynClear(exceptionInfo);
      bDeleteOk = FALSE;
      continue; 
    }
      
    // Get the list of recipe instances of the recipe class
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sClassName, sPvssDpe, exceptionInfo);
    rc = dpGet(sPvssDpe+"ProcessInput.Recipes",recipeInstances);
    if (rc) {
      dpSet(sRecipeDp+".ProcessInput.State", "");
      
      return FALSE;
    }
      
    // Remove the recipe instance of the list of recipes that belong to the recipe class
    if ((iPos=dynContains(recipeInstances, sInstanceName)) > 0) {
      dynRemove(recipeInstances, iPos);
      rc = dpSet(sPvssDpe+"ProcessInput.Recipes", recipeInstances);
      if (rc) {
        dpSet(sRecipeDp+".ProcessInput.State", "");
        bDeleteOk = FALSE;
        continue;	
      }
    }
      
    // Delete the recipe instance
    rc = dpDelete(sRecipeDp);
    if (rc < 0) {
      dpSet(sRecipeDp+".ProcessInput.State", "");   
      bDeleteOk = FALSE;
      continue;
    }
           
    _unRecipeFunctions_writeInRecipeLog("The recipe '" + recipeObject[fwConfigurationDB_RO_META_ORIGNAME] + "' has been deleted.", TRUE); 
  }
  
  unImportDevice_triggerUpdate();
  return bDeleteOk;
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Delete a recipe class and all its recipe instances.
 * Note: The recipe class and its instances will be deleted ONLY if there are no initial recipes of the class.
 * @param sRcpClassDp - [IN] Data point name of the recipe instance to delete.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
public void unRecipeFunctions_deleteRecipeClass(string sRcpClassDp, dyn_string &exceptionInfo) {
  int i, iLen, iRetVal;
  bool bDeleteRecipes = TRUE;
  string sClassName, sInitialRcpDp, sPcoLink, sSystemName;
  dyn_string dsRcpInstances;
  
  unImportDevice_initialize();
  
  // Get the recipe class name
  sClassName = dpGetAlias(sRcpClassDp + ".");
  sSystemName = unGenericDpFunctions_getSystemName(sRcpClassDp);
  
  _unRecipeFunctions_writeInRecipeLog("Removing recipe class '" + sClassName + "' and all its instances ...");
  
  // Get the initial recipe of the recipe class
  iRetVal = _unRecipeFunctions_getInitialRecipeOfClass(sRcpClassDp, sInitialRcpDp, exceptionInfo);
  if (iRetVal != RCP_INITIAL_NO_DEFINED) {
    fwException_raise(exceptionInfo, "ERROR", "The recipe class has some initial recipes defined and it can not be deleted.","");
    _unRecipeFunctions_writeInRecipeLog("Error: The recipe class '" + sClassName + "' has some initial recipes defined and it can not be deleted.");
    return;
  }
  
  dynClear(exceptionInfo);
    
  // Delete the JCOP recipe class and its instances
  fwConfigurationDB_deleteRecipeClass(sSystemName + sClassName, bDeleteRecipes, exceptionInfo);
  if (dynlen(exceptionInfo)) {
    _unRecipeFunctions_writeInRecipeLog("Error deleting the recipe class and its instances: " + exceptionInfo[2], true, true);
    return;
  }
  
  // Get the name of all the recipe instances of the recipe class
  dpGet(sRcpClassDp + ".ProcessInput.Recipes", dsRcpInstances);
  
  // Delete all the recipe instances 
  iLen = dynlen(dsRcpInstances);
  for (i=1; i<=iLen; i++) {
    string sRcpInstanceName, sPvssDpe;
    sRcpInstanceName = sSystemName + sClassName + "/" + dsRcpInstances[i];
    unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sRcpInstanceName, sPvssDpe);  
    dpDelete(sPvssDpe);
  }
  
  // Delete the recipe class
  dpDelete(sRcpClassDp);
  
  // Delete the predefined recipe instance if it exists
  _unRecipeFunctions_deletePredefinedRecipeInstance(sSystemName + sClassName);   
  
  unImportDevice_triggerUpdate();
  
  getValue("PcoList", "text", sPcoLink);
  unRecipeFunctions_loadRecipeClasses(sPcoLink, exceptionInfo);
  
  _unRecipeFunctions_writeInRecipeLog("The recipe class '" + sClassName + "' and all its instances have been removed.", TRUE);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Open the faceplate of a device provided the device alias.
 * @param sAlias        - [IN] Device alias
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_openFaceplate(string sAlias, dyn_string &exceptionInfo) {
  string sDpName;
  unGenericDpFunctions_convert_UNICOSDPE_to_PVSSDPE(sAlias, sDpName);
  
  if (!dpExists(sDpName)) {
    return;
  }
  
  unGenericObject_OpenFaceplate(sDpName, exceptionInfo);
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Creates a dp name that is not existing already
 * @param sDeviceType - [IN] Recipe device type name.
 * @param sFrontEndName - [IN] Front end name where the recipe will be linked.
 * @param sAppName - [IN] Application name.
 * @param sDpName - [OUT] Data point name for the recipe instance.
 * @param iDeviceNumber - [OUT] Device number for the recipe.
 */
void _unRecipeFunctions_createNewRecipeDp(string sDeviceType, string sFrontEndName, string sAppName, string &sDpName, int &iDeviceNumber)
{
  string sRandomNumber, sTmp, sPrefix;
  
  dpGet("_unApplication.devicePrefix", sPrefix);
  sDpName = sPrefix + "-" + sFrontEndName + "-" + sAppName + "-" + sDeviceType + "-";
  
  do {
	  sRandomNumber="";
	  for (int i=1; i<=5; i++) {
		sTmp = rand();
		sRandomNumber += sTmp[strlen(sTmp)-1];    
	  }
  } while (dpExists(sDpName+sRandomNumber));
    
  sDpName += sRandomNumber;
  iDeviceNumber = (int) sRandomNumber;
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Get the data point name of the recipe instance selected in the panel.
 * @param sSelected     - [OUT] DP name of the recipe instance selected in the panel.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 */
void unRecipeFunctions_getSelectedRecipeDp(string &sSelected, dyn_string &exceptionInfo)
{
  sSelected = "";
  string sRecipeClass;
  int iLineCount;
  dyn_int diLines;

  
  if (!shapeExists("RecipeClasses") || !shapeExists("RecipeInstances")) {
  	fwException_raise(exceptionInfo, "ERROR", "Some graphical shapes are missing.","");
	return;
  }
  
  getValue("RecipeClasses", "text", sRecipeClass);
  if (sRecipeClass=="")	{
    return; 
  }
  
  getMultiValue("RecipeInstances", "lineCount", iLineCount,
				"RecipeInstances", "getSelectedLines", diLines);
				
  if (!dynlen(diLines) || diLines[1] >= iLineCount) {
    return;
  }

  getValue("RecipeInstances", "cellValueRC", diLines[1], "RecipeInstancesDP", sSelected);
  
}
//--------------------------------------------------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Methods to edit the color on a table 
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** 
 * Set the background color of a column in a table.
 * @param tableName  - [IN] The name of the table.
 * @param columnName - [IN] The name of the column.
 * @param color		 - [IN] The new color for the column.
 */
void _unRecipeFunctions_setTableColumnBGColor(string tableName, string columnName, string color)
{
  int lines;
  int rc=getValue(tableName, "lineCount", lines);
  if(rc)  return;

  for (int i=0; i<lines; i++) {
    setValue(tableName, "cellBackColRC", i, columnName, color);
  }  
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Set the background color of a row in a table.
 * @param tableName - [IN] The name of the table.
 * @param rowIndex  - [IN] The row index.
 * @param color		- [IN] The new color for the row.
 */
void _unRecipeFunctions_setTableRowBGColor(string tableName, int rowIndex, string color)
{
  int cols, rows; 
  int rc=getMultiValue(tableName, "columnCount", cols,
                       tableName, "lineCount", rows);
  if (rc) {
    return;
  }
  
  if (rowIndex >= rows) {
    return;
  }
  
  for (int i=0; i<cols; i++) {
	string colName;
    getValue(tableName, "columnName", i, colName);
    setValue(tableName, "cellBackColRC", rowIndex, colName, color);
  }  
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Set the background color of a cell in a table.
 * @param tableName  - [IN] The name of the table.
 * @param rowIndex   - [IN] The row index.
 * @param columnName - [IN] The name of the column.
 * @param color		 - [IN] The new color for the row.
 */
void _unRecipeFunctions_setTableCellBGColor(string tableName, int rowIndex, string columnName, string color)
{
  int lines;
  int rc=getValue(tableName, "lineCount", lines);
  if(rc) {
    return;
  }

  setValue(tableName, "cellBackColRC", rowIndex, columnName, color);
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Restore the original colors in the recipe elements table (recipe instance panel).
 */
void _unRecipeFunctions_resetRecipeElementsBgColor() {
  int lines;

  getValue("RecipeElements", "lineCount", lines);
  for (int i=1; i<=lines; i++){
    _unRecipeFunctions_setTableRowBGColor("RecipeElements", (i-1), CELL_BG_COLOR);
	// Resets the font style
	RecipeElements.rowFontType (i-1, makeDynInt(0));
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Set the background color of the rows in the RecipeElements table that contains invalid values.
 * @param rows       - [IN] Indices of the invalid rows.
 * @param sTableName - [IN] Name of the table where the invalid rows must be set.
 */
void _unRecipeFunctions_setInvalidRows(dyn_string rows, string sTableName="RecipeElements") {
  int len=dynlen(rows);
  for (int i=1; i<=len; i++) {
    int row = rows[i];
    _unRecipeFunctions_setTableRowBGColor(sTableName, (row-1), CELL_ERROR_COLOR); 
  }
}

//------------------------------------------------------------------------------------------------------------------------
/** 
 * Method used to trim a label to the appropriate size.
 * @param hiddenShape - [IN] Hidden shape (label) to calculate the width of the string.
 * @param aString 	  - [IN] String to be trimmed if its width is bigger than the size specified.
 * @param width		  - [IN] Width of the shape where the label will be placed.
 * @return A new string that fits in the specified width.
 */
string _unRecipeFunctions_adjustStringToSize(shape hiddenShape, string aString, int width) {
	int x, y;
	string sTmpStr = aString;
	bool bCalculateNewSize, bResized = false;

	do {
		setValue(hiddenShape, "text", sTmpStr);
		getValue(hiddenShape, "size", x, y);
		bCalculateNewSize = (x>width-8);
		if (bCalculateNewSize) {
			sTmpStr = substr(sTmpStr, 0, strlen(sTmpStr)-1);
			bResized = true;
		}
	} while (bCalculateNewSize);
	
	if (bResized) {
	  sTmpStr = sTmpStr + "..";
	}
	
	return sTmpStr;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Compares the elements contained in two lists.
 * @param daList1Orig      - [IN]  First list of elements.
 * @param daList2Orig      - [IN]  Second list of elements.
 * @param daMissingDpNames - [OUT] List of elements contained in daList1 and not contained in daList2.
 * @param daExtraDpNames   - [OUT] List of elements contained in daList2 and not contained in daList1.
 * @return TRUE if the lists contain the same elements, otherwise FALSE.
 */
bool unRecipeFunctions_compareListElements(dyn_anytype daList1Orig, dyn_anytype daList2Orig, dyn_anytype &daMissingDpNames, dyn_anytype &daExtraDpNames, bool bIgnoreSystemName=TRUE) {
  int iLen;
  dyn_anytype daList1, daList2;
  
  if (bIgnoreSystemName) {
    daList1 = _unRecipeFunctions_removeSystemNameFromList(daList1Orig);
    daList2 = _unRecipeFunctions_removeSystemNameFromList(daList2Orig);
  } else {
    daList1 = daList1Orig;
    daList2 = daList2Orig;
  }

  // Look for the elements contained in daList1 and missing in daList2
  iLen = dynlen(daList1);
  for (int i=1; i<=iLen; i++) {
    if (!dynContains(daList2, daList1[i])) {
      dynAppend(daMissingDpNames, daList1Orig[i]);
    }
  }
    
  // Look for the elements contained in daList2 and missing in daList1
  iLen = dynlen(daList2);
  for (int i=1; i<=iLen; i++) {
    if (!dynContains(daList1, daList2[i])) {
      dynAppend(daExtraDpNames, daList2Orig[i]);
    }
  }
  
  return (dynlen(daMissingDpNames) == 0 && dynlen(daExtraDpNames) == 0);
}
//------------------------------------------------------------------------------------------------------------------------
dyn_anytype _unRecipeFunctions_removeSystemNameFromList(dyn_anytype daList) {
  int iLen = dynlen(daList);
  dyn_anytype daRetList;
  
  for (int i=1; i<=iLen; i++) {
    dynAppend(daRetList, unExportTree_removeSystemName(daList[i]));
  }
  
  return daRetList;
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Print a message in the recipe log containing a list of devices (if any).
 * @param dsDeviceDps - [IN] List of device DPs.
 * @param sMessage    - [IN] Message to print if the list is not empty.
 */
void _unRecipeFunctions_printDeviceList(dyn_string dsDeviceDps, string sMessage) {
  int iLen;
  string sRecipeElement;
  
  iLen = dynlen(dsDeviceDps);
  if (iLen == 0) {
    return;
  }
  
  for (int i=1; i<=iLen; i++) {
    unGenericDpFunctions_convert_PVSSDPE_to_UNICOSDPE(dsDeviceDps[i], sRecipeElement);
    sMessage += "\n" + sRecipeElement;
  }
  _unRecipeFunctions_writeInRecipeLog(sMessage, true, true);
}
//------------------------------------------------------------------------------------------------------------------------
/**
 * Function used to set a value in a list of DPEs.
 * @param dsDpNames     - [IN]  List of DP names where to set the values.
 * @param sAppendStr    - [IN]  String to append to the DP names.
 * @param value         - [IN]  Value to set the the DPEs.
 * @param exceptionInfo - [OUT] Standard exception handling variable.
 * @return TRUE if the values were set successfully, otherwise FALSE.
 */
bool unRecipeFunctions_dpMultiSet(dyn_string dsDpNames, string sAppendStr, anytype value, dyn_string &exceptionInfo) {
  int iLen;
  dyn_anytype daValues;
  
  iLen = dynlen(dsDpNames);
  for (int i=1; i<=iLen; i++) {
    dsDpNames[i] += sAppendStr;
    dynAppend(daValues, value);
  }
  
  int iResult = dpSet(dsDpNames, daValues);
  return _unRecipeFunctions_handleDpGetSetError (iResult, exceptionInfo);
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Internal function to normalize a datapoint name (the trailing dot will be removed if it exists).
 * @param sDpName - [IN/OUT] The datapoint name to normalize.
 */
void _unRecipeFunctions_normalizeDp(string &sDpName) {
  int iLen = strlen(sDpName);
  if (iLen==0) {
    return;
  }
  
  if (sDpName[(iLen-1)] == ".") {
    sDpName = substr(sDpName, 0, iLen-1);
  }
}
//------------------------------------------------------------------------------------------------------------------------
/** 
 * Export multiple recipes to a file (comma separated)
 * @param dsRcpInstList [IN] List of recipe names to export.
 */ 
void unRecipeFunctions_exportMultipleRecipes(dyn_string dsRcpInstList) {
  string sExtension, sPath, sFileType, sFileName, sRcpFileName;
  dyn_string exceptionInfo;
  
  if (!dynlen(dsRcpInstList)) {
    _unRecipeFunctions_writeInRecipeLog("No recipe instances selected to export.");
    return;
  }
  
  // 1. Select file
  if (_WIN32) {
    sFileType="xls";
  } else {
    sFileType="txt";
  }
  
  sExtension="." + sFileType; 
  sPath=getPath(DATA_REL_PATH); 
  if (folderSelector(sPath) == 0) {
    return;
  } 
  
  int iLen = dynlen(dsRcpInstList);
  for (int i=1; i<=iLen; i++) {
    string sRcpName, sDpName;
    dyn_dyn_mixed recipeObject;
    
    sDpName  = dsRcpInstList[i];
    sRcpName = dpGetAlias(sDpName + ".");    
    _unRecipeFunctions_getRecipeObjectFromDp(sDpName, recipeObject, exceptionInfo);
    if (dynlen(exceptionInfo)) { 
    	_unRecipeFunctions_writeInRecipeLog("Error loading the recipe : " +exceptionInfo[2]);
    	DebugTN(exceptionInfo); 
    	continue; 
    }
    
    sRcpFileName = sRcpName;
    strreplace(sRcpFileName, "/", "_");
    sFileName = sPath + "/" + sRcpFileName + sExtension;
    file f = fopen(sFileName, "w");
    if (f == 0) {
      _unRecipeFunctions_writeInRecipeLog("The file " + sFileName + " could not be opened for writing");
  	  continue;
    }        

    int jLen = dynlen(recipeObject[fwConfigurationDB_RO_DP_NAME]);
    for (int j=1; j<=jLen; j++) {  
    	dyn_string tmp;
      string dpName, alias, deviceType, description = " : ", dpe, unit="", range="";    
   
      dpName = recipeObject[fwConfigurationDB_RO_DPE_NAME][j];
      alias = recipeObject[fwConfigurationDB_RO_DP_NAME][j];
    	tmp = strsplit(dpName, ".");
    	dpe = tmp[dynlen(tmp)];
    
      if (dpExists(dpName)) {
        description = unGenericDpFunctions_getDescription(unGenericDpFunctions_getDpName(dpName)) + " : " + dpe;
        range = unRecipeFunctions_getDeviceRange(recipeObject[fwConfigurationDB_RO_DPE_NAME][j], exceptionInfo);
        unit = unRecipeFunctions_getDpeUnit(dpName, exceptionInfo);
      } 
    
      int chars = fprintf(f,"%d\t%s\t%s\t%s\t%s\t%s\n", j, alias, description, recipeObject[fwConfigurationDB_RO_VALUE][j], unit, range);
      if (chars<0 || chars==EOF) {
        _unRecipeFunctions_writeInRecipeLog("Error writing to file: " + sFileName);
        break;
      }
    }
    
    fclose(f);
    _unRecipeFunctions_writeInRecipeLog("Recipe instance '" + sRcpName + "' saved to: " + sFileName);
  } 
}

//------------------------------------------------------------------------------------------------------------------------
//@}
